if (PlotType <= 3)
  yrbcolormap;
  grid on;
  axis([-3 3 -1 1 -1 1]);
  daspect([1 1 1]);
  caxis([1 1.25]);

  hideslices;
  showslices('x',2);
  showslices('y',2);
  showslices('z',2);


  th = (0:100)*2*pi/100;
  hold on;
  cx = 0.8*cos(th);
  cy = 0.8*sin(th);
  p_circley = plot3(cx,0*th - 0.05, cy,'k--','LineWidth',2);
  % p_circlez = plot3(cx,cy,0*th + 0.05,'k--','LineWidth',2);

  % showpatchborders;
  str = sprintf('Simple wave at t = %4.2f',t);
  title(str,'FontSize',16);
  xlabel('X','FontSize',16);
  ylabel('Y','FontSize',16);
  set(gca,'FontSize',14);
  hold off;

elseif (PlotType == 4)
  axis([-3 3 0.95 1.35]);
  hold on;

  base_dir = './qref_chombo/';

  uf = '';
  mxc = amrdata(1).mx;
  dxc = amrdata(1).dx;
  rr = dxc/amrdata(end).dx;
  mx = mxc*rr;

  str2 = {};
  dir = [base_dir,'qref_',num2str(mx),'/'];
  [amrdata1d,t1d] = readamrdata(1,Frame/plot_interval,dir);
  if (abs(t1d - t) > 1e-10)
    error('times are incompatible');
  end;
  [q1d,x,p(1)] = plotframe1ez(amrdata1d,1,'r-',uf);
  set(p(1),'LineWidth',2);
  str2{1} = sprintf('1d solution (mx = %d)',mx);

  dir = [base_dir,'qref_fine'];
  [amrdata1d,t1d] = readamrdata(1,Frame/plot_interval,dir);
  if (abs(t1d - t) > 1e-10)
    error('times are incompatible');
  end;
  [q1d,x,p(2)] = plotframe1ez(amrdata1d,1,'k-',uf);
  s1 = sprintf('Fine grid solution');
  str2{2} = s1;
  set(p(2),'LineWidth',2);
  h2 = p;

  h = getlegendinfo;
  str = {};
  for i = 1:length(h),
    str{i} = sprintf('Level %d',i);
  end;

  s1 = cell(0);
  lh = legend([h h2],{str{:} str2{:}});

  set(lh,'FontSize',14);
  hold off;
end;

if (Frame/plot_interval >= 0)
  mx = amrdata(1).mx;
  my = amrdata(1).my;
  mz = amrdata(1).mz;
  dx = amrdata(1).dx;
  dy = amrdata(1).dy;
  dz = amrdata(1).dz;
  xcenter = (0:mx-1)*dx + dx/2;
  ycenter = (0:my-1)*dy + dy/2;
  zcenter = (0:mz-1)*dz + dz/2;
  qmesh = amrdata(1).q;
  kname = ['kappa_',num2str(mx)];
  eval(['load ',kname,'.out;']);
  eval(['kappa = ',kname,';']);
  kappa = permute(reshape(kappa,mx,my,mz),[2 1 3]);
  err_norm = zeros(3,1);
  [en,diff] = compute_err(qmesh,xcenter,ycenter,zcenter,...
      dx,dy,dz,kappa,t,err_norm,level,Frame/plot_interval);
  en(2) = sqrt(en(2));
  fprintf('%16.8e %16.8e %16.8e\n',en(1),en(2),en(3));
end;

clear afterframe;
clear map1d;
