      subroutine compute_basis(z,rot)
      implicit double precision (a-h,o-z)

      double precision z(0:1,0:1,0:1,3)

      double precision Jb(3,3), Jinv(3,3), xcv(3), ycv(3), zcv(3)
      double precision eye(3,3), rot(3,3,3), v(3,3), alpha(3),
     &      rhs(3,3)
      integer IPIV(3)

      common /zpoints/ z000(3),z100(3),z010(3),z001(3), z110(3),
     &      z101(3), z011(3), z111(3)

      common /coeffs/ a000(3),a100(3),a010(3),a001(3), a110(3),
     &      a101(3), a011(3), a111(3)


      do i = 1,3
         do j = 1,3
            if (i == j) then
               eye(i,j) = 1.d0
            else
               eye(i,j) = 0.d0
            endif
         enddo
      enddo


c     Get centers of faces so we can find basis vectors at each face.
      do i = 1,3
         xcv(i) = 0.5d0
         ycv(i) = 0.5d0
         zcv(i) = 0.5d0
      enddo

      xcv(1) = 0.0
      ycv(2) = 0.0
      zcv(3) = 0.0

c     # Make notation easy..
      do i = 1,3
         z000(i) = z(0,0,0,i)
         z100(i) = z(1,0,0,i)
         z010(i) = z(0,1,0,i)
         z001(i) = z(0,0,1,i)
         z110(i) = z(1,1,0,i)
         z101(i) = z(1,0,1,i)
         z011(i) = z(0,1,1,i)
         z111(i) = z(1,1,1,i)
      enddo

c     # Get coefficients for trilinear map.
      do i = 1,3
         a000(i) = z000(i)
         a001(i) = z001(i) - z000(i)
         a010(i) = z010(i) - z000(i)
         a100(i) = z100(i) - z000(i)

         a111(i) = z111(i) - z110(i) - z101(i) -
     &         z011(i) + z100(i) + z010(i) + z001(i) - z000(i)
         a011(i) = z011(i) - z010(i) - z001(i) + z000(i)
         a101(i) = z101(i) - z100(i) - z001(i) + z000(i)
         a110(i) = z110(i) - z100(i) - z010(i) + z000(i)
      enddo

c     # Start computing basis vectors.
      do n = 1,3  !! Loop over each face.

         do i = 1,3 !! Get three rows of Jacobian
            Jb(i,1) = a100(i) + a110(i)*ycv(n)  + a101(i)*zcv(n) +
     &            a111(i)*ycv(n)*zcv(n)
            Jb(i,2) = a010(i) + a110(i)*xcv(n)  + a011(i)*zcv(n) +
     &            a111(i)*xcv(n)*zcv(n)
            Jb(i,3) = a001(i) + a101(i)*xcv(n)  + a011(i)*ycv(n) +
     &            a111(i)*xcv(n)*ycv(n)
         enddo

         do i = 1,3
            IPIV(i) = 0
            do j = 1,3
               rhs(i,j) = eye(i,j)
            enddo
         enddo

c        # Compute inverse of Jacobian.
         call dgesv(3,3,Jb,3,IPIV,rhs,3,info)

         if (info .ne. 0) then
c           # The Jacobian is singular;  set rotation to identity and hope
C           # that it is never really needed in other clawpack calculations...
            do i = 1,3
               ni = mod(i+n-2,3) + 1
               do j = 1,3
                  rot(n,i,j) = eye(ni,j)
               enddo
            enddo

c           # Skip to next face (cycle through n loop)
            cycle

         endif

c        # Now we have to permute Jinv to get normal to face n in first row.
C        # We then orthogonalize other rows against this vector.
         do i = 1,3
            ni = mod(i+n-2,3) + 1
            do j = 1,3
               Jinv(i,j) = rhs(ni,j)
            enddo
         enddo

c        # Do Gram-Schmidt on rows of Jinv, to get orthogonal basis.
         do i = 1,3

c           # Project onto each previous vector v(i,:)
            alpha(1) = 0
            do k = 1,i-1
               alpha(k) = 0
               do j = 1,3
                  alpha(k) = alpha(k) + Jinv(i,j)*v(k,j)
               enddo
            enddo

            sum = 0
            do j = 1,3
               v(i,j) = Jinv(i,j)
               do k = 1,i-1
                  v(i,j) = v(i,j) - alpha(k)*v(k,j)
               enddo
               sum = sum + v(i,j)*v(i,j)
            enddo

            do j = 1,3
               v(i,j) = v(i,j)/dsqrt(sum)
            enddo
         enddo


c        # We now have our three basis vectors for face n. Now, permute them
C        back so that row 1 is always normal in direction xi, row 2 in
C        direction eta, and row3 in direction zeta.

         do i = 1,3
            ni = mod(i+n-2,3) + 1
            do j = 1,3
               rot(n,ni,j) = v(i,j)
            enddo
         enddo
      enddo

      end


      subroutine compute_volume(cube,volume)
      implicit double precision (a-h,o-z)


      double precision dot_cross
      double precision cube(0:1,0:1,0:1,3)
      double precision u(3),v(3),w(3), p(3), q(3), r(3)

      do j = 1,3
         u(j) = cube(0,1,1,j) - cube(0,0,0,j)
         v(j) = cube(1,0,1,j) - cube(0,0,0,j)
         w(j) = cube(1,1,0,j) - cube(0,0,0,j)
      enddo
      volume = dabs(dot_cross(u,v,w))/6.d0

      do j = 1,3
         u(j) = cube(0,0,1,j) - cube(1,1,1,j)
         v(j) = cube(0,1,0,j) - cube(1,1,1,j)
         w(j) = cube(1,0,0,j) - cube(1,1,1,j)
      enddo
      volume = volume + dabs(dot_cross(u,v,w))/6.d0

      do i = 0,1
         do j = 0,1
            do k = 0,1
               ip = mod(i+1,2)
               jp = mod(j+1,2)
               kp = mod(k+1,2)

               do m = 1,3
                  p(m) = cube(ip,j,k,m) - cube(i,j,k,m)
                  q(m) = cube(i,jp,k,m) - cube(i,j,k,m)
                  r(m) = cube(i,j,kp,m) - cube(i,j,k,m)
               enddo

               volume = volume + dabs(dot_cross(p,q,r))/6.d0
            enddo
         enddo
      enddo

      volume = volume/2.d0

      end

      double precision function dot_cross(u,v,w)
      implicit double precision (a-h,o-z)
      double precision u(3), v(3), w(3)

      dot_cross = u(1)*(v(2)*w(3) - v(3)*w(2)) -
     &            u(2)*(v(1)*w(3) - v(3)*w(1)) +
     &            u(3)*(v(1)*w(2) - v(2)*w(1))

      return
      end


c     # This will give us the exact surface area of each face in the case
c     # the surface is flat.  Otherwise, this seems like it should give a
c     # reasonable answer, but this is merely a conjecture!

      subroutine compute_surf_area(cube,area)
      implicit double precision (a-h,o-z)

      double precision dot_cross
      double precision cube(0:1,0:1,0:1,3)
      double precision w(3), p(3), q(3), area(3)
      double precision quad(0:1,0:1,3)

      do n = 1,3
         do i = 0,1
            do j = 0,1
               do m = 1,3
                  if (n == 1) then
                     quad(i,j,m) = cube(0,i,j,m)
                  elseif (n == 2) then
                     quad(i,j,m) = cube(i,0,j,m)
                  else
                     quad(i,j,m) = cube(i,j,0,m)
                  endif
               enddo
            enddo
         enddo

         area(n) = 0
         do i = 0,1
            do j = 0,1
               ip = mod(i+1,2)
               jp = mod(j+1,2)

               do m = 1,3
                  p(m) = quad(ip,j,m) - quad(i,j,m)
                  q(m) = quad(i,jp,m) - quad(i,j,m)
               enddo
               w(1) =   p(2)*q(3) - p(3)*q(2)
               w(2) = -(p(1)*q(3) - p(3)*q(1))
               w(3) =   p(1)*q(2) - p(2)*q(1)

               d = w(1)*w(1) + w(2)*w(2) + w(3)*w(3)
               area(n) = area(n) + dsqrt(d)/2.d0
            enddo
         enddo
         area(n) = area(n)/2.d0

      enddo !! face loop

      end
