C     -*- Mode: Fortran; -*-
c-----------------------------------------------------------------
c     Ravi Samtaney
c     Copyright 2002
c     Princeton Plasma Physics Laboratory
c     All Rights Reserved
c-----------------------------------------------------------------
c     $Log: init.F,v $
c     Revision 1.4  2004/11/23 00:54:17  noel
c     Trying to keep in-line with repo
c
c     Revision 1.2  2002/07/09 19:00:59  samtaney
c     Intermediate commit of files towards a single 2D/3D Ideal MHD code.
c
c     Revision 1.1  2002/06/13 22:16:18  samtaney
c     Original unsplit 8-wave Ideal MHD source (includes B filter).
c
c-----------------------------------------------------------------
c     Subroutine to set up the initial conditions for the
c     "Rotor" problem. 
c---------------------------------------------------------------------------
c     
      subroutine RotorInitialConditions(ux,
     &     iulo,iuhi,julo,juhi,kulo,kuhi,
     &     ilo,ihi,jlo,jhi,
     &     klo,khi,
     &     adx,ady,adz)
c----------------------------------------------------------------------
c
      implicit none
c
#include "EosCommon.fh"
c
c     Incoming arguments
      integer ilo,ihi,jlo,jhi,klo,khi
      integer iulo,iuhi,julo,juhi,kulo,kuhi
      real*8 adx,ady,adz
      real*8 ux(iulo:iuhi,
     &     julo:juhi,
     &     kulo:kuhi,8)
c     
c     Local variables
      integer i,j,k
      real*8 xl,yl,xr,yr
      real*8 rho,press, Bi, Bj, Bk, u,v,w
      real*8 rho1,rho2,rad0,rad1,rad,fac,v0
c     
      real*8 x,y
c     
      write(6,*) 'ROTOR INIT',ilo,ihi,jlo,jhi
      write(6,*) 'ROTOR INIT',iulo,iuhi,julo,juhi
      write(6,*) 'ROTOR INIT',adx,ady,adz,8
      write(6,*) 'ROTOR INIT',gamma,rgas,wmol
c     
      xl=-1.D0
      xr=1.D0
      yl=-1.D0
      yr=1.D0       
c      xl=-0.5D0
c      xr=0.5D0
c      yl=-0.5D0
c      yr=0.5D0       

      rho1=10.D0
      rho2=1.D0
      press=0.5D0
      v0=1.D0
      rad0=0.1D0
      rad1=0.115D0
      w=0.D0
      Bi=1.0D0
      bj=0.00D0
      bk=0.00D0
c     
      do k=klo,khi,1
         do j=jlo,jhi,1
            y = yl+(j - 0.5D0)*ady
            do i=ilo,ihi,1
               x = xl+(i - 0.5D0)*adx
               rad=dsqrt(x*x+y*y)
               fac=(rad1-rad)/(rad1-rad0)
               if(rad.le.rad0) then
                  rho=rho1
                  u=-v0*(y)/rad0
                  v=v0*(x)/rad0
               else 
                  rho=rho2
                  u=0.D0
                  v=0.D0
               endif
c               u=v0
c               v=0.D0
               ux(i,j,k,1)=rho
               ux(i,j,k,2)=rho*u
               ux(i,j,k,3)=rho*v
               ux(i,j,k,4)=rho*w
               ux(i,j,k,5)=Bi
               ux(i,j,k,6)=Bj
               ux(i,j,k,7)=Bk
               ux(i,j,k,8)=press/(gamma-1.D0)+
     &              0.5*rho*(u*u+v*v+w*w)+0.5D0*(bi*bi+bj*bj+bk*bk)
            enddo
         enddo
      enddo
      
      return
      end







