C     -*- Mode: Fortran; -*-
c-----------------------------------------------------------------------
c     Ravi Samtaney
c     Copyright 2002
c     Princeton Plasma Physics Laboratory
c     All Rights Reserved
c-----------------------------------------------------------------------
c     $Log: inviscidfluxrp.F,v $
c     Revision 1.4  2004/11/23 00:54:17  noel
c     Trying to keep in-line with repo
c
c     Revision 1.2  2002/07/09 19:00:59  samtaney
c     Intermediate commit of files towards a single 2D/3D Ideal MHD code.
c
c     Revision 1.1  2002/06/13 22:16:18  samtaney
c     Original unsplit 8-wave Ideal MHD source (includes B filter).
c
c-----------------------------------------------------------------------
      subroutine RiemannSolverMHD(um,ul,ur,
     &     iumlo,iumhi,jumlo,jumhi,kumlo,kumhi,
     &     iullo,iulhi,jullo,julhi,kullo,kulhi,
     &     iurlo,iurhi,jurlo,jurhi,kurlo,kurhi,
     &     ilo,ihi,jlo,jhi,klo,khi,
     &     direction,llo,lhi)
c
      implicit none
c
#include "EosCommon.fh"
c     
c     incoming arguments
      integer iumlo,iumhi,jumlo,jumhi,kumlo,kumhi,
     &     iullo,iulhi,jullo,julhi,kullo,kulhi,
     &     iurlo,iurhi,jurlo,jurhi,kurlo,kurhi,
     &     ilo,ihi,jlo,jhi,klo,khi,
     &     direction,llo,lhi
c
      real*8 ul(iullo:iulhi,jullo:julhi,kullo:kulhi,8)
      real*8 ur(iurlo:iurhi,jurlo:jurhi,kurlo:kurhi,8)
      real*8 um(iumlo:iumhi,jumlo:jumhi,kumlo:kumhi,8)
c
c     Local variables
      integer i,j,k,m
c     
      real*8 vm(llo:lhi,8)
      real*8 vl(llo:lhi,8)
      real*8 vr(llo:lhi,8)
c
c     This subroutine determines the solution to the Riemann problem
c     at the cell interfaces. The solution is based upon linearization
c     and no guarantee can be made that it is accurate.
c     v: Vector of variables - rho, u,v,w,Bx,By,Bz,p
c     rho=density, u=x-velocity, v=y-velocity
c     w=z-velocity
c     Bx,By,Bz=x,y,z components of the magnetic field
c     p=thermal pressure
c     
c     vl(i) = Left state at the i-th interface
c     vr(i) = Right state at the i-th interface
c     
c     
c$$$      write(6,*) 'RiemannSolverMHD', iumlo,iumhi,jumlo,jumhi,
c$$$     &     iullo,iulhi,jullo,julhi,
c$$$     &     iurlo,iurhi,jurlo,jurhi,
c$$$     &     ilo,ihi,jlo,jhi,klo,khi,
c$$$     &     direction,llo,lhi
      if(direction.eq.0) then
         do k=klo,khi,1
            do j=jlo,jhi,1
               do i=ilo,ihi,1
                  do m=1,8,1
                     vl(i,m)=ul(i,j,k,m)
                     vr(i,m)=ur(i,j,k,m)
                  enddo
               enddo
               call RiemannProblem(vm,vl,vr,ilo,ihi,direction)
               do i=ilo,ihi,1
                  do m=1,8,1
                     um(i,j,k,m)=vm(i,m)
                  enddo
               enddo
            enddo
         enddo
      elseif (direction.eq.1) then
         do k=klo,khi,1
            do i=ilo,ihi,1
               do j=jlo,jhi,1
                  do m=1,8,1
                     vl(j,m)=ul(i,j,k,m)
                     vr(j,m)=ur(i,j,k,m)
                  enddo
               enddo
               call RemapToLocalCoords(vl,jlo,jhi,2)
               call RemapToLocalCoords(vr,jlo,jhi,2)
               call RiemannProblem(vm,vl,vr,jlo,jhi,direction)
               call RemapToGlobalCoords(vm,jlo,jhi,2)
               do j=jlo,jhi,1
                  do m=1,8,1
                     um(i,j,k,m)=vm(j,m)
                  enddo
               enddo
            enddo
         enddo
      elseif (direction.eq.2) then
         do j=jlo,jhi,1
            do i=ilo,ihi,1
               do k=klo,khi,1
                  do m=1,8,1
                     vl(k,m)=ul(i,j,k,m)
                     vr(k,m)=ur(i,j,k,m)
                  enddo
               enddo
               call RemapToLocalCoords(vl,klo,khi,3)
               call RemapToLocalCoords(vr,klo,khi,3)
               call RiemannProblem(vm,vl,vr,klo,khi,direction)
               call RemapToGlobalCoords(vm,klo,khi,3)
               do k=klo,khi,1
                  do m=1,8,1
                     um(i,j,k,m)=vm(k,m)
                  enddo
               enddo
            enddo
         enddo
      else
         write(6,*) 'FATAL ERROR: direcion !=0,1,2'
c     
      endif
c     
      return
      end
c     
c-----------------------------------------------------------------------
      subroutine RiemannProblem(vm,vl,vr,ilo,ihi,direction)
c     
      implicit none
c     
#include "EosCommon.fh"
c     
c     Incoming arguments
      integer ilo, ihi, direction
      real*8 vl(ilo:ihi,8)
      real*8 vr(ilo:ihi,8)
      real*8 vm(ilo:ihi,8)
c     
c     Local variables
      real*8 alamda(ILO:IHI,8)
      real*8 evl(ILO:IHI,8,8)
      real*8 evr(ILO:IHI,8,8)
      real*8 durl(ILO:IHI,8)
      real*8 alpha(ILO:IHI,8)
      real*8 utilde(ilo:ihi,8)
      integer i,m
c     
c     v: Vector of variables - rho, u,v,w,Bx,By,Bz,p
c     rho=density, u=x-velocity, v=y-velocity
c     w=z-velocity
c     Bx,By,Bz= Magnetic Field
c     p=pressure
c     
c     utilde = vector of Roe averaged variables.
c     utilde(1,2,3,4,5,6,7,8) = {rhotilde, utilde, vtilde, wtilde, 
c     Bxtilde,Bytilde,Bztilde,ptilde}
c
c     Set the Roe averaged variables depending  on vl,vr
      call SetRoeVariables(utilde,vl,vr,ilo,ihi,direction)
c     
      do i=ilo,ihi,1
         do m=1,8,1
            alamda(i,m)=0.D0
            durl(i,m)=vr(i,m)-vl(i,m)
         enddo
      enddo
c     
      call SetEigenSystem(utilde,evl,evr,alamda,ilo,ihi,direction)
      call SetAlphas(utilde,alpha,durl,evl,evr,ilo,ihi,direction)
      call RPSolve(evl,evr,alamda,alpha,vl,vr,vm,ilo,ihi)
c     
      return
      end

c-----------------------------------------------------------------------
c     Routine to map velocities to local coordinates.
      subroutine RemapToLocalCoords(vx,ilo,ihi,direction)
c     
      implicit none
c     
#include "EosCommon.fh"
c     
c     
c     Incoming arguments
      integer ilo, ihi
      integer direction
      real*8 vx(ilo:ihi,8)
c     
c     Local variables
      real*8 tmp(ilo:ihi,3)
      integer i
c     
      if(direction.eq.2) then
         do i=ilo,ihi,1
            tmp(i,1)=vx(i,3)
            tmp(i,2)=vx(i,4)
            tmp(i,3)=vx(i,2)
            vx(i,2)=tmp(i,1)
            vx(i,3)=tmp(i,2)
            vx(i,4)=tmp(i,3)
            tmp(i,1)=vx(i,6)
            tmp(i,2)=vx(i,7)
            tmp(i,3)=vx(i,5)
            vx(i,5)=tmp(i,1)
            vx(i,6)=tmp(i,2)
            vx(i,7)=tmp(i,3)
         enddo
      elseif(direction.eq.3) then
         do i=ilo,ihi,1
            tmp(i,1)=vx(i,4)
            tmp(i,2)=vx(i,2)
            tmp(i,3)=vx(i,3)
            vx(i,2)=tmp(i,1)
            vx(i,3)=tmp(i,2)
            vx(i,4)=tmp(i,3)
            tmp(i,1)=vx(i,7)
            tmp(i,2)=vx(i,5)
            tmp(i,3)=vx(i,6)
            vx(i,5)=tmp(i,1)
            vx(i,6)=tmp(i,2)
            vx(i,7)=tmp(i,3)
         enddo
      endif
c     
      return
      end
c     
c-----------------------------------------------------------------------
      subroutine RemapToGlobalCoords(vx,ilo,ihi,direction)
      implicit none
c
#include "EosCommon.fh"
c     
c     Routine to map velocities to global coordinates.
c     Incoming arguments
      integer ilo, ihi
      integer direction
      real*8 vx(ilo:ihi,8)
c     
c     Local variables
      real*8 tmp(ilo:ihi,3)
      integer i
c     
      if(direction.eq.2) then
         do i=ilo,ihi,1
            tmp(i,2)=vx(i,2)
            tmp(i,3)=vx(i,3)
            tmp(i,1)=vx(i,4)
            vx(i,2)=tmp(i,1)
            vx(i,3)=tmp(i,2)
            vx(i,4)=tmp(i,3)
            tmp(i,2)=vx(i,5)
            tmp(i,3)=vx(i,6)
            tmp(i,1)=vx(i,7)
            vx(i,5)=tmp(i,1)
            vx(i,6)=tmp(i,2)
            vx(i,7)=tmp(i,3)
         enddo
      elseif(direction.eq.3) then
         do i=ilo,ihi,1
            tmp(i,3)=vx(i,2)
            tmp(i,1)=vx(i,3)
            tmp(i,2)=vx(i,4)
            vx(i,2)=tmp(i,1)
            vx(i,3)=tmp(i,2)
            vx(i,4)=tmp(i,3)
            tmp(i,3)=vx(i,5)
            tmp(i,1)=vx(i,6)
            tmp(i,2)=vx(i,7)
            vx(i,5)=tmp(i,1)
            vx(i,6)=tmp(i,2)
            vx(i,7)=tmp(i,3)
         enddo
      endif
c     
      return
      end
c     
c-----------------------------------------------------------------------
c     The actual Riemann solver
      subroutine RPSolve(evl,evr,alamda,alpha,vl,vr,vm,ilo,ihi)
c
      implicit none
c
#include "EosCommon.fh"
c
c     Incoming arguments
      integer ilo,ihi
      real*8 vl(ilo:ihi,8)
      real*8 vr(ilo:ihi,8)
      real*8 vm(ilo:ihi,8)
      real*8 alamda(ILO:IHI,8)
      real*8 evl(ILO:IHI,8,8)
      real*8 evr(ILO:IHI,8,8)
      real*8 alpha(ILO:IHI,8)
c
c     Local variables
      integer i,k,l,m
      real*8 vml(ilo:ihi,8)
      real*8 vmr(ilo:ihi,8)
c     
c     
      do i=ilo,ihi,1
         do m=1,8,1
            vmr(i,m)=vr(i,m)
            vml(i,m)=vl(i,m)
         enddo
      enddo
      do i=ilo,ihi,1
c     do l=1,8,1
         do k=1,8,1
            if(alamda(i,k).ge.0.D0) then
               do l=1,8,1
                  vmr(i,l)=vmr(i,l)-alpha(i,k)*evr(i,l,k)
               enddo
            else
               do l=1,8,1
                  vml(i,l)=vml(i,l)+alpha(i,k)*evr(i,l,k)
               enddo
            endif
         enddo
c     enddo
      enddo
c     
c     vm=0.5D0*(vml+vmr)
      do i=ilo,ihi,1
         do m=1,8,1
            vm(i,m)=vml(i,m)
         enddo
      enddo
      return
      end
