C -*- Mode: Fortran; -*-
c-----------------------------------------------------------------------
c     Ravi Samtaney
c     Copyright 2002
c     Princeton Plasma Physics Laboratory
c     All Rights Reserved
c-----------------------------------------------------------------------
c     $Log: viscousflux.F,v $
c     Revision 1.4  2004/11/23 00:54:17  noel
c     Trying to keep in-line with repo
c
c     Revision 1.1  2002/06/13 22:16:18  samtaney
c     Original unsplit 8-wave Ideal MHD source (includes B filter).
c
c-----------------------------------------------------------------------
#define TWO_D
      subroutine ViscousFlux(fvis,vx,direction,dt,indx,indy,indz)

#include "EosCommon.fh"
c     
c     This subroutine determines the vector of inviscid fluxes.
      integer indx,indy,indz

#ifndef TWO_D
      real*8 fvis(IXLO+NGHOST:IXHI-NGHOST+indx,
     &                        IYLO+NGHOST:IYHI-NGHOST+indy, 
     &                        IZLO+NGHOST:IZHI-NGHOST+indz,nvar)
#else
      real*8 fvis(IXLO+NGHOST:IXHI-NGHOST+indx,
     &                        IYLO+NGHOST:IYHI-NGHOST+indy, 
     &                        1,nvar)
#endif

      real*8 vx(IXLO:IXHI,IYLO:IYHI,IZLO:IZHI,nvar)
      real*8 dt
      integer direction,ilo,ihi
      integer i,j,k
c     
      real*8 vd(INLO:INHI,nvar)
      real*8 vdjp(INLO:INHI,nvar)
      real*8 vdjm(INLO:INHI,nvar)
      real*8 vdkp(INLO:INHI,nvar)
      real*8 vdkm(INLO:INHI,nvar)
      real*8 etaPrime(INLO:INHI)
      real*8 fx(INLO:INHI,nvar)
      real*8 dxi,dyi,dzi
c     This subroutine determines the vector of viscous fluxes.
c     v: Vector of variables - rho, u,v,w,Bx,By,Bz,p
c     rho=density, u=x-velocity, v=y-velocity
c     w=z-velocity
c     Bx,By,Bz=x,y,z components of the magnetic field
c     p=thermal pressure
c
      dxi=1.D0/dx
      dyi=1.D0/dy
      dzi=1.D0/dz
      fx=0.D0
c     
      select case  (direction)
      case(1)
         do k=1,nzlocal,1
            do j=1,nylocal,1
               ilo=ixlo; ihi=ixhi;
               vd(ilo:ihi,:)=vx(ilo:ihi,j,k,:)
               vdjp(ilo:ihi,:)=vx(ilo:ihi,j+1,k,:)
               vdjm(ilo:ihi,:)=vx(ilo:ihi,j-1,k,:)
c               vdkp(ilo:ihi,:)=vx(ilo:ihi,j,k+1,:)
c               vdkm(ilo:ihi,:)=vx(ilo:ihi,j,k-1,:)
               vdkp(ilo:ihi,:)=vx(ilo:ihi,j,k,:)
               vdkm(ilo:ihi,:)=vx(ilo:ihi,j,k,:)
               ilo=1; ihi=nxlocal+1;
               call ComputeResistivity(etaPrime,vd(inlo,7),ilo,ihi)
               call ViscousFlux1D(vd,vdjp,vdjm,vdkp,
     &	       vdkm,etaPrime,fx,ilo,ihi,dxi,dyi,dzi)
               call PhiRHSV(fx,vd,etaPrime,ilo,ihi,dxi,dyi,dzi)
               fvis(ilo:ihi,j,k,:)=fx(ilo:ihi,:)
c               do i=1,nxlocal+1,1
c                  write(51,*) i,fx(i,:)
c                  write(71,*) i,etaPrime(i)
c                  write(51,*)
c               enddo
c               do i=1,nxlocal,1
c                  write(61,*) i,fx(i+1,:)-fx(i,:)
c                  write(81,*) i,vd(i,7)
c                  write(51,*)
c               enddo
            enddo
         enddo
c         call flush(51)
c         call flush(61)
c         call flush(71)
c         call flush(81)
      case(2)
         do k=1,nzlocal,1
            do i=1,nxlocal,1
               ilo=iylo; ihi=iyhi;
               vd(ilo:ihi,:)=vx(i,ilo:ihi,k,:)
c               vdjp(ilo:ihi,:)=vx(i,ilo:ihi,k+1,:)
c               vdjm(ilo:ihi,:)=vx(i,ilo:ihi,k-1,:)
               vdjp(ilo:ihi,:)=vx(i,ilo:ihi,k,:)
               vdjm(ilo:ihi,:)=vx(i,ilo:ihi,k,:)
               vdkp(ilo:ihi,:)=vx(i+1,ilo:ihi,k,:)
               vdkm(ilo:ihi,:)=vx(i-1,ilo:ihi,k,:)
               call RemapToLocalCoords(vd,ilo,ihi,2)
               call RemapToLocalCoords(vdjp,ilo,ihi,2)
               call RemapToLocalCoords(vdkp,ilo,ihi,2)
               call RemapToLocalCoords(vdjm,ilo,ihi,2)
               call RemapToLocalCoords(vdkm,ilo,ihi,2)
               ilo=1; ihi=nylocal+1;
               call ComputeResistivity(etaPrime,vd(inlo,6),ilo,ihi)
               call ViscousFlux1D(vd,vdjp,vdjm,vdkp,
     &	       vdkm,etaPrime,fx,ilo,ihi,dyi,dzi,dxi)
               call PhiRHSV(fx,vd,etaPrime,ilo,ihi,dyi,dzi,dxi)
               fvis(i,ilo:ihi,k,:)=fx(ilo:ihi,:)
            enddo
         enddo
      case(3)
         do j=1,nylocal,1
            do i=1,nxlocal,1
               ilo=izlo; ihi=izhi;
               vd(ilo:ihi,:)=vx(i,j,ilo:ihi,:)
               vdjp(ilo:ihi,:)=vx(i+1,j,ilo:ihi,:)
               vdjm(ilo:ihi,:)=vx(i-1,j,ilo:ihi,:)
               vdkp(ilo:ihi,:)=vx(i,j+1,ilo:ihi,:)
               vdkm(ilo:ihi,:)=vx(i,j-1,ilo:ihi,:)
               call RemapToLocalCoords(vd,ilo,ihi,3)
               call RemapToLocalCoords(vdjp,ilo,ihi,3)
               call RemapToLocalCoords(vdkp,ilo,ihi,3)
               call RemapToLocalCoords(vdjm,ilo,ihi,3)
               call RemapToLocalCoords(vdkm,ilo,ihi,3)
               ilo=1; ihi=nzlocal+1;
               call ComputeResistivity(etaPrime,vd(inlo,5),ilo,ihi)
               call ViscousFlux1D(vd,vdjp,vdjm,vdkp,
     &	       vdkm,etaPrime,fx,ilo,ihi,dzi,dxi,dyi)
               call PhiRHSV(fx,vd,etaPrime,ilo,ihi,dzi,dxi,dyi)
               fvis(i,j,ilo:ihi,:)=fx(ilo:ihi,:)
            enddo
         enddo
c     
      end select
c     
      return
      end
c
c-----------------------------------------------------------------------
      Subroutine ViscousFlux1D(vd,vdjp,vdjm,vdkp,
     &	       vdkm,etaPrime,fvis,ilo,ihi,adx,ady,adz)

#include "EosCommon.fh"
      real*8 fvis(inlo:inhi,nvar)
      real*8 vd(inlo:inhi,nvar)
      real*8 vdjp(inlo:inhi,nvar)
      real*8 vdjm(inlo:inhi,nvar)
      real*8 vdkp(inlo:inhi,nvar)
      real*8 vdkm(inlo:inhi,nvar)
      real*8 etaPrime(inlo:inhi)
      real*8 tauxx,tauxy,tauxz
      real*8 dbxdx,dbxdy,dbxdz
      real*8 dbydx, dbzdx,dbmagdx,bi,bj,bk
      real*8 dudx,dudy,dudz,div
      real*8 dvdx,dwdx,dvdy,dwdz
      real*8 dtdx
      real*8 adx,ady,adz
      integer ilo,ihi
      integer i,j,k,l
c	 This subroutine determines the vector of inviscid fluxes.
c
c	 v: Vector of variables - rho, u,v,w,Bx,By,Bz,p
c	    rho=density, u=x-velocity, v=y-velocity
c           w=z-velocity
c           Bx,By,Bz= Magnetic Field
c	    p=pressure
c
         do i=ilo,ihi,1
c
            dudx=(vd(i,2)-vd(i-1,2))*adx
            dvdx=(vd(i,3)-vd(i-1,3))*adx
            dwdx=(vd(i,4)-vd(i-1,4))*adx
c
            dudy=0.25D0*(vdjp(i,2)+vdjp(i-1,2)-vdjm(i,2)-vdjm(i-1,2))
     &           *ady
            dvdy=0.25D0*(vdjp(i,3)+vdjp(i-1,3)-vdjm(i,3)-vdjm(i-1,3))
     &           *ady
c            dwdy=0.25D0*(vdjp(i,4)+vdjp(i-1,4)-vdjm(i,4)-vdjm(i-1,4))
c
            dudz=0.25D0*(vdkp(i,2)+vdkp(i-1,2)-vdkm(i,2)-vdkm(i-1,2))
     &           *adz
c            dvdz=0.25D0*(vdkp(i,3)+vdkp(i-1,3)-vdkm(i,3)-vdkm(i-1,3))
            dwdz=0.25D0*(vdkp(i,4)+vdkp(i-1,4)-vdkm(i,4)-vdkm(i-1,4))
     &           *adz
c
            div=(dudx+dvdy+dwdz)/3.D0
c
            tauxx=2.d0*mu*(dudx-div)
            tauxy=mu*(dvdx+dudy)
            tauxz=mu*(dwdx+dudz)
c
            dbxdx=(vd(i,5)-vd(i-1,5))*adx
            dbydx=(vd(i,6)-vd(i-1,6))*adx
            dbzdx=(vd(i,7)-vd(i-1,7))*adx
c
            dbxdy=0.25D0*(vdjp(i,5)+vdjp(i-1,5)-vdjm(i,5)-vdjm(i-1,5))
     &           *ady
            dbxdz=0.25D0*(vdkp(i,5)+vdkp(i-1,5)-vdkm(i,5)-vdkm(i-1,5))
     &           *adz
            dbmagdx=(sum(vd(i,5:7)*vd(i,5:7))-
     &           (sum(vd(i-1,5:7)*vd(i-1,5:7))))*0.5D0*adx
c
            bi=0.5D0*(vd(i,5)+vd(i-1,5))
            bj=0.5D0*(vd(i,6)+vd(i-1,6))
            bk=0.5D0*(vd(i,7)+vd(i-1,7))
c
            dtdx=(vd(i,8)/vd(i,1)-vd(i-1,8)/vd(i-1,1))*adx/rgas
c
            fvis(i,2)=tauxx
            fvis(i,3)=tauxy
            fvis(i,4)=tauxz
            fvis(i,6)=etaPrime(i)*(dbydx-dbxdy)
            fvis(i,7)=etaPrime(i)*(dbzdx-dbxdz)
c            fvis(i,6)=etaPrime(i)*(dbydx)
c            fvis(i,7)=etaPrime(i)*(dbzdx)
            fvis(i,8)=etaPrime(i)*(dbmagdx-bi*dbxdx-bj*dbxdy-bk*dbxdz)+
     &           0.5D0*((vd(i,2)+vd(i-1,2))*tauxx+
     &                  (vd(i,3)+vd(i-1,3))*tauxy+
     &                  (vd(i,4)+vd(i-1,4))*tauxz)+
     &           kappa*dtdx
         enddo
c
c
	return
        end

c-----------------------------------------------------------------------
      Subroutine computeResistivity(etaPrime,vd,ilo,ihi)

#include "EosCommon.fh"
      real*8 etaPrime(inlo:inhi)
      real*8 vd(inlo:inhi)
      integer ilo,ihi
      integer i,j,k,l
c	 This subroutine determines the mag field dependent
c     resistivity
c
         do i=ilo,ihi,1
c
            if(vd(i)+vd(i-1).le.-2.D-04) then
c               etaPrime(i)=eta+100*eta*(1.D0-exp(-vd(i)**2/0.075))
c               etaPrime(i)=eta+100*eta*(1.D0-exp(-(vd(i)/0.075/7.8)**2))
               etaPrime(i)=eta+100.D0*eta*(1.D0-
     &           dexp(-((vd(i)+vd(i-1))/2.D0/0.075D0/7.8D0)**2))
            else
               etaPrime(i)=eta
            endif
         enddo
         return
         end
            
