C       _______              __
C      / ___/ /  ___  __ _  / /  ___
C     / /__/ _ \/ _ \/  ' \/ _ \/ _ \
C     \___/_//_/\___/_/_/_/_.__/\___/ 
C    
C     This software is copyright (C) by the Lawrence Berkeley
C     National Laboratory.  Permission is granted to reproduce
C     this software for non-commercial purposes provided that
C     this notice is left intact.
C     
C     It is acknowledged that the U.S. Government has rights to
C     this software under Contract DE-AC03-765F00098 between
C     the U.S.  Department of Energy and the University of
C     California.
C    
C     This software is provided as a professional and academic
C     contribution for joint exchange. Thus it is experimental,
C     is provided ``as is'', with no warranties of any kind
C     whatsoever, no support, no promise of updates, or printed
C     documentation. By using this software, you acknowledge
C     that the Lawrence Berkeley National Laboratory and
C     Regents of the University of California shall have no
C     liability with respect to the infringement of other
C     copyrights by any part of this software.
C

c     rhs(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
c     n:  length of whole domain
c     amid1, amid2, amid3:  space coordinates of center points
      subroutine initializerhs(rhs, l1, l2, l3, u1, u2, u3, n, s,
     &     h, nballs, amid1, amid2, amid3, rsph, vrbos)
      implicit none
c     --- arguments ---
      integer l1, l2, l3, u1, u2, u3, n, s, nballs
      real*8 rhs(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
      real*8 h, amid1(nballs), amid2(nballs), amid3(nballs)
      real*8 rsph(nballs)
      integer vrbos
c     --- locals ---
      integer dpoly
      parameter (dpoly = 8)
      real*8 polyco(0:dpoly)
      integer i, j, k, iball

c      h = 1d0 / dfloat(n)

c      rsph = 0.45d0

      if (vrbos .gt. 0) then
         print *, 'getting coefficients of poly'
      endif

c     first set to zero
      do k = l3-s, u3+s
         do j = l2-s, u2+s
            do i = l1-s, u1+s
               rhs(i, j, k) = 0d0
            enddo
         enddo
      enddo

      do iball = 1, nballs
         call setballpolyrcoeffs(dpoly, rsph(iball), polyco)

         call addrhsballpolyr(rhs, l1, l2, l3, u1, u2, u3, s,
     &        amid1(iball), amid2(iball), amid3(iball),
     &        h, rsph(iball), dpoly, polyco)
      enddo

      return
      end


c     phi(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
c     n:  length of whole domain
c     amid1, amid2, amid3:  space coordinates of center point
      subroutine getexactphi(phi, l1, l2, l3, u1, u2, u3, n, s,
     &     h, nballs, amid1, amid2, amid3, rsph, vrbos)
      implicit none
c     --- arguments ---
      integer l1, l2, l3, u1, u2, u3, n, s, nballs, iball
      real*8 phi(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
      real*8 h, amid1(nballs), amid2(nballs), amid3(nballs)
      real*8 rsph(nballs)
      integer vrbos
c     --- locals ---
      integer dpoly, i, j, k
      parameter (dpoly = 8)
      real*8 polyco(0:dpoly)

c      h = 1d0 / dfloat(n)

c      rsph = 0.45d0

      if (vrbos .gt. 0) then
         print *, 'getting coefficients of poly'
      endif

      if (vrbos .gt. 0) then
         print *, 'setting exact phi'
      endif

c     first set to zero
      do k = l3-s, u3+s
         do j = l2-s, u2+s
            do i = l1-s, u1+s
               phi(i, j, k) = 0d0
            enddo
         enddo
      enddo

      do iball = 1, nballs
         call setballpolyrcoeffs(dpoly, rsph(iball), polyco)

         call addexactphiballpolyr(phi, l1, l2, l3, u1, u2, u3, s,
     &        amid1(iball), amid2(iball), amid3(iball),
     &        h, rsph(iball), dpoly, polyco)
      enddo

      return
      end


      subroutine setballpolyrcoeffs(deg, rsph, coeffs)
c     Set the coefficients in the polynomial for right-hand side:
c     rhs(r) = sum_{d=0:deg} coeffs(r) * r^d  if r < rsph
c              0                              if r >= rsph
      implicit none
c     --- arguments ---
      integer deg
      real*8 rsph, coeffs(0:deg)
c     --- locals ---
      real*8 rspinv

c     this is for ((r/R - (r/R)^2))^4
c     = (r/R)^4 - 4*(r/R)^3*(r/R)^2 + 6*(r/R)^2*(r/R)^2^2 -
c       4*(r/R)*(r/R)^2^3 + (r/R)^2^4
c     = (1/R^4)*r^4 - (4/R^5)*r^5 + (6/R^6)*r^6 - (4/R^7)*r^7 + (1/R^8)*r^8
      rspinv = 1d0 / rsph
      coeffs(0) = 0d0
      coeffs(1) = 0d0
      coeffs(2) = 0d0
      coeffs(3) = 0d0
      coeffs(4) = (rspinv**4)
      coeffs(5) = -4d0 * (rspinv**5)
      coeffs(6) = 6d0 * (rspinv**6)
      coeffs(7) = -4d0 * (rspinv**7)
      coeffs(8) = (rspinv**8)

      return
      end


      subroutine addrhsballpolyr(rhs, l1, l2, l3, u1, u2, u3, s,
     &     amid1, amid2, amid3, h, rsph, deg, coeffs)
c     Returns the right-hand side
c     rhs(r) = sum_{d=0:deg} coeffs(r) * r^d  if r < rsph
c              0                              if r >= rsph
c     where r = sqrt(x*x + y*y + z*z)
c     and x, y, z are distances in space from (n/2, n/2, n/2),
c     with mesh spacing h.
      implicit none
c     --- arguments ---
      integer l1, l2, l3, u1, u2, u3, s, deg
      real*8 rhs(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
      real*8 amid1, amid2, amid3
      real*8 coeffs(0:deg)
      real*8 h, rsph
c     --- locals ---
      integer i, j, k
      real*8 x, y, z, r
      real*8 dfloat, dsqrt, evalpoly

c      print *, 'adding to rhs, center ', amid1, amid2, amid3,
c     &     ' radius ', rsph
      do k = l3, u3
         z = (dfloat(k) + 0.5) * h - amid3
         do j = l2, u2
            y = (dfloat(j) + 0.5) * h - amid2
            do i = l1, u1
               x = (dfloat(i) + 0.5) * h - amid1
               r = dsqrt(x*x + y*y + z*z)
               if (r .lt. rsph) then
                  rhs(i, j, k) = rhs(i, j, k) +
     &                 evalpoly(coeffs, deg, r)
               endif
            enddo
         enddo
      enddo

      return
      end


      real*8 function evalpoly(coeffs, deg, x)
c     Returns the evaluation of the polynomial
c     sum_{d=0:deg} coeffs(d)*(x**d)
c     coeffs:  input array, coefficients of polynomial
c     deg:  input degree of polynomial
c     x:  input, point of evaluation of polynomial
      implicit none
c     --- arguments ---
      integer deg
      real*8 coeffs(0:deg), x
c     --- locals ---
      integer d
      real*8 sum

      sum = coeffs(deg)
      do d = deg-1, 0, -1
         sum = x * sum + coeffs(d)
      enddo
      evalpoly = sum
      
      return
      end


      subroutine addexactphiballpolyr(phi, l1, l2, l3, u1, u2, u3, s,
     &     amid1, amid2, amid3, h, rsph, deg, coeffs)
c     Returns the exact value of phi that solves
c     Laplacian(phi) = rhs
c     for rhs as set in getrhsballpolyr.
c
c     If the right-hand side is
c     rhs(r) = r^d  if r < rsph
c              0    if r >= rsph
c     then the solution to the Poisson equation is:
c     phi(r) = r^(d+2)/((d+2)*(d+3)) - rsph^(d+2)/(d+2)  if r < rsph
c              -rsph^(d+3)/((d+3)*r)                     if r >= rsph
c     where r = sqrt(x*x + y*y + z*z)
c     and x, y, z are distances in space from (n/2, n/2, n/2),
c     with mesh spacing h.
      implicit none
c     --- arguments ---
      integer l1, l2, l3, u1, u2, u3, s, deg
      real*8 phi(l1-s:u1+s, l2-s:u2+s, l3-s:u3+s)
      real*8 amid1, amid2, amid3, h, rsph
      real*8 coeffs(0:deg)
c     --- locals ---
      integer i, j, k, d
      real*8 x, y, z, r
      real*8 dfloat, dsqrt, evalpoly
      real*8 cincon(0:deg), cinr(0:deg), couter(0:deg)
      real*8 conout, conin

c      print *, 'adding to phiexact, center ', amid1, amid2, amid3,
c     &     ' radius ', rsph
      do d = 0, deg
c     coefficient of r^(d+2) for r < rsph
         cinr(d) = coeffs(d) / dfloat((d+2)*(d+3))
c     coefficient of rsph^d for r < rsph
         cincon(d) = - coeffs(d) * (rsph**2) / dfloat(d+2)
c     coefficient of rsph^d/r for r >= rsph
         couter(d) = - coeffs(d) * (rsph**3) / dfloat(d+3)
      enddo
c     coefficient of 1/r for r >= rsph
      conout = evalpoly(couter, deg, rsph)
c     constant term (a polynomial in rsph) for r < rsph
      conin = evalpoly(cincon, deg, rsph)
      do k = l3-s, u3+s
         z = (dfloat(k) + 0.5) * h - amid3
         do j = l2-s, u2+s
            y = (dfloat(j) + 0.5) * h - amid2
            do i = l1-s, u1+s
               x = (dfloat(i) + 0.5) * h - amid1
               r = dsqrt(x*x + y*y + z*z)
               if (r .gt. rsph) then
                  phi(i, j, k) = phi(i, j, k) +
     &                 conout / r
               else
                  phi(i, j, k) = phi(i, j, k) +
     &                 (r**2) * evalpoly(cinr, deg, r) + conin
               endif
            enddo
         enddo
      enddo

      return
      end
