/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


#include "Box.H"
#include "IntVectSet.H"
#include "GodunovPhysics.H"
#include "Vector.H"

//////
bool
GodunovPhysics::isDefined() const
{
  return m_isDefined;
}
///////
GodunovPhysics::GodunovPhysics()
{
  m_isDefined = false;
  m_isBCSet = false;
  m_bc = NULL;
}
//////
PhysIBC*
GodunovPhysics::getPhysIBC() const
{
  assert(m_isBCSet);
  return m_bc;
}
///////////
void
GodunovPhysics::setPhysIBC(PhysIBC* a_bc)
{
  // Delete old boundary condition object - if any
  if (m_bc != NULL)
  {
    delete m_bc;
  }

  // Store new boundary condition object
  m_bc = a_bc->new_physIBC();

  // just in case we're re-defining the BC
  if (m_isDefined)
    m_bc->define(m_domain, m_dx);


  m_isBCSet = true;
}
///////////
GodunovPhysics::~GodunovPhysics()
{
  if(m_bc != NULL)
    delete m_bc;
}
///////////////
void
GodunovPhysics::define(const ProblemDomain& a_domain,
                       const Real& a_dx)
{
  m_domain = a_domain;
  m_dx     = a_dx;
  m_isDefined = true;
  if(m_bc != NULL)
    m_bc->define(m_domain, m_dx);
}
//////////////

