/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DFMartin, Sun, May 5, 2002

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"

#include "BiCGStabSmoother.H"

#include <cstdio>  // for debugging sprintfs

#include "Timer.H"
#include "NoelsPrettyPrint.H"
extern Timer Everything;
Timer TBOT_doBottomSmooth("indoBottomSmooth", 107);

Timer TBOT_init              ("init",           TBOT_doBottomSmooth);
Timer TBOT_compInitResidual  ("comp init res",  TBOT_doBottomSmooth);
Timer TBOT_stuff             ("stuff",          TBOT_doBottomSmooth);
Timer TBOT_DotProduct1       ("DotProduct1",    TBOT_doBottomSmooth);
Timer TBOT_badDotProd_plus   ("badDotProd plus",TBOT_doBottomSmooth);
Timer TBOT_stuff2            ("stuff2 b4 pc",   TBOT_doBottomSmooth);
Timer TBOT_preconditioner1   ("precond1",       TBOT_doBottomSmooth);
Timer TBOT_app1              ("applyOp1",       TBOT_doBottomSmooth);
Timer TBOT_DotProduct2       ("DotProduct2",    TBOT_doBottomSmooth);
Timer TBOT_incrementSolution ("inc solution",   TBOT_doBottomSmooth);
Timer TBOT_computeRes_norm   ("comp res norm",  TBOT_doBottomSmooth);
Timer TBOT_preconditioner2   ("precond2",       TBOT_doBottomSmooth);
Timer TBOT_app2              ("applyOp2",       TBOT_doBottomSmooth);
Timer TBOT_omega             ("comp omega",     TBOT_doBottomSmooth);
Timer TBOT_increment_residual("inc residual",   TBOT_doBottomSmooth);
Timer TBOT_check_convergence ("check converg",  TBOT_doBottomSmooth);
Timer TBOT_doublecheck_convergence("double check",TBOT_doBottomSmooth);

extern double global_BS_tolerance;

BiCGStabSmoother::BiCGStabSmoother()
{
  //m_maxIter = 40;
  //m_solverTol = 1.0e-7;

  // try this:
  m_maxIter = 103;
  m_solverTol = global_BS_tolerance;

#ifdef CH_USE_FLOAT
  m_small = 1.0e-30;
#endif
#ifdef CH_USE_DOUBLE
  m_small = 1.0e-60;
#endif
  // m_convergeSmall = 1.0e-60;
  m_convergeSmall = 1.0e-2;

  m_enableRestart = true;

  // m_verbose = true;
  m_verbose = false;
}

BiCGStabSmoother::~BiCGStabSmoother()
{
}

void BiCGStabSmoother::setMaxIter(int a_maxIter)
{
  m_maxIter = a_maxIter;
}

void BiCGStabSmoother::setConvergenceMetric(Real a_metric, int a_comp)
{
  // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size())
  {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++)
      {
        m_convergenceMetrics[i] = tempVect[i];
      }
    for (int i=tempVect.size(); i<a_comp; i++)
      {
        m_convergenceMetrics[i] = 1.0;
      }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
}

void BiCGStabSmoother::setSolverTol(Real a_solverTol)
{
  m_solverTol = a_solverTol;
}

void BiCGStabSmoother::setEnableRestart(bool a_enableRestart)
{
  m_enableRestart = a_enableRestart;
}

void BiCGStabSmoother::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}

BaseBottomSmoother* BiCGStabSmoother::new_bottomSmoother() const
{
  BiCGStabSmoother* newsmoother = new BiCGStabSmoother();
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL)
    {
      MayDay::Error("Out of Memory in BiCGStabSmoother::new_bottomSmoother");
    }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

/***********************/
// This does a GSRB Pre/Conditioned BiCGStab on a level
// for the bottom solver.
/***********************/
void BiCGStabSmoother::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                                      const LevelData<FArrayBox>& a_rhs,
                                      LevelOp*                    a_levelopPtr)
{
  //pout() << " BiCGStabSmoother::doBottomSmooth" << endl;
  TBOT_doBottomSmooth.start();

  Timer BS_timer;
  BS_timer.start();

  TBOT_init.start();
  bool alreadyRestarted = false;

  // if we won't allow resarts, then act like we already have...
  if (!m_enableRestart)
    {
      alreadyRestarted = true;
    }

  Real small2 = 1.0e-8;
  int ncomp = a_rhs.nComp();
  char tempchar[200];

  assert(a_rhs.nComp() == a_phi.nComp());

  // in the case where the convergence metrics haven't already been
  // defined, default to norm(rhs) for the bottom smooth
  if (m_convergenceMetrics.size() != ncomp)
    {
      m_convergenceMetrics = Vector<Real>(ncomp, 0);
      for (int comp=0; comp<ncomp; comp++)
        {
          Interval thisInterval(comp,comp);
          int normType = 0;
          m_convergenceMetrics[comp] = norm(a_rhs, thisInterval, normType);
        } // end loop over components
    } // end if we need to resize convergence metrics

  const DisjointBoxLayout grids = a_rhs.getBoxes();

  //pout() << " BiCGStabSmoother::doBottomSmooth: grids.size() " << grids.size() << endl;

  assert (grids == a_phi.getBoxes());

  LevelData<FArrayBox> corremf(grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> residmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> rtwidmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> shatmf (grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> phatmf (grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> pmf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> vmf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> tmf    (grids, ncomp, IntVect::TheZeroVector());

  DataIterator dit = residmf.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      // this is necessary to prevent uninitialized memory reads
      residmf[dit()].setVal(0.0);
      rtwidmf[dit()].setVal(0.0);
      shatmf[dit()] .setVal(0.0);
      phatmf[dit()] .setVal(0.0);
      pmf[dit()]    .setVal(0.0);
      tmf[dit()]    .setVal(0.0);
      corremf[dit()].setVal(0.0);
      vmf[dit()]    .setVal(0.0);
    }
  TBOT_init.stop();

  // compute initial residual
  TBOT_compInitResidual.start();
  a_levelopPtr->applyOpH(a_phi,residmf);

  for (dit.begin(); dit.ok(); ++dit)
    {
      residmf[dit()] -= a_rhs[dit()];
      residmf[dit()].negate();
    }
  TBOT_compInitResidual.stop();

  Vector<Real> initnorm(ncomp,0.0); // chombo no have this
  Vector<Real> oldNorm (ncomp,0.0);
  Vector<Real> rnorm   (ncomp,0.0);
  Vector<Real> norms   (ncomp,0.0);

  TBOT_stuff.start();
  // compute residual and rhs norms
  // finish is norm(residual)/nrom(rhs) < m_solverTol
  // for all components
  for (int comp = 0; comp < ncomp; comp++)
    {
      Interval compInterval(comp,comp);

      rnorm[comp] = norm(residmf, compInterval, 0);

      oldNorm[comp] = rnorm[comp];
      initnorm[comp] = rnorm[comp]; // chombo no have this

//       if (m_verbose)
//         {
//           pout() << "Bottom Smoother norm of initial residual [" << comp
//                  << "] = " << rnorm[comp] << endl;
//         }
//       sprintf(tempchar, "  BiCGStab BottomSmooth    initNorm=%12.4e  tol=%10.4e cmet=%10.4e small=%10.4e maxIter=%d\n",
//               rnorm[comp], m_solverTol, m_convergenceMetrics[comp],
//               m_small, m_maxIter);
//       pout() << tempchar;
    }

  for (dit.begin(); dit.ok(); ++dit)
    {
      rtwidmf[dit()].copy(residmf[dit()],0,0,ncomp);
    }

  Vector<Vector<Real> > rho(ncomp);
  Vector<Vector<Real> > alpha(ncomp);
  // Vector<Vector<Real> > beta(ncomp);
  Vector<Vector<Real> > omega(ncomp);

  // dummy omega[0], alpha[0] (not used)
  for (int comp=0; comp<ncomp; comp++)
    {
      omega[comp].push_back(0.0);
      alpha[comp].push_back(0.0);
      // beta[comp].push_back(0.0);
    }
  TBOT_stuff.stop();

  // not sure if this is the right thing to do.
  // option 1: if any rhoDots are bad, add in current correction and exit
  // option 2: keep going with other components?
  bool initialResidualZero = true;

  // if initial residual is already 0, don't do anything here
  for (int comp=0; comp<ncomp; comp++)
    {
      if (rnorm[comp] > m_small) initialResidualZero = false;
    }

  // my crap:
  bool normTest = false;
  bool hangingTest = false;
  double IterationTimeSum = 0.0;
  Timer IterationTimer;
  Real previous_rnorm = rnorm[0];
  // end my crap.

  bool finished = initialResidualZero;
  bool badDotProd = false;

  int iter;
  for (iter = 1; (iter <= m_maxIter) && !finished; iter++)
    {
      normTest = false;
      hangingTest = false;

      IterationTimer.clear();
      IterationTimer.start();
      TBOT_DotProduct1.start();
      Vector<Real> rhodot(ncomp);
      badDotProd = false;

      for (int comp = 0; comp < ncomp; comp++)
        {
          Interval compInterval(comp,comp);

          rhodot[comp] = DotProduct(residmf, rtwidmf, grids, compInterval);

          if (Abs(rhodot[comp]) < m_small)
            {
              // cerr << "bicbstab failed in procedure ";
              // cerr << "PoissonOp::bottomSmoother" << endl;
              // cerr << "at iteration " << iter;
              badDotProd = true;
            }

          rho[comp].push_back(rhodot[comp]);
        } // end loop over components
      TBOT_DotProduct1.stop();

      if (badDotProd)
        {
          TBOT_badDotProd_plus.start();
          for (dit.begin(); dit.ok(); ++dit)
            {
              a_phi[dit()].plus(corremf[dit()],0,0,ncomp);
            }

          if (m_verbose)
            {
              pout() << "BiCGStabSmoother: BadDotProd -- returning!  iter = "
                     << iter << endl;
            }
          TBOT_badDotProd_plus.stop();
          //return;
      goto done;
        }

      // if we're restarting the computation, act as if it was
      // the first iteration
      if (iter == 1 || alreadyRestarted)
        {
          for (dit.begin(); dit.ok(); ++dit)
            {
              pmf[dit()].copy(residmf[dit()]);
            }
        }
      else
        {
          TBOT_stuff2.start();
          for (int comp = 0; comp < ncomp; comp++)
            {
              Real betaim1 = (rho[comp][iter-1] / rho[comp][iter-2])
                           * (alpha[comp][iter-1] / omega[comp][iter-1]);

              // beta[comp].push_back(betaim1);

              Real omegaim1 = omega[comp][iter-1];

              // p(i) = r(i-1) + beta(i-1)*(p(i-1) -omega(i-1)*v(i-1))

              for (dit.begin(); dit.ok(); ++dit)
                {
                  FArrayBox& pfab = pmf[dit()];
                  FArrayBox& vfab = vmf[dit()];
                  FArrayBox& rfab = residmf[dit()];

                  Box fabbox = grids.get(dit());
                  FArrayBox temp(fabbox, 1);

                  temp.copy(vfab,comp,0,1);

                  temp *= omegaim1;
                  temp.negate();
                  temp.plus(pfab,comp,0,1);
                  temp *= betaim1;
                  temp.plus(rfab,comp,0,1);

                  pfab.copy(temp,0,comp,1);
                } // end loop over grids
            } // end loop over components
          TBOT_stuff2.stop();
        } // end if i ==  1 's else

      // solve M(phat) = p(i)
      TBOT_preconditioner1.start();
      a_levelopPtr->levelPreconditioner(phatmf, pmf);
      TBOT_preconditioner1.stop();

      // v(i) = A(phat)
      TBOT_app1.start();
      a_levelopPtr->applyOpH(phatmf,vmf);
      TBOT_app1.stop();

      // alpha(i) = rho(i-1)/(rtwid^T v(i))
      TBOT_DotProduct2.start();
      bool badCorrection = false;
      for (int comp = 0; comp < ncomp; comp++)
        {
          Interval compInt(comp,comp);
          Real denom = DotProduct(vmf, rtwidmf, grids,compInt);
          Real alphai;

          if (Abs(denom) > small2*Abs(rho[comp][iter-1]))
            {
              alphai = rho[comp][iter-1]/denom;
            }
          else
            {
              alphai = 0.0;
              badCorrection = true;
            }

          alpha[comp].push_back(alphai);

          // s = r(i-1) - alpha(i)*v(i)
        } // end loop over components
      TBOT_DotProduct2.stop();

      if (!badCorrection)
        {
          // increment residual and solution
          TBOT_incrementSolution.start();
          for (dit.begin(); dit.ok(); ++dit)
            {
              FArrayBox& rfab    = residmf[dit()];
              FArrayBox& vfab    = vmf[dit()];
              FArrayBox& phatfab = phatmf[dit()];
              FArrayBox& corrfab = corremf[dit()];

              for (int comp = 0; comp < ncomp; comp++)
                {
                  // increment residual: resid = resid - alpha*v
                  rfab.plus(vfab,-1.0*alpha[comp][iter],comp,comp,1);

                  // increment correction: corr = corr + alpha*phat
                  corrfab.plus(phatfab,alpha[comp][iter],comp,comp,1);
                }
            } // end loop over grids
          TBOT_incrementSolution.stop();
        } // end if the first correction step is OK
      else
        {
          // if correction was bad, don't increment anything,
          // but set residual to zero (forces recomputation of resid)
          for (dit.begin(); dit.ok(); ++dit)
            {
              residmf[dit()].setVal(0.0);
            }
        }

      // loop over components and compute residuals
      TBOT_computeRes_norm.start();
      for (int comp = 0; comp < ncomp; comp++)
        {
          Interval compInt(comp,comp);

          norms[comp] = norm(residmf, compInt, 0);
        } // end loop over comps
      TBOT_computeRes_norm.stop();

      // check norm of s.
      // if small enough, step to end (forces recomputation of residual, etc)
      bool allDone = true;
      for (int comp = 0; comp < ncomp; comp++)
        {
          if (norms[comp] > m_solverTol*m_convergenceMetrics[comp])
            {
              allDone = false;
            }
        }

      // if not "done" here, do second correction step
      if (!allDone)
        {
          // solve Mshat = s
      TBOT_preconditioner2.start();
          a_levelopPtr->levelPreconditioner(shatmf, residmf);
      TBOT_preconditioner2.stop();
          // t = A(shat)
      TBOT_app2.start();
          a_levelopPtr->applyOpH(shatmf,tmf);
      TBOT_app2.stop();
          // w(i) = tTresid/tTt
      TBOT_omega.start();
          for (int comp=0; comp<ncomp; comp++)
            {
              Interval compInt(comp,comp);
              Real numerw = DotProduct(tmf, residmf, grids,compInt);
              Real denomw = DotProduct(tmf, tmf, grids,compInt);
              Real omegai = 0;

              if (Abs(denomw) > m_small)
                {
                  omegai = numerw/denomw;
                }

              omega[comp].push_back(omegai);
            } // end loop over comps
          TBOT_omega.stop();
          // corr(i) = corr(i-1) + omega(i)*shat
          // resid(i) = resid(i) - omega(i)*t
          TBOT_increment_residual.start();
          for (dit.begin(); dit.ok(); ++dit)
            {
              FArrayBox& corrfab = corremf[dit()];
              FArrayBox& shatfab = shatmf[dit()];
              FArrayBox& tfab    = tmf[dit()];

              for (int comp = 0; comp < ncomp; comp++)
                {
                  // increment correction: corr = corr+omega*shat
                  corrfab.plus(shatfab,omega[comp][iter],
                               comp,comp,1);

                  // increment residual: resid = resid - omega*t
                  residmf[dit()].plus(tfab,-1.0*omega[comp][iter],
                                      comp,comp,1);
                }
            } // end loop over grids
      TBOT_increment_residual.stop();

          // reset this just in case (if we made it this
          // far, then we can assume that we're not caught in a
          // loop
          // (dfm, 5/29/02) actually, don't reset this -- only
          // allow one restart.
          // alreadyRestarted = false;
        } // end second update loop
      else
        {
          // if we skipped second update step, still need to store something
          // in omega.  try using 0
          for (int comp = 0; comp < ncomp; comp++)
            {
              omega[comp].push_back(0.0);
              // beta[comp].push_back(0.0);
            }
        }

      // now check residual.
      TBOT_check_convergence.start();
      allDone = true;
      for (int comp = 0; comp < ncomp; comp++)
        {
          Interval compInt(comp,comp);

          rnorm[comp] = norm(residmf, compInt, 0);

          if (m_verbose && (rnorm[comp] > 0))
            {
              //pout() << "BottomSmoother: after " << iter
              //     << " iterations, residnorm[" << comp << "] = "
              //     << rnorm[comp] << endl;
            }

          if ((rnorm[comp] > m_solverTol*m_convergenceMetrics[comp]) &&
              (Abs(omega[comp][iter]) > m_small))
            {
              allDone = false;
            }

#if 0
          // commented out version with hanging test.  also, I think
          // the second && should be an ||? (DFM 4/8/04)

          // second test is to eliminate solver hanging
          allDone = allDone && ( ((rnorm[comp]
                                   < m_solverTol*m_convergenceMetrics[comp])
                                 && (Abs(omega[comp][iter]) < m_small))
                                 || (rnorm[comp]
                                     > oldNorm[comp]*(1.0-m_convergeSmall)));
#endif
          oldNorm[comp] = rnorm[comp];
        } // end loop over comps
      TBOT_check_convergence.stop();

      // if residnorm is small enough (or if correction is
      // small enough), then recompute residual from scratch
      // and check again (to avoid roundoff-drift issues).  if
      // we _still_ think we're done, then exit
      TBOT_doublecheck_convergence.start();
      if (allDone)
        {
          // recompute residual from scratch -- first increment
          // a_phi with correction -- also reset correction to zero
          for (dit.begin(); dit.ok(); ++dit)
            {
              a_phi[dit()].plus(corremf[dit()],0,0,ncomp);
              corremf[dit()].setVal(0.0);
            }

          a_levelopPtr->applyOpH(a_phi,residmf);

          for (dit.begin(); dit.ok(); ++dit)
            {
              residmf[dit()] -= a_rhs[dit()];
              residmf[dit()] *= -1.0;
            }

          // now recompute residual norms and check for convergence
          allDone = true;
          for (int comp=0; comp<ncomp; comp++)
            {
              Interval compInterval(comp,comp);
              rnorm[comp] = norm(residmf, compInterval, 0);

              if (rnorm[comp] > m_solverTol*m_convergenceMetrics[comp])
                {
                  allDone = false;
                }
            } // end loop over components for residnorm checking

          // if we've already restarted and haven't gotten
          // anywhere, then we've stalled and should probably
          // exit.  also, if newly-computed residual still indicates
          // convergence, then we're done.
          if (alreadyRestarted || allDone)
            {
              finished = true;
            }
          else
            {
              // otherwise, go back and try again with new residual
              if (m_verbose)
                {
                  pout() << "Bottom Smoother -- restarting iterations"
                         << endl;
                }

              // also reset \tilde{r}
              for (dit.begin(); dit.ok(); ++dit)
                {
                  rtwidmf[dit()].copy(residmf[dit()],0,0,ncomp);
                }

              finished = false;
              alreadyRestarted = true;
            }

          if (finished && m_verbose)
            {
              for (int comp=0; comp<ncomp; comp++)
                {
                  pout() << "Bottom Smoother final Norm[" << comp << "] = "
                         << rnorm[comp] << endl;
                }
            }
        } // end if we're recomputing the residual
      TBOT_doublecheck_convergence.stop();

      // would do this over components if u had more than one...
      //#define DIAG_PRINT
#ifdef DIAG_PRINT
      Real convMetric;
      convMetric = m_solverTol*m_convergenceMetrics[0];
      if ( rnorm[0] < convMetric) {
        normTest = true;
      }
      Real rnormprint=rnorm[0];
      if (rnorm[0] == 0.0) rnormprint=1.0e-15;

      // hanging test that is embedded in the crazy logic above
      if ( rnorm[0] > oldNorm[0]*(1.0-m_convergeSmall) ) {
        hangingTest = true;
      }

      IterationTimer.stop();
      IterationTimeSum += IterationTimer.wc_time();
      sprintf(tempchar, "  BiCGStab BS after %3d iters, residnorm=%11.4e (%8.3f) n%d h%d a%d w=%5.3f\n",
              iter, rnorm[0], previous_rnorm/rnormprint, normTest,
              hangingTest, allDone, IterationTimer.wc_time());
      pout() << tempchar;
      previous_rnorm = rnorm[0];
#endif
    } // end main loop over BiCGStab iterations

 done:
  bool converged = true;

  for (int comp = 0; comp < ncomp; comp++)
    {
      if (rnorm[comp] >= m_solverTol*m_convergenceMetrics[comp])
        {
          converged = false;
        }
    }

  if (badDotProd == true || initialResidualZero == true)
    {
      // if bad dot product or initial residual is zero, then don't bother
      // telling me it's not converged.
      // and set converged to be true for the sake of diagnostics below.
      converged = true;
    }
  else
    {
      if (!converged && m_verbose)
        {
          pout() << "BiCGStabSmoother not Converged!  iter = " << iter << endl;

          for (int comp = 0; comp < ncomp; comp++)
            {
              pout() << "comp = " << comp
                     << " resid/metric = "
                     << rnorm[comp]/m_convergenceMetrics[comp]
                     << endl;
            }
        }
    }

  BS_timer.stop();

#ifdef DIAG_PRINT
  // debug print statements
  int Npoints=0, Nboxes=0, Nghosts=0;
  NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);
  // Nboxes should now be same as grids.size()
  char stuff[180];
  //sprintf(stuff, "   end doBS L%d Npoints=%6d Nboxes=%4d iter=%2d conv=%1d zI=%1d badDP=%1d hang=%1d rnorm=%9.3e w=%5.3f wc=%6.3f\n",
  //      GlobalLevel, Npoints, Nboxes, iter, converged,
  //      initialResidualZero, badDotProd, hangingTest, rnorm[0],
  //      IterationTimeSum, BS_timer.wc_time());
  sprintf(stuff, "   end doBS L%d Npoints=%6d Nboxes=%4d iter=%2d conv=%1d zI=%1d badDP=%1d hang=%1d rnorm=%9.3e tol=%9.3e wc=%6.3f\n",
          GlobalLevel, Npoints, Nboxes, iter, converged,
          initialResidualZero, badDotProd, hangingTest, rnorm[0],
          m_solverTol, BS_timer.wc_time());
  pout() << stuff;

  // bool myConverge = true;
  //   for (int comp=0; comp<ncomp; comp++) {
  //     if(rnorm[comp] > m_solverTol*initnorm[comp]) {
  //       myConverge = false;
  //     }
  //     if (!myConverge) pout() << "Bottom Smoother NOT converged!!!"
  //                            << " comp" << comp << " iter:" << iter
  //                            << " Npoints=" << Npoints
  //                            << " Nboxes=" << Nboxes
  //                            << " " << rnorm[comp] << " " << m_solverTol*initnorm[comp]
  //                            << endl;
  //   }

#endif

  TBOT_doBottomSmooth.stop();

  NoelsSooperIterationCounter[BIterationFlavor].Count++;
  NoelsSooperIterationCounter[BIterationFlavor].Value+=iter;
  if (!converged) {
    NoelsSooperIterationCounter[BIterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[BIterationFlavor].wc += BS_timer.wc_time();
#ifdef COUNTER_DEBUG
  pout() << " BiCGStabSmoother::doBottomSmooth BIterationFlavor=" << BIterationFlavor
         << " n=" << NoelsSooperIterationCounter[BIterationFlavor].Count
         << " wc=" << BS_timer.wc_time()
         << endl;
#endif
  BIterationFlavor = 0;

}
