/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DotProduct.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"

#include "BiCGStabSmootherSimple.H"

#include <cstdio>  // for debugging sprintfs

extern double global_BS_tolerance;

#include "Timer.H"
#include "NoelsPrettyPrint.H"
extern Timer Everything;
Timer TBOTS_doBottomSmooth("in doBS simple", 108);

Timer TBOTS_init              ("init",           TBOTS_doBottomSmooth);
Timer TBOTS_compInitResidual  ("comp init res",  TBOTS_doBottomSmooth);
Timer TBOTS_compNorm          ("comp norm",      TBOTS_doBottomSmooth);
Timer TBOTS_DotProduct1       ("DotProduct1",    TBOTS_doBottomSmooth);
Timer TBOTS_badDotProd_plus   ("badDotProd plus",TBOTS_doBottomSmooth);
Timer TBOTS_iter1stuff        ("iter1 stuff",    TBOTS_doBottomSmooth);
Timer TBOTS_stuff             ("stuff",          TBOTS_doBottomSmooth);
Timer TBOTS_preconditioner1   ("precond1",       TBOTS_doBottomSmooth);
Timer TBOTS_applyOp1          ("applyOp1",       TBOTS_doBottomSmooth);
Timer TBOTS_DotProduct2       ("DotProduct2",    TBOTS_doBottomSmooth);
Timer TBOTS_incrementSolution ("inc solution",   TBOTS_doBottomSmooth);
Timer TBOTS_computeRes_norm   ("comp res norm",  TBOTS_doBottomSmooth);
Timer TBOTS_preconditioner2   ("precond2",       TBOTS_doBottomSmooth);
Timer TBOTS_applyOp2          ("applyOp2",       TBOTS_doBottomSmooth);
Timer TBOTS_DotProduct3       ("DotProduct3",    TBOTS_doBottomSmooth);
Timer TBOTS_increment_residual("inc residual",   TBOTS_doBottomSmooth);
Timer TBOTS_check_convergence ("check converg",  TBOTS_doBottomSmooth);
Timer TBOTS_doublecheck_convergence("double check",TBOTS_doBottomSmooth);

BiCGStabSmootherSimple::BiCGStabSmootherSimple()
{
  m_maxIter = 103;
  m_solverTol = global_BS_tolerance;

#ifdef CH_USE_FLOAT
  m_small = 1.0e-30;
#endif
#ifdef CH_USE_DOUBLE
  m_small = 1.0e-60;
#endif
  m_convergeSmall = 1.0e-2;
  m_enableRestart = true;
  m_verbose = false;
}

BiCGStabSmootherSimple::~BiCGStabSmootherSimple(){}

void BiCGStabSmootherSimple::setMaxIter(int a_maxIter) {
  m_maxIter = a_maxIter;
}

void BiCGStabSmootherSimple::setConvergenceMetric(Real a_metric, int a_comp) {

  // may need to resize this vector
  if (a_comp >= m_convergenceMetrics.size()) {
    Vector<Real> tempVect = m_convergenceMetrics;
    m_convergenceMetrics.resize(a_comp+1);
    for (int i=0; i<tempVect.size(); i++) {
      m_convergenceMetrics[i] = tempVect[i];
    }
    for (int i=tempVect.size(); i<a_comp; i++) {
      m_convergenceMetrics[i] = 1.0;
    }
  } // end if we need to resize vector

  m_convergenceMetrics[a_comp] = a_metric;
}

void BiCGStabSmootherSimple::setSolverTol(Real a_solverTol) {
  m_solverTol = a_solverTol;
}

void BiCGStabSmootherSimple::setEnableRestart(bool a_enableRestart) {
  m_enableRestart = a_enableRestart;
}

void BiCGStabSmootherSimple::setVerbose(bool a_verbose) {
  m_verbose = a_verbose;
}

BaseBottomSmoother* BiCGStabSmootherSimple::new_bottomSmoother() const {
  BiCGStabSmootherSimple* newsmoother = new BiCGStabSmootherSimple();
  newsmoother->setVerbose(m_verbose);

  if (newsmoother == NULL) {
    MayDay::Error("Out of Memory in BiCGStabSmootherSimple::new_bottomSmoother");
  }

  return static_cast<BaseBottomSmoother*>(newsmoother);
}

void
BiCGStabSmootherSimple::doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                                       const LevelData<FArrayBox>& a_rhs,
                                       LevelOp*                    a_levelopPtr) {
  TBOTS_doBottomSmooth.start();
  Timer BS_timer;
  BS_timer.start();
  TBOTS_init.start();
  Real small2 = 1.0e-8;
  int ncomp = a_rhs.nComp(); // 1
  char tempchar[128];

  // in the case where the convergence metrics haven't already been
  // defined, default to norm(rhs) for the bottom smooth
  if (m_convergenceMetrics.size() != ncomp) {
    m_convergenceMetrics = Vector<Real>(ncomp, 0);
    Interval thisInterval(0,0);
    int normType = 0;
    m_convergenceMetrics[0] = norm(a_rhs, thisInterval, normType);
  } // end if we need to resize convergence metrics

  const DisjointBoxLayout grids = a_rhs.getBoxes();

  LevelData<FArrayBox> corremf(grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> shatmf (grids, ncomp, IntVect::TheUnitVector());
  LevelData<FArrayBox> phatmf (grids, ncomp, IntVect::TheUnitVector());

  LevelData<FArrayBox> residmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> rtwidmf(grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> pmf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> vmf    (grids, ncomp, IntVect::TheZeroVector());
  LevelData<FArrayBox> tmf    (grids, ncomp, IntVect::TheZeroVector());

  DataIterator dit = residmf.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)    {
    // this is necessary to prevent uninitialized memory reads
    corremf[dit()].setVal(0.0);
    shatmf[dit()] .setVal(0.0);
    phatmf[dit()] .setVal(0.0);

    residmf[dit()].setVal(0.0);
    rtwidmf[dit()].setVal(0.0);
    pmf[dit()]    .setVal(0.0);
    vmf[dit()]    .setVal(0.0);
    tmf[dit()]    .setVal(0.0);
  }
  TBOTS_init.stop();

  // compute initial residual   residual = b - Ax
  TBOTS_compInitResidual.start();
  a_levelopPtr->applyOpH(a_phi,residmf);   // residual = Ax

  for (dit.begin(); dit.ok(); ++dit) {
    residmf[dit()] -= a_rhs[dit()];     // residual -= b
    residmf[dit()].negate();            // residual = -residual
  }
  TBOTS_compInitResidual.stop();

  // compute residual and rhs norms
  TBOTS_compNorm.start();
  Real rho, rho_previous, rnorm, alpha, omega;
  Interval compInterval(0,0);
  rnorm = norm(residmf, compInterval, 0);

  // if initial residual is already 0, don't do anything here
  bool initialResidualZero = false;
  if (rnorm < m_small) {
    initialResidualZero = true;
  } else {
    // set r-twiddle to be the initial residual.
    for (dit.begin(); dit.ok(); ++dit) {
      rtwidmf[dit()].copy(residmf[dit()], 0, 0, ncomp);
    }
  }
  TBOTS_compNorm.stop();

  //sprintf(tempchar, "      BiCGStabSimple BS        initNorm=%12.4e  tol=%10.4e small=%10.4e  maxIter=%d\n",
  //      rnorm, m_solverTol, m_small, m_maxIter);
  //pout() << tempchar;
  Timer IterationTimer;
  Real previous_rnorm = rnorm;

  bool finished = initialResidualZero;
  bool badDotProd = false;
  int iter;
  for (iter = 1; (iter <= m_maxIter) && !finished; iter++) {

    IterationTimer.clear();
    IterationTimer.start();

    TBOTS_DotProduct1.start();
    rho = DotProduct(residmf, rtwidmf, grids, compInterval);

    if (Abs(rho) < m_small) {
      badDotProd = true;
    } else {
      badDotProd = false;
    }
    TBOTS_DotProduct1.stop();

    if (badDotProd) {
      TBOTS_badDotProd_plus.start();
      if (iter>1) { // corremf has nothing in it on 1st iteration
        for (dit.begin(); dit.ok(); ++dit) {
          a_phi[dit()].plus(corremf[dit()], 0, 0, ncomp);
        }
      }
      TBOTS_badDotProd_plus.stop();
      goto done;
    }

    if (iter == 1) {
      TBOTS_iter1stuff.start();
      for (dit.begin(); dit.ok(); ++dit) {
        pmf[dit()].copy(residmf[dit()]);
      }
      TBOTS_iter1stuff.stop();
      rho_previous = rho;
    } else {

      TBOTS_stuff.start();
      Real betaim1 = (rho/rho_previous)*(alpha/omega);

      // p(i) = r(i-1) + beta(i-1)*[p(i-1) - omega(i-1)*v(i-1)]
      for (dit.begin(); dit.ok(); ++dit) {
        vmf[dit()] *= -omega;
        pmf[dit()] += vmf[dit()];
        pmf[dit()] *= betaim1;
        pmf[dit()] += residmf[dit()];

        //pmf[dit()].plus(vmf[dit()], -omega, 0,0,1);
        //pmf[dit()] *= betaim1;
        //pmf[dit()] += residmf[dit()];
      } // end loop over grids
      TBOTS_stuff.stop();
    } // end if i ==  1 's else

    // solve M(phat) = p(i)
    TBOTS_preconditioner1.start();
    a_levelopPtr->levelPreconditioner(phatmf, pmf);
    TBOTS_preconditioner1.stop();

    // v(i) = A(phat)
    TBOTS_applyOp1.start();
    a_levelopPtr->applyOpH(phatmf, vmf);
    TBOTS_applyOp1.stop();

    // alpha(i) = rho(i-1)/(rtwid^T v(i))
    TBOTS_DotProduct2.start();
    Real denom = DotProduct(vmf, rtwidmf, grids, compInterval);
    bool badCorrection = false;
    if (Abs(denom) > small2*Abs(rho)) {
      alpha = rho/denom;
    } else {
      alpha = 0.0;
      badCorrection = true;
      pout() << " badCorrection denom=" << denom
             << "  small2*Abs(rho) = "  << small2*Abs(rho) << endl;
    }
    rho_previous = rho;
    TBOTS_DotProduct2.stop();

    // s = r(i-1) - alpha(i)*v(i)
    if (!badCorrection)        {
      // increment residual:   resid = resid - alpha*v
      // increment correction: corr  = corr  + alpha*phat
      TBOTS_incrementSolution.start();
      for (dit.begin(); dit.ok(); ++dit) {
        residmf[dit()].plus(vmf[dit()], -1.0*alpha, 0, 0, 1);
        corremf[dit()].plus(phatmf[dit()], alpha, 0, 0, 1);
      }
      TBOTS_incrementSolution.stop();
    } else {
      // if correction was bad, don't increment anything,
      // but set residual to zero (forces recomputation of resid) (not happening)
      for (dit.begin(); dit.ok(); ++dit) {
        residmf[dit()].setVal(0.0);
      }
    }

    // compute residuals
    TBOTS_computeRes_norm.start();
    rnorm = norm(residmf, compInterval, 0);
    TBOTS_computeRes_norm.stop();

    // check norm of s.
    // if small enough, step to end (forces recomputation of residual, etc)
    bool allDone = true;
    if (rnorm > m_solverTol*m_convergenceMetrics[0]) {
      allDone = false;
    }

    // if not "done" here, do second correction step
    if (!allDone) {

      // solve Mshat = s
      TBOTS_preconditioner2.start();
      a_levelopPtr->levelPreconditioner(shatmf, residmf);
      TBOTS_preconditioner2.stop();

      // t = A(shat)
      TBOTS_applyOp2.start();
      a_levelopPtr->applyOpH(shatmf, tmf);
      TBOTS_applyOp2.stop();

      // w(i) = tTresid/tTt
      TBOTS_DotProduct3.start();
      Real numerw = DotProduct(tmf, residmf, grids, compInterval);
      Real denomw = DotProduct(tmf, tmf,     grids, compInterval);

      if (Abs(denomw) > m_small) {
        omega = numerw/denomw;
      } else {
        omega = 0.0;
      }
      TBOTS_DotProduct3.stop();

      // increment correction: corr(i)  = corr(i-1)  + omega(i)*shat
      // increment residual:   resid(i) = resid(i-1) - omega(i)*t
      TBOTS_increment_residual.start();
      for (dit.begin(); dit.ok(); ++dit) {
        corremf[dit()].plus(shatmf[dit()],  omega, 0, 0, 1);
        residmf[dit()].plus(tmf[dit()],    -omega, 0, 0, 1);
      }
      TBOTS_increment_residual.stop();

    }

    // now check residual.
    TBOTS_check_convergence.start();
    allDone = true;
    rnorm = norm(residmf, compInterval, 0);

    if ((rnorm > m_solverTol*m_convergenceMetrics[0]) &&
        (Abs(omega) > m_small)) {
      allDone = false;
    }
    TBOTS_check_convergence.stop();

    // if residnorm is small enough (or if correction is
    // small enough), then recompute residual from scratch
    // and check again (to avoid roundoff-drift issues).  if
    // we _still_ think we're done, then exit
    TBOTS_doublecheck_convergence.start();
    if (allDone) {
      // recompute residual from scratch -- first increment
      // a_phi with correction
      for (dit.begin(); dit.ok(); ++dit) {
        a_phi[dit()].plus(corremf[dit()], 0, 0, ncomp);
      }

      a_levelopPtr->applyOpH(a_phi, residmf);

      for (dit.begin(); dit.ok(); ++dit) {
        residmf[dit()].minus(a_rhs[dit()], 0, 0, 1);
      }

      // now recompute residual norms and check for convergence
      allDone = true;
      rnorm = norm(residmf, compInterval, 0);

      if (rnorm > m_solverTol*m_convergenceMetrics[0]) {
        allDone = false;
      }

      if (allDone) {
        finished = true;
      } else {
        pout() << " BiCGStabSimple: allDone -- but need to try again " << endl;
        // otherwise, go back and try again with new residual

        // also reset \tilde{r} and correction
        for (dit.begin(); dit.ok(); ++dit) {
          rtwidmf[dit()].copy(residmf[dit()],0,0,ncomp);
          corremf[dit()].setVal(0.0);
        }

        finished = false;
      }

      //if (finished) {
      //pout() << "Bottom SmootherSimple final Norm = " << rnorm << endl;
      //}
    } // end if we're recomputing the residual
    TBOTS_doublecheck_convergence.stop();

    IterationTimer.stop();
    //sprintf(tempchar, "      BiCGStabSimple BS after %3d iters, rnorm=%11.4e (%8.2f) w=%5.3f\n",
    //      iter, rnorm, previous_rnorm/rnorm,
    //      IterationTimer.wc_time());
    //pout() << tempchar;
    previous_rnorm = rnorm;

  } // end main loop over BiCGStab iterations

 done:
  bool converged = true;
  if (rnorm >= m_solverTol*m_convergenceMetrics[0]) {
    converged = false;
  }

  if (badDotProd == true || initialResidualZero == true) {
    // if bad dot product or initial residual is zero, then don't bother
    // telling me it's not converged.
    // and set converged to be true for the sake of diagnostics below.
    converged = true;
  } else {
    if (!converged && m_verbose)   {
      pout() << "BiCGStabSmootherSimple not Converged! iter = " << iter << endl;
      pout() << " resid/metric = "
             << rnorm/(m_solverTol*m_convergenceMetrics[0]) << endl;
    }
  }
  TBOTS_doBottomSmooth.stop();

  BS_timer.stop();

#if 0
  // debug print statements
  int Npoints=0, Nboxes=0, Nghosts=0;
  NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);
  // Nboxes should now be same as grids.size()
  char stuff[180];
  sprintf(stuff, "    end doBSsimp L%d Npoints=%6d Nboxes=%4d iter=%2d conv=%1d zI=%1d badDP=%1d rnorm=%9.3e tol=%7.1e met=%9.3e wc=%6.3f\n",
          GlobalLevel, Npoints, Nboxes, iter, converged,
          initialResidualZero, badDotProd, rnorm,
          m_solverTol, m_solverTol*m_convergenceMetrics[0],
          BS_timer.wc_time());
  pout() << stuff;
#endif

  NoelsSooperIterationCounter[BIterationFlavor].Count++;
  NoelsSooperIterationCounter[BIterationFlavor].Value+=iter;
  if (!converged) {
    NoelsSooperIterationCounter[BIterationFlavor].NmaxIteration++;
  }
  NoelsSooperIterationCounter[BIterationFlavor].wc += BS_timer.wc_time();
  BIterationFlavor = 0;

}
