/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// from PoissonOp by DTGraves, Weds, July 21, 1999
// petermc, 20 Feb 2003
#ifdef HYPRE_AS_BS

#include "HyprePoissonOp.H"
#include "HypreBottomSmoother.H"
#include "memusage.H"

LevelOp*
HyprePoissonOp::new_levelop() const
{
  //pout() << " HyprePoissonOp::new_levelop() " << get_memory_usage_from_OS() << endl;
  //only check to see if the boundary conditions are
  //defined.  not necessary to have whole thing defined.
  //the boundary conditions need to be defined because
  //the interface does not know about the boundary
  //condtions.  the solvers call the other define functions
  //in our inimitable two-stage construction
  assert(m_isBCDefined);
  HyprePoissonOp* newop = new HyprePoissonOp();
  if(newop == NULL)
    {
      MayDay::Error("Out of Memory in HyprePoissonOp::new_levelop");
    }
  newop->setDomainGhostBC(m_domghostbc);
  // do you need this?
  if (m_bottomSmootherPtr != NULL)
    {
      newop->setBottomSmoother(*m_bottomSmootherPtr);
    }
  // added 20 feb 2003
  assert(m_bottomSmootherPtr != NULL);
  return static_cast<LevelOp*>(newop);
}

HyprePoissonOp::HyprePoissonOp():PoissonOp()
{
  //pout() << " HyprePoissonOp::HyprePoissonOp():PoissonOp() " << get_memory_usage_from_OS() << endl;
  // first calls PoissonOp(); you can't avoid this.
  setHypreValues();
}

HyprePoissonOp::~HyprePoissonOp()
{
  //pout() << " HyprePoissonOp::~HyprePoissonOp() " << get_memory_usage_from_OS() << endl;
  clearMemory();
  if (m_bottomSmootherPtr != NULL)
    {
      delete m_bottomSmootherPtr;
      m_bottomSmootherPtr = NULL;
    }
  //if(m_bottomOp_ptr != NULL)
  //{
  //  delete m_bottomOp_ptr;
  //  m_bottomOp_ptr = NULL;
  //}
}

/***********************/
// return to undefined state
/***********************/
void
HyprePoissonOp::setHypreValues()
{
  // replace the bottom smoother
  if (m_bottomSmootherPtr != NULL) delete m_bottomSmootherPtr;
  m_bottomSmootherPtr = new HypreBottomSmoother;
}

/***********************/
// return to undefined state
/***********************/
void
HyprePoissonOp::clearMemory()
{
  //pout() << " HyprePoissonOp::clearMemory() " << get_memory_usage_from_OS() << endl;
  m_dxLevel = -1.0;
  m_refRatio = -1;
  m_isDefined = false;
  m_ihcfiEnabled = false;
  // don't touch bottom smoother here, since we want it to be able
  // to pass through the define statement unaltered

  // huh? (ndk)
  m_bottomSmootherPtr->clearMemory();
}

/***********************/
/***********************/

void
HyprePoissonOp::define(const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real  a_dxLevel,
                       int a_refRatio,
                       const Box& a_domain,
                       bool a_homogeneousOnly,
                       int a_ncomp)
{
  //pout() << " HyprePoissonOp::define1 " << get_memory_usage_from_OS() << endl;
  ProblemDomain probdomain(a_domain);
  define(a_grids, a_baseBAPtr, a_dxLevel, a_refRatio, probdomain,
         a_homogeneousOnly, a_ncomp);
}

/***********************/
/***********************/

void
HyprePoissonOp::define(const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real  a_dxLevel,
                       int   a_refRatio,
                       const ProblemDomain& a_domain,
                       bool  a_homogeneousOnly,
                       int   a_ncomp)
{
  //pout() << " HyprePoissonOp::define2 " << get_memory_usage_from_OS() << endl;
  PoissonOp::define(a_grids, a_baseBAPtr, a_dxLevel, a_refRatio, a_domain,
                    a_homogeneousOnly, a_ncomp);
}

/***********************/
// a_reftofine is the refinement between this operator and a_opfine
/***********************/
void
HyprePoissonOp::define(const LevelOp* a_opfine, int a_reftoFine)
{
  //pout() << " HyprePoissonOp::define3 " << get_memory_usage_from_OS() << endl;
  PoissonOp::define(a_opfine, a_reftoFine);
}

void HyprePoissonOp::setBottomSmoother(const BaseBottomSmoother& a_bottomSmoother)
{
  if (m_bottomSmootherPtr != NULL)
    {
      delete m_bottomSmootherPtr;
    }

  m_bottomSmootherPtr = a_bottomSmoother.new_bottomSmoother();

  const HypreBottomSmoother* oldBSPtr = static_cast<const HypreBottomSmoother*>(&a_bottomSmoother);
  HypreBottomSmoother::hypre_solver_type solverType = oldBSPtr->solverType();

  HypreBottomSmoother* hypreBSPtr = static_cast<HypreBottomSmoother*>(m_bottomSmootherPtr);
  hypreBSPtr->define(m_dxLevel, m_dxCrse, m_domain, solverType);
}
#endif // #ifdef HYPRE_AS_BS
