/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Sat, July 18, 1999

#include "LevelMG.H"
#include "LevelMGF_F.H"

#include "Timer.H"
extern Timer Everything;

Timer TF_INTERPMG_LevelMG  ("INTERPMG LMG", Everything, DTIMEFORT, true);

// mgRelax called recursively, so can't do things here as normal.
//  measure mgRelax_all in LevelSolver.cpp and AMRLevelMG.cpp which
//  are the only places it's called from.
Timer TLMG_mgRelax_all     ("call mgRelax", 120);
Timer TLMG_downsmooth      ("down smooth",      TLMG_mgRelax_all);
Timer TLMG_applyOp         ("applyOp",          TLMG_mgRelax_all);
Timer TLMG_residualAlgebra ("residual algebra", TLMG_mgRelax_all);
Timer TLMG_avgTocoarse     ("avgToCoarse",      TLMG_mgRelax_all);
Timer TLMG_crseCorrect     ("crseCorrect",      TLMG_mgRelax_all);
Timer TLMG_upsmooth        ("up smooth",        TLMG_mgRelax_all);
Timer TLMG_1x1smooth       ("1x1 smooth",       TLMG_mgRelax_all);
Timer TLMG_bottomsmooth    ("bottom smooth",    TLMG_mgRelax_all);
Timer TLMG_bottomSmootherCG("bottomSmootherCG", TLMG_mgRelax_all);
Timer TLMG_lastsmooth      ("last smooth",      TLMG_mgRelax_all);

Timer TLevelMG_define("LevelMG define", Everything, DTIMEDEFINES, true);

extern int global_mgRelax_numBottomGSRB;
extern int global_mgRelax_numSmoothUp;
extern int global_mgRelax_numSmoothDown;

void LevelMG::setDefaultValues()
{
  // set defaults
  //m_numBottomGSRB = 4;
  //m_numSmoothUp = 4;
  //m_numSmoothDown = 4;

  m_numBottomGSRB = global_mgRelax_numBottomGSRB;
  m_numSmoothUp   = global_mgRelax_numSmoothUp;
  m_numSmoothDown = global_mgRelax_numSmoothDown;

  m_lCoarsePtr = NULL;
  m_levelopPtr = NULL;

  m_dxLevel = -1;

  m_isDefined = false;
}

bool LevelMG::isDefined() const
{
  return m_isDefined;
}

void LevelMG::clearMemory()
{
  if (m_lCoarsePtr != NULL)
    {
      delete m_lCoarsePtr;
      m_lCoarsePtr = NULL;
    }

  if (m_levelopPtr != NULL)
    {
      delete m_levelopPtr;
      m_levelopPtr = NULL;
    }

  m_isDefined = false;
}

// default constructor
LevelMG::LevelMG()
{
  setDefaultValues();
}

// Constructor with DisjointBoxLayout, number of coarser levels
// this calls the levelmg big define function
LevelMG::LevelMG(const DisjointBoxLayout& a_ba,
                 const DisjointBoxLayout* a_baseBaPtr,
                 Real                     a_dxLevel,
                 int                      a_refRatio,
                 const Box&               a_domain,
                 int                      a_nCoarserLevels,
                 const LevelOp* const     a_opin,
                 int                      a_ncomp)
{
  setDefaultValues();

  ProblemDomain physdomain(a_domain);

  define(a_ba,a_baseBaPtr, a_dxLevel, a_refRatio, physdomain,
         a_nCoarserLevels, a_opin, a_ncomp);
}

// Constructor with DisjointBoxLayout, number of coarser levels
// this calls the levelmg big define function
LevelMG::LevelMG(const DisjointBoxLayout& a_ba,
                 const DisjointBoxLayout* a_baseBaPtr,
                 Real                     a_dxLevel,
                 int                      a_refRatio,
                 const ProblemDomain&     a_domain,
                 int                      a_nCoarserLevels,
                 const LevelOp* const     a_opin,
                 int                      a_ncomp)
{
  setDefaultValues();

  define(a_ba,a_baseBaPtr, a_dxLevel, a_refRatio, a_domain,
         a_nCoarserLevels, a_opin, a_ncomp);
}

// corresponding define function
void LevelMG::define(const DisjointBoxLayout& a_ba,
                     const DisjointBoxLayout* a_baseBaPtr,
                     Real                     a_dxLevel,
                     int                      a_refRatio,
                     const  Box&              a_domain,
                     int                      a_nCoarserLevels,
                     const LevelOp* const     a_opin,
                     int                      a_ncomp)
{
  ProblemDomain physdomain(a_domain);

  define(a_ba, a_baseBaPtr, a_dxLevel, a_refRatio, a_domain,
         a_nCoarserLevels, a_opin, a_ncomp);
}

// corresponding define function
void LevelMG::define(const DisjointBoxLayout& a_ba,
                     const DisjointBoxLayout* a_baseBaPtr,
                     Real                     a_dxLevel,
                     int                      a_refRatio,
                     const  ProblemDomain&    a_domain,
                     int                      a_nCoarserLevels,
                     const LevelOp* const     a_opin,
                     int                      a_ncomp)
{
  TLevelMG_define.start();
  assert(a_opin != NULL);
  assert(a_nCoarserLevels >= 0);
  assert(!a_domain.isEmpty());
  assert(a_ba.checkPeriodic(a_domain));

  //either there are no coarser levels
  //or we have a legitimate ref ratio
  assert((a_refRatio > 0) || (a_baseBaPtr == NULL));
  assert(a_dxLevel > 0);

  clearMemory();

  m_ba = a_ba;
  m_baseBaPtr = a_baseBaPtr;

  m_dxLevel = a_dxLevel;
  m_domain  = a_domain;

  m_nCoarserLevels = a_nCoarserLevels;

  m_isDefined = true;
  m_resid.define(m_ba, a_ncomp, IntVect::TheZeroVector());
  m_levelopPtr = a_opin->new_levelop();

  // no need to do inhomgeneous CFInterp here
  // all levelmg operations (smooth, etc) use
  // only homogeneous cf interpolation
  bool homogeneousOnly = true;

  m_levelopPtr->define(m_ba,  m_baseBaPtr,
                       m_dxLevel,  a_refRatio,
                       m_domain, homogeneousOnly, a_ncomp);
  // recursively define MG hierarchy
  if (m_nCoarserLevels != 0)
    {

      //this constructor is called at the finest mg level.
      //this means that it might need an averaging operator
      //but it does not need an interpolation operator
      m_refToCoar = 2;

      if (m_baCoarsened.isClosed())
      {
        m_baCoarsened = DisjointBoxLayout();
      }

      coarsen(m_baCoarsened, m_ba, m_refToCoar);

      m_crseCorr .define(m_baCoarsened, a_ncomp, IntVect::TheUnitVector());
      m_crseResid.define(m_baCoarsened, a_ncomp, IntVect::TheZeroVector());

      m_averageOp.define(m_ba, m_baCoarsened, a_ncomp, m_refToCoar);

      m_lCoarsePtr = new LevelMG(*this, m_refToCoar, a_opin);

      if(m_lCoarsePtr == NULL)
        {
          MayDay::Error("Out of Memory in LevelMG::define");
        }
    }
  TLevelMG_define.stop();
}

// --------------------------------------------------------------
//  constructor for coarsened version of levelMG
// refcoarse is the ratio from coarsened thing to this thing
// --------------------------------------------------------------
LevelMG::LevelMG(const LevelMG& a_level,
                 int            a_refCoarse,
                 const LevelOp* a_opin)
{
  // set pointers to NULL so define does not try to delete them
  setDefaultValues();

  define(a_level, a_refCoarse, a_opin);
}

// --------------------------------------------------------------
// constructor for coarsened version of levelMG
// refcoarse is the ratio from coarsened thing to this thing
// --------------------------------------------------------------
void LevelMG::define(const LevelMG&       a_level,
                     int                  a_refCoarse,
                     const LevelOp* const a_opin)
{
  TLevelMG_define.start();
  assert(a_refCoarse == 2);
  assert(a_opin != NULL);

  clearMemory();

  m_isDefined = true;
  m_ba = a_level.m_baCoarsened;

  Real dxcoarse = a_refCoarse*a_level.m_dxLevel;
  int ncomp = a_level.m_resid.nComp();

  ProblemDomain dprobc = coarsen(a_level.m_domain, a_refCoarse);

  m_nCoarserLevels = a_level.m_nCoarserLevels - 1;

  assert(m_nCoarserLevels >= 0);

  m_baseBaPtr = a_level.m_baseBaPtr;
  m_dxLevel = dxcoarse;
  m_domain = dprobc;

  m_resid.define(m_ba, ncomp, IntVect::TheZeroVector());

  m_levelopPtr = a_level.m_levelopPtr->new_levelop();

  m_levelopPtr->define(a_level.m_levelopPtr, a_refCoarse);

  m_numSmoothDown = a_level.m_numSmoothDown;
  m_numSmoothUp = a_level.m_numSmoothUp;

  // recursively define MG hierarchy
  if (m_nCoarserLevels != 0)
    {
      m_refToCoar = 2;

      coarsen(m_baCoarsened, m_ba, m_refToCoar);

      m_crseCorr .define(m_baCoarsened, ncomp, IntVect::TheUnitVector());
      m_crseResid.define(m_baCoarsened, ncomp, IntVect::TheZeroVector());

      m_averageOp.define(m_ba, m_baCoarsened, ncomp, m_refToCoar);

      m_lCoarsePtr = new LevelMG(*this, m_refToCoar, a_opin);
      if (m_lCoarsePtr == NULL)
        {
          MayDay::Error("Out of Memory in LevelMG::define");
        }
    }
  else
    {
      m_lCoarsePtr = NULL;
    }
  TLevelMG_define.stop();
}

// ------------------------------------------------------------
// destructor
// ------------------------------------------------------------
LevelMG::~LevelMG()
{
  clearMemory();
}

// -------------------------------------------------------------
void
LevelMG::mgRelax(LevelData<FArrayBox>&       a_soln,
                 const LevelData<FArrayBox>& a_rhs,
                 bool                        a_bottomSolveFlag)
{
  assert(isDefined());

  int ncomp = a_rhs.nComp();

  assert(a_soln.nComp() == ncomp);
  assert (ncomp == m_resid.nComp());
  assert( a_soln.ghostVect() >= IntVect::TheUnitVector());

  // now, either coarsen or call bottom solver
  if (m_lCoarsePtr != NULL)
    {
      // first, smooth on this level
      TLMG_downsmooth.start();
      for (int iter=0; iter< m_numSmoothDown; iter++)
        {
          // calls one levelGSRB
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      TLMG_downsmooth.stop();

      TLMG_applyOp.start();
      DataIterator ditCoar = m_crseCorr.dataIterator();
      for (ditCoar.reset(); ditCoar.ok(); ++ditCoar)
        {
          m_crseCorr[ditCoar()].setVal(0.0);
        }

      // define residual on this grid
      // using homogeneous cf bcs
      m_levelopPtr->applyOpH(a_soln,m_resid);
      TLMG_applyOp.stop();

      TLMG_residualAlgebra.start();
      DataIterator dit = m_resid.dataIterator();
      for (dit.reset(); dit.ok(); ++dit)
        {
          m_resid[dit()] -= a_rhs[dit()];
          m_resid[dit()].negate();
        }
      TLMG_residualAlgebra.stop();

      // coarsen residual and physical boundary condition
      TLMG_avgTocoarse.start();
      assert(m_averageOp.isDefined());
      m_averageOp.averageToCoarse(m_crseResid, m_resid);
      TLMG_avgTocoarse.stop();

      // relax recursively on coarsened level
      m_lCoarsePtr->mgRelax(m_crseCorr, m_crseResid, a_bottomSolveFlag);

      TLMG_crseCorrect.start();
      // interpolate correction back up and modify solution
      crseCorrect(a_soln, m_crseCorr, m_refToCoar);
      TLMG_crseCorrect.stop();

      // smooth again on this level
      TLMG_upsmooth.start();
      for (int iter=0; iter < m_numSmoothUp; iter++)
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      TLMG_upsmooth.stop();
    }
  else
  if (m_domain.domainBox().numPts() == 1)
    {
      // grid is 1x1, just need to smooth
      TLMG_1x1smooth.start();
      for (int iter = 0; iter < m_numBottomGSRB; iter++)
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      TLMG_1x1smooth.stop();
    }
  else
  if (a_bottomSolveFlag)
    {
      // first, smooth on this level
      TLMG_bottomsmooth.start();
      for (int iter = 0; iter < m_numBottomGSRB; iter++)
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      TLMG_bottomsmooth.stop();

      // then call bottom smoother
      TLMG_bottomSmootherCG.start();
      // this is the only place that bottomSmoother() is called.  (noel)
      m_levelopPtr->bottomSmoother(a_soln,a_rhs);
      TLMG_bottomSmootherCG.stop();
    }
  else
    {
      // if we get here, then there is no coarser level, but
      // we're not at the bottom.  This means that we're at the bottom
      // of a mini-V-cycle between AMR levels.  In this case, we just
      // call smooth on this level.  Do this max(numSmoothUp,numSmoothDown)
      // in case one or the other is zero
      int numSmooth = std::max(m_numSmoothUp, m_numSmoothDown);

      // do numSmoothDown iterations
      TLMG_lastsmooth.start();
      for (int iter = 0; iter < numSmooth; iter++)
        {
          m_levelopPtr->smooth(a_soln,a_rhs);
        }
      TLMG_lastsmooth.stop();
    }
}

void LevelMG::crseCorrect(LevelData<FArrayBox>&       a_fine,
                          const LevelData<FArrayBox>& a_crse,
                          int                         a_refRat)
{
  DataIterator crseIt = a_crse.dataIterator();

  const DisjointBoxLayout& crseBoxes = a_crse.getBoxes();
  const DisjointBoxLayout& fineBoxes = a_fine.getBoxes();

  for (crseIt.reset(); crseIt.ok(); ++crseIt)
    {
      const Box& crseBox = crseBoxes[crseIt()];
      const FArrayBox& crseFab = a_crse[crseIt()];

      FArrayBox& fineFab = a_fine[crseIt()];
      Box fineBox = fineBoxes[crseIt()];

      fineBox.coarsen(a_refRat);
      fineBox &= crseBox;

      // c set refinement box
      Box nrefbox(IntVect::TheZeroVector(),
                  (a_refRat-1)*IntVect::TheUnitVector());

      if (!fineBox.isEmpty())
        {
          TF_INTERPMG_LevelMG.start();
          FORT_INTERPMG(CHF_FRA(fineFab),
                        CHF_FRA(crseFab),
                        CHF_BOX(fineBox),
                        CHF_CONST_INT(a_refRat),
                        CHF_BOX(nrefbox));
          TF_INTERPMG_LevelMG.stop();
         }
    }
}

LevelOp* LevelMG::levelOpPtr()
{
  return m_levelopPtr;
}

LevelMG* LevelMG::lCoarsePtr()
{
  return m_lCoarsePtr;
}

/// Set number of smoothing steps on the way up V-cycle
void
LevelMG::setNumSmoothUp(int a_numSmoothUp)
{
  m_numSmoothUp = a_numSmoothUp;
  // may need to pass this down the chain
  if (lCoarsePtr() != NULL)
    {
      lCoarsePtr()->setNumSmoothUp(a_numSmoothUp);
    }
}

/// Set number of smoothing steps on the way down V-cycle
void
LevelMG::setNumSmoothDown(int a_numSmoothDown)
{
  m_numSmoothDown = a_numSmoothDown;
  // may need to pass this down the chain
  if (lCoarsePtr() != NULL)
    {
      lCoarsePtr()->setNumSmoothDown(a_numSmoothDown);
    }
}

void
LevelMG::setConvergenceMetric(Real a_metric, int a_comp)
{
  // first set on this level, then pass down to coarser level
  assert (m_levelopPtr != NULL);
  m_levelopPtr->setConvergenceMetric(a_metric, a_comp);

  if (m_lCoarsePtr != NULL)
    {
      m_lCoarsePtr->setConvergenceMetric(a_metric, a_comp);
    }
}
