/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _MGINTERP_H_
#define _MGINTERP_H_

#include "REAL.H"
#include "LevelData.H"
#include "BaseFab.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"

class DisjointBoxLayout;

/// Internal Chombo class to interpolate coarse level data to a fine level
/**
    Class to interpolate coarse level data to a fine level
    data for multigrid.  This class should be considered internal
    to AMRSolver and not a part of the Chombo API.
*/
class MGInterp
{
public:
  /// Null constructor
  MGInterp();

  /// Destructor
  ~MGInterp();

  /// Full constructor (calls full define)
  MGInterp(const DisjointBoxLayout& a_fineDomain,
           int                      a_numcomps,
           int                      a_refRatio,
           const Box&               a_problemDomain);

  /// Full constructor (calls full define)
  MGInterp(const DisjointBoxLayout& a_fineDomain,
           int                      a_numcomps,
           int                      a_refRatio,
           const ProblemDomain&     a_problemDomain);

  /// Full define
  /**
      Specify the fine level domain and refinement ratio.
  */
  void define(const DisjointBoxLayout& a_fineDomain,     // the fine level domain
              int                      a_numcomps,       // the number of components
              int                      a_refRatio,       // the refinement ratio
              const Box&               a_problemDomain); // problem domain

  /// Full define
  /**
      Specify the fine level domain and refinement ratio.
  */
  void define(const DisjointBoxLayout& a_fineDomain,     // the fine level domain
              int                      a_numcomps,       // the number of components
              int                      a_refRatio,       // the refinement ratio
              const ProblemDomain&     a_problemDomain); // problem domain

  ///
  /**
      Interpolate coarse level data to fine level
  */
  void interpToFine(LevelData<FArrayBox>&       a_fineData,
                    const LevelData<FArrayBox>& a_coarseData);

  /// Is the object defined
  bool isDefined() const;

protected:
  bool m_isDefined;

  // the refinement ratio
  int m_refRatio;

  // work array for the coarse level data in a domain that is the
  // outline of the fine level domain on the coarse level
  //
  LevelData<FArrayBox> m_coarsenedFineData;
  DisjointBoxLayout m_grids;

  // problem domain
  ProblemDomain m_problemDomain;
};

#endif
