/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DTGraves, Weds, July 21, 1999

//#define WRITE_BOTTOMSMOOTH_FILE

#include <fstream>
#include <iostream>
#include <iomanip>
using namespace std;

// my broken out INTERPHOMO
#include "InterpHomoF_F.H"

#include "PoissonOp.H"
#include "PoissonOpF_F.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
#include "BiCGStabSmoother.H"
#include "BiCGStabSmootherSimple.H"
#include "CGSmoother.H"
#include "CGSquaredSmoother.H"
#include "GSRBSmoother.H"

#include "AMRIO.H"
int NbottomSmooth=0;

#include <cstdio>
//#include "GSRBF_F.H"
#include "Timer.H"
#include "NoelsPrettyPrint.H"
#ifdef COUNT_PEX
extern double pex_delta_time[NPEX];
extern unsigned int pex_Nlocal[NPEX];
extern unsigned int pex_Nto[NPEX];
extern unsigned int pex_Nfrom[NPEX];
extern unsigned int pex_GSRBvol[NPEX];
#endif

extern int write_bottomsmoother_file;  // hack for now (noel)
extern int num_levelGSRBs_inPC;  // hack for now (noel)
extern int global_relaxation_method;
extern int global_BS_algorithm;
extern Real global_GSRB_relaxFactor;

extern Timer Everything;
Timer TF_GSRBLEVELLAP               ("GSRBLEVELLAP",     Everything, DTIMEFORT, true);
Timer TF_INTERPHOMOPOI              ("INTERPHOMOP",      Everything, DTIMEFORT, true);
Timer TF_applyOpI_OPERATORLAP       ("aOpI_OLAP",        Everything, DTIMEFORT, true);
Timer TF_applyOpIcfHphys_OPERATORLAP("aOpIcfHphys_OLAP", Everything, DTIMEFORT, true);
Timer TF_applyOpH_OPERATORLAP       ("aOpH_OLAP",        Everything, DTIMEFORT, true);
Timer TF_applyOpHcfIphys_OPERATORLAP("aOpHcfIphys_OLAP", Everything, DTIMEFORT, true);

Timer TMPI_Poi_levelGSRB_tempBarrier  ("P ex tempBarr",   Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_tempBarrierL1("P ex tempBarrL1", Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_tempBarrierL2("P ex tempBarrL2", Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_tempBarrierL3("P ex tempBarrL3", Everything, DTIMECOMM, true);

Timer TMPI_Poi_levelGSRB_exchange      ("PlevelGSRB ex",    Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_exchangeL1    ("PlevelGSRB exL1",  Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_exchangeL2    ("PlevelGSRB exL2",  Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelGSRB_exchangeL3    ("PlevelGSRB exL3",  Everything, DTIMECOMM, true);
Timer TMPI_Poi_applyOpI_exchange       ("PapplyOpI ex",     Everything, DTIMECOMM, true);
Timer TMPI_Poi_applyOpIcfHphys_exchange("PapplyOpIcfH ex",  Everything, DTIMECOMM, true);
Timer TMPI_Poi_applyOpH_exchange       ("PapplyOpH ex",     Everything, DTIMECOMM, true);
Timer TMPI_Poi_applyOpHcfIphys_exchange("PapplyOpHcfI ex",  Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelPreconditioner_cp  ("PlevelPreC cp",    Everything, DTIMECOMM, true);
Timer TMPI_Poi_levelPreconditioner_copier("PlevelPreC cop", Everything, DTIMECOMM, true);

Timer TPlevelGSRB("in PlevelGSRB", 130);
Timer TPlevelGSRB_B1       ("temp B1",       TPlevelGSRB);
Timer TPlevelGSRB_CFInterp ("homo CFInterp", TPlevelGSRB);
Timer TPlevelGSRB_B2       ("temp B2",       TPlevelGSRB);
Timer TPlevelGSRB_ex       ("hot exchange",  TPlevelGSRB);
Timer TPlevelGSRB_B3       ("temp B3",       TPlevelGSRB);
Timer TPlevelGSRB_applyLoop("applyBC Loop",  TPlevelGSRB);
Timer TPlevelGSRB_B4       ("temp B4",       TPlevelGSRB);
Timer TPlevelGSRB_LAPloop  ("LAP loop",      TPlevelGSRB);
Timer TPlevelGSRB_B5       ("temp B5",       TPlevelGSRB);

Timer TGSRB_Poi_smooth ("Poi smooth",  Everything, DTIMEALLGSRB, true);
Timer TGSRB_Poi_PreCon1("Poi PreCon1", Everything, DTIMEALLGSRB, true);
Timer TGSRB_Poi_PreCon2("Poi PreCon2", Everything, DTIMEALLGSRB, true);

void PoissonOp::setDomainGhostBC(const DomainGhostBC& a_dombcIn)
{
  m_domghostbc = a_dombcIn;
  m_isBCDefined = true;
}

void PoissonOp::setBottomSmoother(const BaseBottomSmoother& a_bottomSmoother)
{
  if (m_bottomSmootherPtr != NULL)
    {
      delete m_bottomSmootherPtr;
    }

  m_bottomSmootherPtr = a_bottomSmoother.new_bottomSmoother();
}

bool PoissonOp::isDefined() const
{
  return (m_isDefined && m_isBCDefined);
}

LevelOp* PoissonOp::new_levelop() const
{
  // only check to see if the boundary conditions are
  // defined.  not necessary to have whole thing defined.
  // the boundary conditions need to be defined because
  // the interface does not know about the boundary
  // condtions.  the solvers call the other define functions
  // in our inimitable two-stage construction
  assert(m_isBCDefined);

  PoissonOp* newop = new PoissonOp();

  if (newop == NULL)
    {
      MayDay::Error("Out of Memory in PoissonOp::new_levelop");
    }

  newop->setDomainGhostBC(m_domghostbc);

  if (m_bottomSmootherPtr != NULL)
    {
      newop->setBottomSmoother(*m_bottomSmootherPtr);
    }

  return static_cast<LevelOp*>(newop);
}

/***********************/
// return to undefined state
/***********************/
void PoissonOp::setDefaultValues()
{
  m_dxLevel = -1.0;
  m_dxCrse = -1.0;
  m_refRatio = -1;

  m_ihcfiEnabled = false;
  m_isBCDefined = false;

  // default is to use BiCGStab as bottom smoother.
  switch (global_BS_algorithm) {
  case 0:
    m_bottomSmootherPtr = new BiCGStabSmoother;
    break;
  case 1:
    m_bottomSmootherPtr = new BiCGStabSmootherSimple;
    break;
  case 2:
    m_bottomSmootherPtr = new CGSmoother;
    break;
  case 3:
    m_bottomSmootherPtr = new CGSquaredSmoother;
    break;
  case 4:
    m_bottomSmootherPtr = new GSRBSmoother;
    break;
  default:
    m_bottomSmootherPtr = new BiCGStabSmoother;
  }

  m_isDefined = false;
}

/***********************/
// return to undefined state
/***********************/
void PoissonOp::clearMemory()
{
  m_dxLevel = -1.0;
  m_dxCrse = -1.0;
  m_refRatio = -1;

  m_ihcfiEnabled = false;

  // don't touch bottom smoother here, since we want it to be able
  // to pass through the define statement unaltered

  m_isDefined = false;
}

/***********************/
// default constructor
/***********************/
PoissonOp::PoissonOp()
  :LevelOp()
{
  setDefaultValues();
}

/***********************/
/***********************/
void PoissonOp::define(const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real                     a_dxLevel,
                       int                      a_refRatio,
                       const Box&               a_domain,
                       bool                     a_homogeneousOnly,
                       int                      a_ncomp)
{
  ProblemDomain probdomain(a_domain);

  define(a_grids, a_baseBAPtr, a_dxLevel, a_refRatio, probdomain,
         a_homogeneousOnly, a_ncomp);
}

void PoissonOp::define(const DisjointBoxLayout& a_grids,
                       const DisjointBoxLayout* a_baseBAPtr,
                       Real                     a_dxLevel,
                       int                      a_refRatio,
                       const ProblemDomain&     a_domain,
                       bool                     a_homogeneousOnly,
                       int                      a_ncomp)
{
  clearMemory();

  m_isDefined = true;

  assert(!a_domain.isEmpty());
  assert(a_grids.checkPeriodic(a_domain));

  m_dxLevel = a_dxLevel;
  m_domain = a_domain;
  m_grids = a_grids;

  m_exchangeCopier.define(a_grids, a_grids, IntVect::Unit, true);

  m_refRatio = a_refRatio;
  m_dxCrse = m_refRatio*m_dxLevel;

  m_ihcfiEnabled = (!a_homogeneousOnly);

  m_ncomp = a_ncomp;

  if (m_ihcfiEnabled)
    {
      m_quadCFI.define(a_grids,a_baseBAPtr, a_dxLevel,
                       a_refRatio, m_ncomp, m_domain);
    }

  if (a_baseBAPtr != NULL)
    {
      m_baseBA = *a_baseBAPtr;
    }

  DataIterator lit = m_grids.dataIterator();
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loCFIVS[idir].define(m_grids);
      m_hiCFIVS[idir].define(m_grids);

      for (lit.begin(); lit.ok(); ++lit)
        {
          m_loCFIVS[idir][lit].define(m_domain, m_grids.get(lit),
                                      m_grids, idir,Side::Lo);
          m_hiCFIVS[idir][lit].define(m_domain, m_grids.get(lit),
                                      m_grids, idir,Side::Hi);
        }
    }
}

// special define for use in BottomSmoother Test (noel)
void PoissonOp::define(const DisjointBoxLayout& a_grids,
                       Real                     a_dxLevel,
                       Real                     a_dxCoarse,
                       const ProblemDomain&     a_domain,
                       int                      a_ncomp)
{
  clearMemory();

  m_isDefined = true;

  assert(!a_domain.isEmpty());
  assert(a_grids.checkPeriodic(a_domain));

  m_dxLevel = a_dxLevel;
  m_domain = a_domain;
  m_grids = a_grids;

  m_exchangeCopier.define(a_grids, a_grids, IntVect::Unit, true);

  //m_refRatio = a_refRatio;
  m_dxCrse = a_dxCoarse;

  m_ihcfiEnabled = false;

  m_ncomp = a_ncomp;

  DataIterator lit = m_grids.dataIterator();
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loCFIVS[idir].define(m_grids);
      m_hiCFIVS[idir].define(m_grids);

      for (lit.begin(); lit.ok(); ++lit)
        {
          m_loCFIVS[idir][lit].define(m_domain, m_grids.get(lit),
                                      m_grids, idir,Side::Lo);
          m_hiCFIVS[idir][lit].define(m_domain, m_grids.get(lit),
                                      m_grids, idir,Side::Hi);
        }
    }
}

// a_reftofine is the refinement between this operator and a_opfine
void PoissonOp::define(const LevelOp* a_opfine,
                       int            a_reftoFine)
{
  clearMemory();

  const PoissonOp* opfineptr = dynamic_cast<const PoissonOp*>(a_opfine);

  if (opfineptr == NULL)
    {
      MayDay::Error("PoissonOp::define: casting failed");
    }

  const PoissonOp& opfine = *opfineptr;

  assert(opfine.isDefined());
  assert(a_reftoFine > 0);

  m_isDefined = true;
  m_ihcfiEnabled = false;

  setDomainGhostBC(opfine.m_domghostbc);

  m_dxLevel = (opfine.m_dxLevel)*a_reftoFine;
  m_domain = coarsen(opfine.m_domain,a_reftoFine);

  coarsen(m_grids, opfine.m_grids, a_reftoFine);

  m_exchangeCopier.define(m_grids, m_grids, IntVect::Unit, true);

  m_dxCrse = opfine.m_dxCrse;
  m_ncomp = opfine.m_ncomp;

  // this refratio is the ratio between this level and the
  // next level in the amr hierarchy
  m_refRatio = -1;

  // don't need to define the cfinterp since inhomogeneous cfinterpolaton
  // is never done here.  We shall leave m_quadCFI and base_ba undefined and
  // and just leave the refratio -1.  We DO need fineinterpIVS for doing
  // homogeneous cfinterpolation
  DataIterator lit = m_grids.dataIterator();
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loCFIVS[idir].define(m_grids);
      m_hiCFIVS[idir].define(m_grids);

      for (lit.begin(); lit.ok(); ++lit)
        {
          m_loCFIVS[idir][lit()].define(m_domain, m_grids.get(lit()),
                                        m_grids, idir,Side::Lo);
          m_hiCFIVS[idir][lit()].define(m_domain, m_grids.get(lit()),
                                        m_grids, idir,Side::Hi);
        }
    }

  setBottomSmoother(*(opfine.m_bottomSmootherPtr));
}

/***********************/
//  apply coarse-fine boundary conditions -- assume that phi grids
//  are grown by one
/***********************/
void PoissonOp::CFInterp(LevelData<FArrayBox>&       a_phif,
                         const LevelData<FArrayBox>& a_phic)
{
  assert(isDefined());
  assert(m_ihcfiEnabled);
  assert( a_phif.ghostVect() >= IntVect::TheUnitVector());
  assert(m_quadCFI.isDefined());

  m_quadCFI.coarseFineInterp(a_phif, a_phic);
}

/***********************/
// does homogeneous coarse/fine interpolation
/***********************/
void PoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif)
{
  assert(isDefined());
  assert(a_phif.ghostVect() >= IntVect::TheUnitVector());

  DataIterator dit = a_phif.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      const DataIndex& datInd = dit();

      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              homogeneousCFInterp(a_phif,datInd,idir,sit());
            }
        }
    }
}

/***********************/
// does homogeneous coarse/fine interpolation
/***********************/
void PoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif,
                                    const DataIndex&      a_datInd,
                                    int                   a_idir,
                                    Side::LoHiSide        a_hiorlo)
{
  assert(isDefined());
  assert((a_idir >= 0) && (a_idir  < SpaceDim));
  assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  assert( a_phif.ghostVect() >= IntVect::TheUnitVector());
  //  assert (m_ncomp == a_phif.nComp());

  const CFIVS* cfivsPtr = NULL;

  if (a_hiorlo == Side::Lo)
    {
      cfivsPtr = &m_loCFIVS[a_idir][a_datInd];
    }
  else
    {
      cfivsPtr = &m_hiCFIVS[a_idir][a_datInd];
    }

  if (cfivsPtr->isPacked())
    {
      int ihiorlo = sign(a_hiorlo);
      //TF_INTERPHOMOPOI.start(); //calledTooMuch
      FORT_OLDINTERPHOMO(CHF_FRA(a_phif[a_datInd]),
                         CHF_BOX(cfivsPtr->packedBox()),
                         CHF_CONST_REAL(m_dxLevel),
                         CHF_CONST_REAL(m_dxCrse),
                         CHF_CONST_INT(a_idir),
                         CHF_CONST_INT(ihiorlo));
      //TF_INTERPHOMOPOI.stop(); //calledTooMuch
    }
  else
    {
      const IntVectSet& interpIVS = cfivsPtr->getFineIVS();

      if (!interpIVS.isEmpty())
        {
          //TimePoissonOpinterpOnIVS.start();
          // Assuming homogenous, interpolate on fine ivs
          interpOnIVSHomo(a_phif, a_datInd, a_idir,
                          a_hiorlo, interpIVS);
          //TimePoissonOpinterpOnIVS.stop();
        }
    }
}

void PoissonOp::interpOnIVSHomo(LevelData<FArrayBox>& a_phif,
                                const DataIndex&      a_datInd,
                                const int             a_idir,
                                const Side::LoHiSide  a_hiorlo,
                                const IntVectSet&     a_interpIVS)
{
  assert(isDefined());
  assert((a_idir >= 0) && (a_idir  < SpaceDim));
  assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  assert(a_phif.ghostVect() >= IntVect::TheUnitVector());

  IVSIterator fineIVSit(a_interpIVS);
  FArrayBox& a_phi = a_phif[a_datInd];

  assert(m_ncomp = a_phi.nComp());

  // much of these scalar values can be precomputed and stored if
  // we ever need to speed-up this function (ndk)
  Real x1 = m_dxLevel;
  Real x2 = 0.5*(3.0*m_dxLevel+m_dxCrse);

  Real denom = 1.0-((x1+x2)/x1);
  Real idenom = 1/(denom); // divide is more expensive usually

  Real x = 2.0*m_dxLevel;
  Real xsquared = x*x;

  Real m1 = 1.0/(x1*x1);
  Real m2 = 1.0/(x1*(x1-x2));

  Real q1 = 1.0/(x1-x2);
  Real q2 = x1+x2;

  int ihilo = sign(a_hiorlo);
  Real pa,pb,a,b;

  IntVect ivf;
  for (fineIVSit.begin(); fineIVSit.ok(); ++fineIVSit)
    {
      ivf = fineIVSit();

      // quadratic interpolation
      for(int ivar = 0; ivar < m_ncomp; ivar++)
        {
          ivf[a_idir]-=2*ihilo;
          pa = a_phi(ivf, ivar);

          ivf[a_idir]+=ihilo;
          pb = a_phi(ivf, ivar);

          a = ((pb-pa)*m1 - (pb)*m2)*idenom;
          b = (pb)*q1 - a*q2;

          a_phi(fineIVSit(), ivar) = a*xsquared + b*x + pa;

          // this returns the ivf to the original point
          ivf[a_idir]+=ihilo;
        } //end loop over components
    } //end loop over fine intvects
}

PoissonOp::~PoissonOp()
{
  clearMemory();

  if (m_bottomSmootherPtr != NULL)
    {
      delete m_bottomSmootherPtr;
      m_bottomSmootherPtr = NULL;
    }
}

/***********************/
// this smoother assumes problem has
// already been put into residual-correction
//  form, so that crse/fine BC's are homogeneous
/***********************/
void
PoissonOp::smooth(LevelData<FArrayBox>&       a_phi,
                  const LevelData<FArrayBox>& a_rhs)
{
  assert(isDefined());
  TGSRB_Poi_smooth.start();
  levelGSRB(a_phi,a_rhs);
  TGSRB_Poi_smooth.stop();
}

/**************************/
// this preconditioner first initializes phihat to (IA)phihat = rhshat
// (diagonization of L -- A is the matrix version of L)
// then smooths with a couple of passes of levelGSRB
/**************************/
void
PoissonOp::levelPreconditioner(LevelData<FArrayBox>&       a_phihat,
                               const LevelData<FArrayBox>& a_rhshat)
{
  // diagonal term of this operator is 4/h/h in 2D, 6/h/h in 3D,
  // so inverse of this is our initial multiplier
  Real mult = -m_dxLevel*m_dxLevel/(2.0*SpaceDim);
  Interval comps = a_phihat.interval();

  assert(a_phihat.nComp() == a_rhshat.nComp());

  TMPI_Poi_levelPreconditioner_cp.start();
  // don't need to use a Copier -- plain copy will do
  DataIterator dit = a_phihat.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      a_phihat[dit()].copy(a_rhshat[dit()].box(), comps,
                           a_rhshat[dit()].box(), a_rhshat[dit()],
                           comps);

      a_phihat[dit()] *= mult;
    }
  TMPI_Poi_levelPreconditioner_cp.stop();

  if (num_levelGSRBs_inPC >= 1) {
    TGSRB_Poi_PreCon1.start();
    levelGSRB(a_phihat, a_rhshat);
    TGSRB_Poi_PreCon1.stop();
  }
  if (num_levelGSRBs_inPC >= 2) {
    for (int i=1; i < num_levelGSRBs_inPC; i++) {
      TGSRB_Poi_PreCon2.start();
      levelGSRB(a_phihat, a_rhshat);
      TGSRB_Poi_PreCon2.stop();
    }
  }

}

/***********************/
//  levelGSRB does GSRB on a level.  it has no knowledge of overlying
//  finer grids, and assumes coarse/fine BC's are homogeneous
/***********************/
void PoissonOp::levelGSRB(LevelData<FArrayBox>&       a_phi,
                          const LevelData<FArrayBox>& a_rhs)
{
  TPlevelGSRB.start();
  assert(isDefined());
  assert(a_phi.isDefined());
  assert(a_rhs.isDefined());
  assert(a_phi.ghostVect() >= IntVect::TheUnitVector());
  assert(a_phi.nComp() == a_rhs.nComp());

#ifdef COUNT_PEX
  static double previousTime_exchange=0.0;
  double deltaTime_exchange=0.0;
#endif

#if 0
  char tempstring[180];
  static double previousTime_CFInterp=0.0;
  double deltaTime_CFInterp=0.0;
  static double previousTime_apply=0.0;
  double deltaTime_apply=0.0;
  static double previousTime_gsrb=0.0;
  double deltaTime_gsrb=0.0;
#endif

  // do first red, then black passes
  // global_relaxation_method = GS 0   GSRB 1 (default GSRB)
  for (int whichPass = 0; whichPass <= global_relaxation_method; whichPass++)
    {

#ifdef TIMED_BARR
      TPlevelGSRB_B1.start();
      barrier();
      TPlevelGSRB_B1.stop();
#endif

      TPlevelGSRB_CFInterp.start(); //calledTooMuch
      // do coarse/fine and copy bc's
      //should be done patch by patch.
      homogeneousCFInterp(a_phi);
      TPlevelGSRB_CFInterp.stop(); //calledTooMuch
      //deltaTime_CFInterp = TPlevelGSRB_CFInterp.wc_time() - previousTime_CFInterp;
      //previousTime_CFInterp = TPlevelGSRB_CFInterp.wc_time();

      // now step through grids...
      Real dx = m_dxLevel;
      DataIterator dit = a_phi.dataIterator();

#ifdef TIMED_BARR
      // time the barrier
      TPlevelGSRB_B2.start();
      if (GlobalLevel == 1)      TMPI_Poi_levelGSRB_tempBarrierL1.start();
      else if (GlobalLevel == 2) TMPI_Poi_levelGSRB_tempBarrierL2.start();
      else if (GlobalLevel == 3) TMPI_Poi_levelGSRB_tempBarrierL3.start();
      else                       TMPI_Poi_levelGSRB_tempBarrier.start();
      barrier();
      if (GlobalLevel == 1)      TMPI_Poi_levelGSRB_tempBarrierL1.stop();
      else if (GlobalLevel == 2) TMPI_Poi_levelGSRB_tempBarrierL2.stop();
      else if (GlobalLevel == 3) TMPI_Poi_levelGSRB_tempBarrierL3.stop();
      else                       TMPI_Poi_levelGSRB_tempBarrier.stop();
      TPlevelGSRB_B2.stop();
#endif

      if (GlobalLevel == 1)      TMPI_Poi_levelGSRB_exchangeL1.start();
      else if (GlobalLevel == 2) TMPI_Poi_levelGSRB_exchangeL2.start();
      else if (GlobalLevel == 3) TMPI_Poi_levelGSRB_exchangeL3.start();
      else                       TMPI_Poi_levelGSRB_exchange.start();
      TPlevelGSRB_ex.start(); //calledTooMuch
      //fill in intersection of ghostcells and a_phi's boxes
#ifdef USE_MYEXCHANGE_IN_POISSONOP
      a_phi.myexchange(a_phi.interval(), m_exchangeCopier);
#else
      a_phi.exchange(a_phi.interval(), m_exchangeCopier);
#endif
      if (GlobalLevel == 1)      TMPI_Poi_levelGSRB_exchangeL1.stop();
      else if (GlobalLevel == 2) TMPI_Poi_levelGSRB_exchangeL2.stop();
      else if (GlobalLevel == 3) TMPI_Poi_levelGSRB_exchangeL3.stop();
      else                       TMPI_Poi_levelGSRB_exchange.stop();
      TPlevelGSRB_ex.stop(); //calledTooMuch

#ifdef TIMED_BARR
      TPlevelGSRB_B3.start();
      barrier();
      TPlevelGSRB_B3.stop();
#endif

      TPlevelGSRB_applyLoop.start(); //calledTooMuch
      for (dit.begin(); dit.ok(); ++dit)
        {
          // invoke physical BC's where necessary
          m_domghostbc.applyHomogeneousBCs(a_phi[dit()],
                                           m_domain,
                                           m_dxLevel);
        }
      TPlevelGSRB_applyLoop.stop(); //calledTooMuch

#ifdef TIMED_BARR
      TPlevelGSRB_B4.start();
      barrier();
      TPlevelGSRB_B4.stop();
#endif

      //deltaTime_apply = TPlevelGSRB_applyLoop.wc_time() - previousTime_apply;
      //previousTime_apply = TPlevelGSRB_applyLoop.wc_time();

      //int volTot=0;
      //Timer gsrb;
      TPlevelGSRB_LAPloop.start(); //calledTooMuch
      TF_GSRBLEVELLAP.start();
      for (dit.begin(); dit.ok(); ++dit)
        {
#ifndef NDEBUG
          FArrayBox& thisPhi = a_phi[dit()];
          const FArrayBox& thisRhs = a_rhs[dit()];
          const Box& thisBox = m_grids.get(dit());

          assert(thisRhs.box().contains(thisBox));
          assert(thisPhi.box().contains(grow(thisBox,1)));
#endif
          //gsrb.clear();
          //gsrb.start();
          // if GS, then global_relaxation_method will be 0 and
          // we are only doing one pass.  need to tell fortran that
          // which i'll do by passing in a "2" into the whichPass flag.
          int whichPassPrime=whichPass;
          if (global_relaxation_method==0) whichPassPrime=2;
#ifdef TEST_MYGSRBLEVELLAP
          FORT_MYGSRBLEVELLAP(CHF_FRA(a_phi[dit()]),
                              CHF_CONST_FRA(a_rhs[dit()]),
                              CHF_BOX(m_grids.get(dit())),
                              CHF_CONST_REAL(dx),
                              CHF_CONST_INT(whichPassPrime),
                              CHF_CONST_REAL(global_GSRB_relaxFactor));
#else
          FORT_OLDGSRBLEVELLAP(CHF_FRA(a_phi[dit()]),
                               CHF_CONST_FRA(a_rhs[dit()]),
                               CHF_BOX(m_grids.get(dit())),
                               CHF_CONST_REAL(dx),
                               CHF_CONST_INT(whichPassPrime),
                               CHF_CONST_REAL(global_GSRB_relaxFactor));
#endif
          //gsrb.stop();

          //const Box& thisBox = m_grids.get(dit());
          //int vol = thisBox.numPts();
          //volTot+=vol;
          //     Real gwc = gsrb.wc_time()*1000;
          //      if (gwc==0) gwc=0.0001;
          //      sprintf(tempstring,
          //              "k %5d %5d  j %5d %5d  i %5d %5d  vol %6d wc %10.3f  r %7.0f\n",
          //              thisBox.smallEnd(2), thisBox.bigEnd(2),
          //              thisBox.smallEnd(1), thisBox.bigEnd(1),
          //              thisBox.smallEnd(0), thisBox.bigEnd(0),
          //              vol, gwc, vol/gwc);
          //      pout() << tempstring;

        } // end loop through grids
      TF_GSRBLEVELLAP.stop();
      TPlevelGSRB_LAPloop.stop(); //calledTooMuch
      //deltaTime_gsrb = TPlevelGSRB_LAPloop.wc_time() - previousTime_gsrb;
      //previousTime_gsrb = TPlevelGSRB_LAPloop.wc_time();

#ifdef TIMED_BARR
      TPlevelGSRB_B5.start();
      barrier();
      TPlevelGSRB_B5.stop();
#endif

#ifdef COUNT_PEX
      deltaTime_exchange = TPlevelGSRB_ex.wc_time() - previousTime_exchange;
      previousTime_exchange = TPlevelGSRB_ex.wc_time();

      int index = TPlevelGSRB_ex.Count();
      pex_delta_time[index] = deltaTime_exchange;
      pex_Nlocal[index] = m_exchangeCopier.numLocalCellsToCopy();
      pex_Nto[index] = m_exchangeCopier.numToCellsToCopy();
      pex_Nfrom[index] = m_exchangeCopier.numFromCellsToCopy();

      //pex_GSRBvol[index] = volTot;
      pout() << " Local=" << m_exchangeCopier.numLocalCellsToCopy()
             << " From=" << m_exchangeCopier.numFromCellsToCopy()
             << " To=" << m_exchangeCopier.numToCellsToCopy()
             << " wc=" << deltaTime_exchange
             << endl;
#endif

      //      sprintf(tempstring, " rb%d cfInterp %8.3f  ex %8.3f  apply %8.3f  gsrb %8.3f  tot %8.3f\n",
      //      whichPass, deltaTime_CFInterp*1000.0, deltaTime_exchange*1000.0,
      //      deltaTime_apply*1000.0, deltaTime_gsrb*1000.0,
      //      (deltaTime_CFInterp+deltaTime_apply+deltaTime_exchange+deltaTime_gsrb)*1000.);
      //pout() << tempstring;

      //       sprintf(tempstring, " plot ex %8.3f  gsrb %8.3f\n",
      //               deltaTime_exchange*1000.0,
      //               deltaTime_gsrb*1000.0);
      //       pout() << tempstring;

    } // end loop through red-black
  TPlevelGSRB.stop();
}
/***************/
/***************/
void PoissonOp::levelGSMC(LevelData<FArrayBox>&       a_phi,
                          const LevelData<FArrayBox>& a_rhs)
{
  int comp = 0;
  int ncomp = 1;
  LevelData<FArrayBox> residLD(m_grids, ncomp, IntVect::Zero);
  LevelData<FArrayBox> lofphi(m_grids, ncomp, IntVect::Zero);

  IntVect color = IntVect::TheZeroVector();
  IntVect limit = IntVect::TheUnitVector();

  Real weight = 0.0;

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      weight += -2.0 / (m_dxLevel*m_dxLevel);
    }
  weight = 1.0/weight;

  // Loop over all possibilities (in all dimensions)
  while (color[SpaceDim-1] <= limit[SpaceDim-1])
    {
      // Get the residual everywhere
      applyOpH(a_phi, lofphi);
      for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
        {
          residLD[dit()].copy(a_rhs[dit()]);
          residLD[dit()] -= lofphi[dit()];
        }

      int ibox = 0;
      for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
        {
          FArrayBox& phi = a_phi[dit()];
          const Box& box = m_grids.get(dit());

          FArrayBox phiSave(box, ncomp);
          phiSave.copy(phi);

          const FArrayBox& resid = residLD[dit()];

          IntVect loIV = box.smallEnd();
          IntVect hiIV = box.bigEnd();

          for (int idir = 0; idir < SpaceDim; idir++)
            {
              if (loIV[idir] % 2 != color[idir])
                {
                  loIV[idir]++;
                }
            }

          if (loIV <= hiIV)
            {
              Box coloredBox(loIV,hiIV);

              FORT_GSMULTICOLOR(CHF_FRA1(phi,comp),
                                CHF_CONST_FRA1(phiSave,comp),
                                CHF_CONST_REAL(weight),
                                CHF_CONST_FRA1(resid,comp),
                                CHF_BOX(coloredBox));
            }

          ibox++;
        }

      color[0]++;

      // If greater that the maximum it can be...
      if (color[0] > limit[0])
        {
          // For all dimensions...
          for (int i = 0; i < SpaceDim-1; i++)
            {
              color[i] = 0;
              color[i+1]++;

              // If the next dimension's power is okay go on
              if (color[i+1] <= limit[i+1])
                {
                  break;
                }
            }
        }
    }
}

/***********************/
// evaluate operator, inhomogeneous C/F boundary conditions
// note that operator applies C/F interpolation and copy bc's
/***********************/
void PoissonOp::applyOpI(LevelData<FArrayBox>&       a_phi,
                         const LevelData<FArrayBox>* a_phicPtr,
                         LevelData<FArrayBox>&       a_lofPhi)
{
  assert(isDefined());
  assert(m_ihcfiEnabled);
  assert(a_phi.nComp() == a_lofPhi.nComp());
  assert(a_phi.nComp() == m_ncomp);

  if (a_phicPtr != NULL)
    {
      // apply C/F boundary conditions...
      CFInterp(a_phi,*a_phicPtr);
    }

  //fill in intersection of ghostcells and phi's boxes
  TMPI_Poi_applyOpI_exchange.start();
  a_phi.exchange(a_phi.interval(), m_exchangeCopier);
  TMPI_Poi_applyOpI_exchange.stop();

  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_domghostbc.applyInhomogeneousBCs(a_phi[dit()],
                                         m_domain,
                                         m_dxLevel);
    }

  //fill in intersection of ghostcells and phi's boxes
  // for a 5-point operator, this should not be necessary
  //a_phi.exchange(a_phi.interval(), m_exchangeCopier);

  Real dx = m_dxLevel;

  TF_applyOpI_OPERATORLAP.start();
  for (dit.begin(); dit.ok(); ++dit)
    {
#ifndef NDEBUG
      FArrayBox& thisPhi = a_phi[dit()];
      FArrayBox& thisLPhi = a_lofPhi[dit()];
      const Box& thisBox = m_grids.get(dit());

      assert(thisPhi.box().contains(grow(thisBox,1)));
      assert(thisLPhi.box().contains(thisBox));
#endif

      FORT_OLDOPERATORLAP(CHF_FRA(a_lofPhi[dit()]),
                          CHF_CONST_FRA(a_phi[dit()]),
                          CHF_BOX(m_grids.get(dit())),
                          CHF_CONST_REAL(dx));
    }
  TF_applyOpI_OPERATORLAP.stop();
}

/***********************/
// evaluate operator, inhomogeneous C/F boundary conditions
// note that operator applies C/F interpolation and copy bc's
// homogeneous physical bcs
/***********************/
void PoissonOp::applyOpIcfHphys(LevelData<FArrayBox>&       a_phi,
                                const LevelData<FArrayBox>* a_phicPtr,
                                LevelData<FArrayBox>&       a_lofPhi)
{
  assert(isDefined());
  assert(m_ihcfiEnabled);
  assert(a_phi.nComp() == a_lofPhi.nComp());
  assert(a_phi.nComp() == m_ncomp);

  if (a_phicPtr != NULL)
    {
      // apply C/F boundary conditions...
      CFInterp(a_phi,*a_phicPtr);
    }

  // fill in intersection of ghostcells and phi's boxes
  TMPI_Poi_applyOpIcfHphys_exchange.start();
  a_phi.exchange(a_phi.interval(), m_exchangeCopier);
  TMPI_Poi_applyOpIcfHphys_exchange.stop();

  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_domghostbc.applyHomogeneousBCs(a_phi[dit()],
                                       m_domain,
                                       m_dxLevel);
    }

  // fill in intersection of ghostcells and phi's boxes
  // for a 5 point operator, this shouldn't be necessary
  // a_phi.exchange(a_phi.interval(), m_exchangeCopier);

  Real dx = m_dxLevel;

  TF_applyOpIcfHphys_OPERATORLAP.start();
  for (dit.begin(); dit.ok(); ++dit)
    {
#ifndef NDEBUG
      FArrayBox& thisPhi = a_phi[dit()];
      FArrayBox& thisLPhi = a_lofPhi[dit()];
      const Box& thisBox = m_grids.get(dit());

      assert(thisPhi.box().contains(grow(thisBox,1)));
      assert(thisLPhi.box().contains(thisBox));
#endif

      FORT_OLDOPERATORLAP(CHF_FRA(a_lofPhi[dit()]),
                          CHF_CONST_FRA(a_phi[dit()]),
                          CHF_BOX(m_grids.get(dit())),
                          CHF_CONST_REAL(dx));
    }
  TF_applyOpIcfHphys_OPERATORLAP.stop();
}

/***********************/
// evaluate operator, inhomogeneous C/F boundary conditions
// note that operator applies C/F interpolation and copy bc's
// homogeneous physical bcs
/***********************/
void PoissonOp::applyOpH(LevelData<FArrayBox>& a_phi,
                         LevelData<FArrayBox>& a_lofPhi)
{
  assert(isDefined());
  assert (a_phi.nComp() == a_lofPhi.nComp());
  //assert (a_phi.nComp() == m_ncomp);

  //timercomment//TCFI_Poi_applyOpH.start();
  homogeneousCFInterp(a_phi);
  //timercomment//TCFI_Poi_applyOpH.stop();

  //if (NbottomSmooth==100) {
  //  NoelsPrettyPrint(&a_phi, "phi applyOpH");
  //}

  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_domghostbc.applyHomogeneousBCs(a_phi[dit()],
                                       m_domain,
                                       m_dxLevel);
    }

  // fill in intersection of ghostcells and phi's boxes
  TMPI_Poi_applyOpH_exchange.start();
  a_phi.exchange(a_phi.interval(), m_exchangeCopier);
  TMPI_Poi_applyOpH_exchange.stop();

  //if (NbottomSmooth==BS_STOP) {
  //NoelsPrettyPrint(&a_phi, "phi applyOpH");
  //}

  Real dx = m_dxLevel;

  //int NN=0;
  TF_applyOpH_OPERATORLAP.start();
  for (dit.begin(); dit.ok(); ++dit)
    {

#ifndef NDEBUG
      FArrayBox& thisPhi = a_phi[dit()];
      FArrayBox& thisLPhi = a_lofPhi[dit()];
      const Box& thisBox = m_grids.get(dit());

      assert(thisPhi.box().contains(grow(thisBox,1)));
      assert(thisLPhi.box().contains(thisBox));
#endif

      //if (NbottomSmooth==BS_STOP) {
      //cout << endl << "NN" << NN++ << endl;
      //NoelsPrettyPrint(&(a_lofPhi[dit()]));
      //cout << endl << " a_phi:" << endl;
      //NoelsPrettyPrint(&(a_phi[dit()]));
      //cout << endl << "m_grids.get()=" << m_grids.get(dit()) << endl;
      //}

      FORT_OLDOPERATORLAP(CHF_FRA(a_lofPhi[dit()]),
                          CHF_CONST_FRA(a_phi[dit()]),
                          CHF_BOX(m_grids.get(dit())),
                          CHF_CONST_REAL(dx));

      //if (NbottomSmooth==BS_STOP) {
      //cout << endl << "NN" << NN++ << endl;
      //NoelsPrettyPrint(&(a_lofPhi[dit()]));
      //cout << endl << " a_phi:" << endl;
      //NoelsPrettyPrint(&(a_phi[dit()]));
      //cout << endl << "m_grids.get()=" << m_grids.get(dit()) << endl;
      //exit(0);
      //}
    }
  TF_applyOpH_OPERATORLAP.stop();

  //if (NbottomSmooth==100) {
  //NoelsPrettyPrint(&a_lofPhi, "a_lofPhi applyOpH");
  //}
}

/***********************/
// evaluate operator, homogeneous C/F boundary conditions
// note -- grids for phi need not correspond to bx
// inhomogeneous physical boundary conditions
/***********************/
void PoissonOp::applyOpHcfIphys(LevelData<FArrayBox>& a_phi,
                                LevelData<FArrayBox>& a_lofPhi)
{
  assert(isDefined());
  assert (a_phi.nComp() == a_lofPhi.nComp());
  assert (a_phi.nComp() == m_ncomp);

  //timercomment//TCFI_Poi_applyOpHcfIphys.start();
  homogeneousCFInterp(a_phi);
  //timercomment//TCFI_Poi_applyOpHcfIphys.stop();

  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_domghostbc.applyInhomogeneousBCs(a_phi[dit()],
                                         m_domain,
                                         m_dxLevel);
    }

  // fill in intersection of ghostcells and phi's boxes
  TMPI_Poi_applyOpHcfIphys_exchange.start();
  a_phi.exchange(a_phi.interval(), m_exchangeCopier);
  TMPI_Poi_applyOpHcfIphys_exchange.stop();

  Real dx = m_dxLevel;

  TF_applyOpHcfIphys_OPERATORLAP.start();
  for (dit.begin(); dit.ok(); ++dit)
    {

#ifndef NDEBUG
      FArrayBox& thisPhi = a_phi[dit()];
      FArrayBox& thisLPhi = a_lofPhi[dit()];
      const Box& thisBox = m_grids.get(dit());

      assert(thisPhi.box().contains(grow(thisBox,1)));
      assert(thisLPhi.box().contains(thisBox));
#endif

      FORT_OLDOPERATORLAP(CHF_FRA(a_lofPhi[dit()]),
                          CHF_CONST_FRA(a_phi[dit()]),
                          CHF_BOX(m_grids.get(dit())),
                          CHF_CONST_REAL(dx));
    }
  TF_applyOpHcfIphys_OPERATORLAP.stop();
}

/***********************/
// This calls whatever has been set as the bottom solver.
/***********************/
void PoissonOp::bottomSmoother(LevelData<FArrayBox>&       a_phi,
                               const LevelData<FArrayBox>& a_rhs)
{
  assert(isDefined());
  assert(m_bottomSmootherPtr != NULL);
  NbottomSmooth++;

  BIterationFlavor = COUNTER_bottomSmoother_Poisson + GlobalLevel;

  int Npoints=0, Nboxes=0, Nghosts=0;
  NoelsLDFABgetUsefulInfo(a_rhs, Nboxes, Npoints, Nghosts);

  //char tempy[150];
  //sprintf(tempy,
  //    "  PoissonOp::bottomSmoother%04d L%d Nboxes=%4d Npoints=%7d  dxCoarse=%10.6f dxLevel=%10.6f\n",
  //    NbottomSmooth, GlobalLevel, Nboxes, Npoints, m_dxCrse, m_dxLevel);
  //pout() << tempy;

  //pout() << "  PoissonOp::bottomSmoother" << NbottomSmooth << endl;

  //#define WRITE_BOTTOMSMOOTH_FILE
#ifdef WRITE_BOTTOMSMOOTH_FILE
  if (write_bottomsmoother_file == 1) {
    pout() << " writing BOTTOMSMOOTH_FILE NbottomSmooth=" << NbottomSmooth << endl;
    //void WriteAMRHierarchyHDF5(    //This is blocking
    //     const string& filename,
    //     const Vector<DisjointBoxLayout>& a_vectGrids, //grids at each level
    //     const Vector<LevelData<FArrayBox>* > & a_vectData, //data at each level
    //     const Vector<string>& a_vectNames,       //names of variables
    //     const Box& a_domain,                   //domain at coarsest level
    //     const Real& a_dx,     //grid spacing at coarsest level
    //     const Real& a_dt,    //time step at coarsest level
    //     const Real& a_time,
    //     const Vector<int>& a_vectRatio,    //refinement ratio at all levels
    //     const int& a_numLevels);//number of levels to output

    if (NbottomSmooth>1 ) {
      int numComp = 2;  // for phi and rhs
      Vector<LevelData<FArrayBox>*> debugData(2, NULL);

      Vector<DisjointBoxLayout> vectGrids(1);
      vectGrids[0] = a_rhs.getBoxes(); // same as a_phi

      debugData[0] = new LevelData<FArrayBox>(a_rhs.getBoxes(), numComp);

      Interval rhsInterval(0,0);
      a_rhs.copyTo(a_rhs.interval(), *debugData[0], rhsInterval);

      Interval phiInterval(1,1);
      a_phi.copyTo(a_phi.interval(), *debugData[0], phiInterval);

      Box domainBox = m_domain.domainBox();

      Vector<string> dataNames(numComp);
      dataNames[0] = "rhs";
      dataNames[1] = "phi";

      Vector<int> vectRefRatio(2,1);  // what's 2,1 for?
      vectRefRatio[0] = m_refRatio;

      char filename[45];
      sprintf(filename, "bs%dd-%03d.hdf5", CH_SPACEDIM, NbottomSmooth);

      //char filename2[45];
      //sprintf(filename2, "bsl%03d.hdf5", NbottomSmooth);

      //writeLevelname(&a_rhs, filename2);
      WriteAMRHierarchyHDF5(filename,
                            vectGrids,       // vector of 1
                            debugData,       //
                            dataNames,       //
                            domainBox,
                            m_dxLevel,       // dt, hack it with dxLevel
                            m_dxCrse,
                            0.0,             // newTime, doesn't matter
                            vectRefRatio,    //
                            1);              // one level

      //cout << " m_dxLevel=" << m_dxLevel << endl;
      //cout << " m_dxCrse=" << m_dxCrse << endl;
      //cout << " domainBox=" << domainBox << endl;

      //NoelsPrettyPrint(&a_rhs, "rhs");
      //NoelsPrettyPrint(&a_phi, "phi");
    }
  }
#endif

  m_bottomSmootherPtr->doBottomSmooth(a_phi, a_rhs, this);

  //if (NbottomSmooth==100) {
  //NoelsPrettyPrint(&a_rhs, "rhs");
  //NoelsPrettyPrint(&a_phi, "phi");
  //}

  //exit(0);

  // #undef OPT_BVS_BOTTOMSMOOTH_REDIST //not helping
  // #ifdef OPT_BVS_BOTTOMSMOOTH_REDIST
  //   //  if(numProc() > 4){
  //    if(m_bottomOp_ptr == NULL){
  //      m_bottomOp_ptr = new PoissonOp();
  //      DisjointBoxLayout gatheredDBL;
  //      gatheredDBL.deepCopy(m_grids);
  //      for(LayoutIterator it=gatheredDBL.layoutIterator(); it.ok(); ++it)
  //        gatheredDBL.setProcID(it(), 0);
  //      gatheredDBL.close();
  //      m_bottomOp_ptr->define(gatheredDBL, &m_baseBA,
  //                                  m_dxLevel, m_refRatio, m_domain,
  //                                  true, m_ncomp);
  //           m_bottomOp_ptr->m_dxCrse = m_dxCrse;
  //      m_bottomOp_ptr->setDomainGhostBC(m_domghostbc);
  //    }
  //    LevelData<FArrayBox> phi(m_bottomOp_ptr->m_grids, m_ncomp, a_phi.ghostVect());
  //    LevelData<FArrayBox> rhs(m_bottomOp_ptr->m_grids, m_ncomp, a_rhs.ghostVect());
  //         DataIterator newLevelDit = phi.dataIterator();
  //         for (newLevelDit.begin(); newLevelDit.ok(); ++newLevelDit)
  //           {
  //             phi[newLevelDit()].setVal(0.0);
  //           }
  //    Interval comps(0, m_ncomp-1);
  //    a_phi.copyTo(comps, phi, comps);
  //    a_rhs.copyTo(comps, rhs, comps);
  //    m_bottomSmootherPtr->doBottomSmooth(phi, rhs, m_bottomOp_ptr);
  //    phi.copyTo(comps, a_phi, comps);
  // //   }else
  // //     {
  // //       m_bottomSmootherPtr->doBottomSmooth(a_phi, a_rhs, this);
  // //     }
  // #else
  //   m_bottomSmootherPtr->doBottomSmooth(a_phi, a_rhs, this);
  // #endif
}

/**
   get flux( == flux at THIS level)
   The fluxes live on the cell faces with direction dir.
   Fluxes are computed for all interior edges of data.
   The flux fab is resized inside the routine.
*/
void PoissonOp::getFlux(FArrayBox&       a_fineFlux,
                        const FArrayBox& a_data,
                        const DataIndex& a_datInd,
                        int              a_dir)
{
  assert(isDefined());
  assert(a_dir >= 0);
  assert(a_dir <  SpaceDim);
  assert(!a_data.box().isEmpty());

  Box edgebox = surroundingNodes(a_data.box(),a_dir);
  edgebox.grow(a_dir, -1);

  // if this fails, the data box was too small (one cell wide, in fact)
  assert(!edgebox.isEmpty());

  a_fineFlux.resize(edgebox, a_data.nComp());
  BoxIterator bit(edgebox);

  IntVect iv, shiftiv, ivlo, ivhi;
  Real phihi, philo, gradphi;
  Real im_dxLevel = 1.0/m_dxLevel;

  for( bit.begin(); bit.ok(); bit.next())
    {
      iv = bit();
      shiftiv = BASISV(a_dir);
      ivlo = iv - shiftiv;
      ivhi = iv;

      assert(a_data.box().contains(ivlo));
      assert(a_data.box().contains(ivhi));

      for (int ivar = 0; ivar < a_data.nComp(); ivar++)
        {
          phihi = a_data(ivhi,ivar);
          philo = a_data(ivlo,ivar);
          //Real gradphi =(phihi-philo)/m_dxLevel;
          gradphi =(phihi-philo) * im_dxLevel;

          a_fineFlux(iv,ivar) = gradphi;
        }
    }
}

void
PoissonOp::setConvergenceMetric(Real a_metric, int a_comp)
{
  m_bottomSmootherPtr->setConvergenceMetric(a_metric, a_comp);
}
