/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _LOHICENTER_H_
#define _LOHICENTER_H_

#include "Box.H"
#include "ProblemDomain.H"

///
/**
   This function is used when in direction a_dir a 3 point stencil of cell-
   centered data is being used to compute something at the cell center of the
   center cell of the stencil.  The data for the stencil is valid in
   a_inBox.  It divides a_inBox into a cell-centered box where the full
   stencil can be used, a_centerBox, and cell-centered boxes where 2 point
   stencils must be used, a_loBox and a_hiBox based on the current problem
   domain, a_domain, and the stencil direction, a_dir.  The union of these 2
   and 3 point stencil boxes is returned as a_entireBox.  The 2 point stencil
   boxes are one wide, at most, and if they have been defined then the
   corresponding flag, a_hasLo or a_hasHi, is set to one, otherwise these
   flags are zero.  All output boxes lie within the domain.
 */
extern void loHiCenter(Box&                 a_loBox,
                       int&                 a_hasLo,
                       Box&                 a_hiBox,
                       int&                 a_hasHi,
                       Box&                 a_centerBox,
                       Box&                 a_entireBox,
                       const Box&           a_inBox,
                       const ProblemDomain& a_domain,
                       const int&           a_dir);

///
/**
   This function is used when in direction a_dir a 2 point stencil of cell-
   centered data is being used to compute something at the cell face between
   the cell centers of the stencil.  The data for the stencil is valid in
   a_inBox.  It uses a_inBox to compute a box (face-centered in a_dir) where
   the full stencil can be used, a_centerBox, and boxes (face-centered in
   a_dir) where a 1 point stencil must be used, a_loBox and a_hiBox based on
   the current problem domain, a_domain, and the stencil direction, a_dir.
   The union of these 1 and 2 point stencil boxes is returned as a_entireBox
   (face-centered in a_dir).  The 1 point stencil boxes are one wide, at most,
   and if they have been defined then the corresponding flag, a_hasLo or
   a_hasHi, is set to one, otherwise these flags are zero.  All output boxes
   lie within the domain.
 */
extern void loHiCenterFace(Box&                 a_loBox,
                           int&                 a_hasLo,
                           Box&                 a_hiBox,
                           int&                 a_hasHi,
                           Box&                 a_centerBox,
                           Box&                 a_entireBox,
                           const Box&           a_inBox,
                           const ProblemDomain& a_domain,
                           const int&           a_dir);

#endif
