/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "QuadCFInterp.H"
#include "LayoutIterator.H"
#include "DataIterator.H"

#include "Timer.H"
extern Timer Everything;
Timer TMPI_tempQCFIBarrier     ("temp QCFI Barr",   Everything, DTIMECOMM, true);
Timer TMPI_tempQCFIBarrierL1   ("temp QCFI BarrL1", Everything, DTIMECOMM, true);
Timer TMPI_tempQCFIBarrierL2   ("temp QCFI BarrL2", Everything, DTIMECOMM, true);

Timer TMPI_coarseFineInterp_cp1  ("cfInterp1 cp",   Everything, DTIMECOMM, true);
Timer TMPI_coarseFineInterp_cp1L1("cfInterp1 cpL1", Everything, DTIMECOMM, true);
Timer TMPI_coarseFineInterp_cp1L2("cfInterp1 cpL2", Everything, DTIMECOMM, true);

Timer TMPI_coarseFineInterp_cp2  ("cfInterp2 cp",   Everything, DTIMECOMM, true);
Timer TMPI_coarseFineInterp_cp2L1("cfInterp2 cpL1", Everything, DTIMECOMM, true);
Timer TMPI_coarseFineInterp_cp2L2("cfInterp2 cpL2", Everything, DTIMECOMM, true);

Timer TQCFInterp_define("QCFInterp define", Everything, DTIMEDEFINES, true);

Timer TQCFI_all("QCFI cfi all", 178);
Timer TQCFI_Barrier("QCFI cfi Barr", TQCFI_all);
Timer TQCFI_cp1("QCFI cfi cp1", TQCFI_all);
Timer TQCFI_cp2("QCFI cfi cp2", TQCFI_all);
Timer TQCFI_work("QCFI cfi work", TQCFI_all);

extern int GlobalLevel;

//Timer q1("q1", TQCFInterp_define);
//Timer q2("q2", TQCFInterp_define);
//Timer q3("q3", TQCFInterp_define);
//Timer q4("q4", TQCFInterp_define);

#include <cstdio>
#include "NoelsPrettyPrint.H"

//Timer TimeQuadCFinterpOnIVS("QuadCF interp", Everything, 8);

/***********************/
// default constructor
/***********************/
void
QuadCFInterp::clear()
{
  m_isDefined = false;
  m_level = -1;
  m_dxFine = -1;
}

QuadCFInterp::QuadCFInterp()
{
  clear();
}

/***********************/
/***********************/
QuadCFInterp::QuadCFInterp(
                           const DisjointBoxLayout& a_fineBoxes,
                           const DisjointBoxLayout* a_coarBoxes,
                           Real  a_dxFine,
                           int a_refRatio, int a_nComp,
                           const Box& a_domf)
{
  ProblemDomain fineProbDomain(a_domf);
  define(a_fineBoxes,a_coarBoxes, a_dxFine,a_refRatio,a_nComp, fineProbDomain);
}

/***********************/
/***********************/
QuadCFInterp::QuadCFInterp(
                           const DisjointBoxLayout& a_fineBoxes,
                           const DisjointBoxLayout* a_coarBoxes,
                           Real  a_dxFine,
                           int a_refRatio, int a_nComp,
                           const ProblemDomain& a_domf)
{
  define(a_fineBoxes,a_coarBoxes, a_dxFine,a_refRatio,a_nComp, a_domf);
}

/***********************/
/***********************/

void
QuadCFInterp::define(
                     const DisjointBoxLayout& a_fineBoxes,
                     const DisjointBoxLayout* a_coarBoxesPtr,
                     Real  a_dxLevel,
                     int a_refRatio, int a_nComp,
                     const ProblemDomain& a_domf)
{
  TQCFInterp_define.start();
  clear();
  m_isDefined = true;
  assert(a_nComp > 0);

  assert (!a_domf.isEmpty());
  // consistency check
  assert (a_fineBoxes.checkPeriodic(a_domf));

  m_domainFine = a_domf;
  m_dxFine = a_dxLevel;
  m_refRatio = a_refRatio;
  m_nComp = a_nComp;
  m_inputFineLayout = a_fineBoxes;
  if(a_coarBoxesPtr == NULL)
    m_level = 0;
  else
    m_level = 1;

  if(m_level > 0)
    {
      // (DFM) only check for valid refRatio if a coarser level exists
      assert(a_refRatio >= 1);
      const DisjointBoxLayout& coarBoxes = *a_coarBoxesPtr;
      m_inputCoarLayout = coarBoxes;
      assert (coarBoxes.checkPeriodic(coarsen(a_domf,a_refRatio)));

      //q1.start();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          m_loQCFS[idir].define(a_fineBoxes);
          m_hiQCFS[idir].define(a_fineBoxes);
          m_loCoarBoxes[idir].deepCopy(a_fineBoxes);
          m_hiCoarBoxes[idir].deepCopy(a_fineBoxes);
        }
      //q1.stop();

      //locoarboxes and hicoarboxes are now open
      //and have same processor mapping as a_fineboxes
      //make boxes for coarse buffers
      //q2.start();
      LayoutIterator lit = a_fineBoxes.layoutIterator();
      for(lit.begin(); lit.ok(); ++lit)
        {
          const Box& fineBox = a_fineBoxes[lit()];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Box coarBox   = coarsen(fineBox, m_refRatio);
              Box coarBoxHi = adjCellHi(coarBox,idir,1);
              Box coarBoxLo = adjCellLo(coarBox,idir,1);
              //give em a ghost cell on non-idir sides
              coarBoxLo.grow(1);
              coarBoxLo.grow(idir, -1);
              coarBoxHi.grow(1);
              coarBoxHi.grow(idir, -1);

              m_loCoarBoxes[idir][lit()] = coarBoxLo;
              m_hiCoarBoxes[idir][lit()] = coarBoxHi;
            }
        }
      //q2.stop();

      //q3.start();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          m_loCoarBoxes[idir].close();
          m_hiCoarBoxes[idir].close();
          m_loCoarBuffer[idir].define(m_loCoarBoxes[idir],m_nComp);
          m_loCoarCopiers[idir].define(coarBoxes, m_loCoarBoxes[idir]);
          m_hiCoarBuffer[idir].define(m_hiCoarBoxes[idir],m_nComp);
          m_hiCoarCopiers[idir].define(coarBoxes, m_hiCoarBoxes[idir]);
        }
      //q3.stop();

      //make cfstencils and boxes for coarse buffers
      //q4.start();
      DataIterator dit = a_fineBoxes.dataIterator();
      for(dit.begin(); dit.ok(); ++dit)
        {
          const Box& fineBox = a_fineBoxes[dit()];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              //low side cfstencil
              m_loQCFS[idir][dit()].define(a_domf,
                                           fineBox,
                                           a_fineBoxes,
                                           coarBoxes,
                                           a_refRatio,
                                           idir,
                                           Side::Lo);
              //high side cfstencil
              m_hiQCFS[idir][dit()].define(a_domf,
                                           fineBox,
                                           a_fineBoxes,
                                           coarBoxes,
                                           a_refRatio,
                                           idir,
                                           Side::Hi);

            }
        }
      //q4.stop();
    }
  TQCFInterp_define.stop();
}

/***********************/
//  apply coarse-fine boundary conditions -- assume that phi grids
//  are grown by one
/***********************/
void
QuadCFInterp::coarseFineInterp(BaseFab<Real> & a_phif,
                               const BaseFab<Real> & a_phic,
                               const QuadCFStencil& a_qcfs,
                               const Side::LoHiSide a_hiorlo,
                               const int a_idir,
                               const Interval& a_variables) const
{
  assert(isDefined());
  //nothing happens if m_level == 0
  if(m_level > 0)
    {
      if(!a_qcfs.isEmpty())
        {
          BaseFab<Real>  phistar;

          //first find extended value phistar
          //includes finding slopes of coarse solution bar
          getPhiStar(phistar, a_phic, a_qcfs, a_hiorlo, a_idir, a_variables);

          //given phistar, interpolate on fine ivs
          //TimeQuadCFinterpOnIVS.start();
          interpOnIVS(a_phif, phistar, a_qcfs, a_hiorlo, a_idir, a_variables);
          //TimeQuadCFinterpOnIVS.stop();
        }
    }
}

/***********************/
//get extended phi (lives next to interpivs)
/***********************/
void
QuadCFInterp::getPhiStar(BaseFab<Real> & a_phistar,
                         const BaseFab<Real> & a_phic,
                         const QuadCFStencil& a_qcfs,
                         const Side::LoHiSide a_hiorlo,
                         const int a_idir,
                         const Interval& a_variables) const
{

  assert(isDefined());
  assert(a_qcfs.isDefined());
  Real dxf = m_dxFine;
  Real dxc = m_refRatio*dxf;

  // if we think of a_idir as the "me" direction, then
  // the other directions can be "you1" and "you2"
#if(CH_SPACEDIM == 3)
  int you1, you2;
  if (a_idir == 0)
    {
      you1 = 1;
      you2 = 2;
    }
  else if (a_idir == 1)
    {
      you1 = 0;
      you2 = 2;
    }
  else
    {
      you1 = 0;
      you2 = 1;
    }
#else // (CH_SPACEDIM == 2)
  int you1;
  if (a_idir == 0)
    {
      you1 = 1;
    }
  else
    {
      you1 = 0;
    }
#endif

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty())
    {

      assert(m_level > 0);
      const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
      const IntVectSet& coarsl_ivs = a_qcfs.getCoarIVS();
      Box coarinterpbox = coarsl_ivs.minBox();
      int ncomp = a_phic.nComp();
      assert(ncomp == m_nComp);
      assert(a_phic.box().contains((coarinterpbox)));

      // allocate phistar here
      int ihilo = sign(a_hiorlo);
      Box phistarbox = interp_ivs.minBox();
      phistarbox.shift(a_idir, ihilo);
      a_phistar.resize(phistarbox, ncomp);

      for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++)
        {

          //phi = phino + slope*x + half*x*x*curvature
          BaseFab<Real>  coarslope(coarinterpbox, SpaceDim);
          BaseFab<Real>  coarcurva(coarinterpbox, SpaceDim);
#if(CH_SPACEDIM == 3)
          BaseFab<Real>  coarmixed(coarinterpbox, 1);
#endif
          coarslope.setVal(0.);

          //first find extended value phistar. get slopes of coarse solution
          IVSIterator coar_ivsit(coarsl_ivs);
          for(coar_ivsit.begin(); coar_ivsit.ok(); ++coar_ivsit)
            {
              IntVect coariv = coar_ivsit();

              coarslope(coariv, a_idir) = 0.0;
              coarcurva(coariv, a_idir) = 0.0;

              coarslope(coariv, you1) =
                a_qcfs.computeFirstDerivative (a_phic, you1, ivar, coariv, dxc);
              coarcurva(coariv, you1) =
                a_qcfs.computeSecondDerivative(a_phic, you1, ivar, coariv, dxc);

#if(CH_SPACEDIM == 3)
              coarslope(coariv, you2) =
                a_qcfs.computeFirstDerivative (a_phic, you2, ivar, coariv, dxc);
              coarcurva(coariv, you2) =
                a_qcfs.computeSecondDerivative(a_phic, you2, ivar, coariv, dxc);

              coarmixed(coariv) =
                a_qcfs.computeMixedDerivative(a_phic, ivar, coariv, dxc);
#endif
            } //end loop over coarse intvects

          IntVect ivf, ivc, ivstar;
          int jf, jc;
          Real xf, xc, x1;
          IVSIterator fine_ivsit(interp_ivs);
          for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
            {
              ivf = fine_ivsit();
              ivc = coarsen(ivf, m_refRatio);
              ivstar = ivf;
              ivstar.shift(a_idir, ihilo);
              a_phistar(ivstar, ivar)  = a_phic(ivc,ivar);

              jf = ivf[you1];
              jc = ivc[you1];
              xf = (jf+0.5)*dxf;
              xc = (jc+0.5)*dxc;
              x1 = xf-xc;
              a_phistar(ivstar, ivar) +=
                x1*coarslope(ivc, you1) + 0.5*x1*x1*coarcurva(ivc, you1);

#if(CH_SPACEDIM==3)
              Real x2;
              jf = ivf[you2];
              jc = ivc[you2];
              xf = (jf+0.5)*dxf;
              xc = (jc+0.5)*dxc;
              x2 = xf-xc;
              a_phistar(ivstar, ivar) +=
                x2*coarslope(ivc, you2) + 0.5*x2*x2*coarcurva(ivc, you2);

              //add in mixed derivative component
              a_phistar(ivstar, ivar) += x1*x2*coarmixed(ivc);
#endif
            } //end loop over fine intvects
        } //end loop over variables
    } //end if(level>0 && !hocfs.isempty())
} //end function getphistar

/***********************/
/***********************/
bool
QuadCFInterp::isDefined() const
{
  return m_isDefined;
}

void
QuadCFInterp::interpOnIVS(BaseFab<Real> & a_phif,
                          const BaseFab<Real> & a_phistar,
                          const QuadCFStencil& a_qcfs,
                          const Side::LoHiSide a_hiorlo,
                          const int a_idir,
                          const Interval& a_variables) const
{
  assert(isDefined());
  assert(a_qcfs.isDefined());

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty())
    {
      //if there IS something to interpolate, the level ident
      //had better be greater than zero.  Otherwise a null
      //was sent in as coarse grids on construction
      assert(m_level > 0);
      const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
      IVSIterator fine_ivsit(interp_ivs);
      assert(a_phistar.nComp() == a_phif.nComp());
      assert(a_phistar.nComp() == m_nComp);

      int nref = m_refRatio;
      int ihilo = sign(a_hiorlo);
      for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
        {
          IntVect ivf = fine_ivsit();
          // quadratic interpolation
          for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++)
            {
              Real pa =    a_phif(ivf -2*ihilo*BASISV(a_idir), ivar);
              Real pb =    a_phif(ivf -  ihilo*BASISV(a_idir), ivar);
              Real ps = a_phistar(ivf +  ihilo*BASISV(a_idir), ivar);
              //phi = ax**2 + bx + c, x = 0 at pa
              Real h = m_dxFine;
              Real a = (2./h/h)*(2.*ps + pa*(nref+1.0) -pb*(nref+3.0))/
                (nref*nref + 4*nref + 3.0);
              Real b = (pb-pa)/h - a*h;
              Real c = pa;
              Real x = 2.*h;
              a_phif(ivf,ivar) = a*x*x + b*x + c;
            } //end loop over components
        } //end loop over fine intvects
    } //end if(level>0 && !oscfs.isempty())
} //end function interponivs

/***********************/
//  apply coarse-fine boundary conditions -- assume that phi grids
//  are grown by one
/***********************/
void
QuadCFInterp::coarseFineInterp(LevelData<FArrayBox>& a_phif,
                               const LevelData<FArrayBox>& a_phic)
{
  assert(isDefined());
  assert(a_phif.nComp() == m_nComp);
  assert(a_phic.nComp() == m_nComp);
  assert(a_phif.ghostVect() >= IntVect::TheUnitVector());
  assert(a_phic.boxLayout() == m_inputCoarLayout);
  assert(a_phif.boxLayout() == m_inputFineLayout);

  TQCFI_all.start();

  Interval variables = a_phic.interval();

#ifdef TIMED_BARR
  //Timer tb;
  //tb.start();
  TQCFI_preBarrier.start();
  if (GlobalLevel == 1)      TMPI_tempQCFIBarrierL1.start();
  else if (GlobalLevel == 2) TMPI_tempQCFIBarrierL2.start();
  else                       TMPI_tempQCFIBarrier.start();
  barrier();
  if (GlobalLevel == 1)      TMPI_tempQCFIBarrierL1.stop();
  else if (GlobalLevel == 2) TMPI_tempQCFIBarrierL2.stop();
  else                       TMPI_tempQCFIBarrier.stop();
  TQCFI_preBarrier.stop();
  //tb.stop();

  //int Nboxes, Npoints, Nghosts;
  //NoelsLDFABgetUsefulInfo(a_phif, Nboxes, Npoints, Nghosts);
  //char temp[234];
  //sprintf(temp, " QuadCFInterp::coarseFineInterp lev%d gl=%d  fineboxes=%4d fp=%8d barr=%5.3f\n",
  //      m_level, GlobalLevel,  Nboxes, Npoints, tb.wc_time());
  //pout() << temp;
#endif

  ////Timer a, b, c;
  if(m_level > 0)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          //put coarse data into buffers
          ////a.start();
          TQCFI_cp1.start();
          if (GlobalLevel == 1)      TMPI_coarseFineInterp_cp1L1.start();
          else if (GlobalLevel == 2) TMPI_coarseFineInterp_cp1L2.start();
          else                       TMPI_coarseFineInterp_cp1.start();
          a_phic.copyTo(a_phic.interval(),
                        m_loCoarBuffer[idir],
                        m_loCoarBuffer[idir].interval(),
                        m_loCoarCopiers[idir]);
          if (GlobalLevel == 1)      TMPI_coarseFineInterp_cp1L1.stop();
          else if (GlobalLevel == 2) TMPI_coarseFineInterp_cp1L2.stop();
          else                       TMPI_coarseFineInterp_cp1.stop();
          TQCFI_cp1.stop(); TQCFI_cp2.start();
          ////a.stop(); b.start();
          if (GlobalLevel == 1)      TMPI_coarseFineInterp_cp2L1.start();
          else if (GlobalLevel == 2) TMPI_coarseFineInterp_cp2L2.start();
          else                       TMPI_coarseFineInterp_cp2.start();
          a_phic.copyTo(a_phic.interval(),
                        m_hiCoarBuffer[idir],
                        m_hiCoarBuffer[idir].interval(),
                        m_hiCoarCopiers[idir]);
          if (GlobalLevel == 1)      TMPI_coarseFineInterp_cp2L1.stop();
          else if (GlobalLevel == 2) TMPI_coarseFineInterp_cp2L2.stop();
          else                       TMPI_coarseFineInterp_cp2.stop();
          TQCFI_cp2.stop();  TQCFI_work.start();
          ////b.stop(); c.start();
          DataIterator ditFine = a_phif.dataIterator();
          for(ditFine.begin(); ditFine.ok(); ++ditFine)
            {
              DataIndex datIndGlo =ditFine();
              BaseFab<Real> & phif = a_phif[datIndGlo];
              //lo side cfinterp
              //recall that buffers have fine processor mapping
              {
                const BaseFab<Real> & loPhiC = m_loCoarBuffer[idir][datIndGlo];
                const QuadCFStencil& loQCFS = m_loQCFS[idir][datIndGlo];
                coarseFineInterp(phif,loPhiC,loQCFS, Side::Lo, idir, variables);
              }
              //hi side cfinterp
              {
                const BaseFab<Real> & hiPhiC = m_hiCoarBuffer[idir][datIndGlo];
                const QuadCFStencil& hiQCFS = m_hiQCFS[idir][datIndGlo];
                coarseFineInterp(phif,hiPhiC, hiQCFS, Side::Hi, idir, variables);
              }
            }//end iteration over boxes in fine grid
          ////c.stop();
          TQCFI_work.stop();
        } //end iteration over directions
      ////sprintf(temp, " ls0=%4d hs0=%4d a=%5.3f b=%5.3f c=%5.3f \n",
      ////    m_loCoarBuffer[0].boxLayout().size(),
      ////    m_hiCoarBuffer[0].boxLayout().size(),
      ////    a.wc_time(), b.wc_time(), c.wc_time());
      ////pout()<< temp;

    }
  TQCFI_all.stop();
}
/***********************/
/***********************/
QuadCFInterp::~QuadCFInterp()
{
  clear();
}
