/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "QuadCFInterp.H"
#include "LayoutIterator.H"
#include "DataIterator.H"

#ifdef OPT_QUAD_INTERPONIVS

#include "Timer.H"
extern Timer Everything;
Timer TQCF_getPhiStar("get Phi star", Everything, 182);
//Timer q0("q0", TQCF_getPhiStar);
//Timer q1("q1", TQCF_getPhiStar);

Timer TQCF_IVS("QCF IVS", Everything, 183);
//Timer qa("qa", TQCF_IVS);
//Timer qb("qb", TQCF_IVS);

void QuadCFInterp::mygetPhiStar(BaseFab<Real> & a_phistar,
                                const BaseFab<Real> & a_phic,
                                const QuadCFStencil& a_qcfs,
                                const Side::LoHiSide a_hiorlo,
                                const int a_idir,
                                const Interval& a_variables) const {
  TQCF_getPhiStar.start();
  assert(isDefined());
  assert(a_qcfs.isDefined());
  Real dxf = m_dxFine;
  Real dxc = m_refRatio*dxf;

  // if we think of a_idir as the "me" direction, then
  // the other two directions can be "you1" and "you2"
  int you1, you2;
  if (a_idir == 0) {
    you1 = 1;
    you2 = 2;
  } else if (a_idir == 1) {
    you1 = 0;
    you2 = 2;
  } else {
    you1 = 0;
    you2 = 1;
  }

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty()) {

    assert(m_level > 0);
    const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
    const IntVectSet& coarsl_ivs = a_qcfs.getCoarIVS();
    Box coarinterpbox = coarsl_ivs.minBox();
    int ncomp = a_phic.nComp();
    assert(ncomp == m_nComp);
    assert(a_phic.box().contains((coarinterpbox)));

    // allocate phistar here
    int ihilo = sign(a_hiorlo);
    Box phistarbox = interp_ivs.minBox();
    phistarbox.shift(a_idir, ihilo);
    a_phistar.resize(phistarbox, ncomp);

    for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++) {

      //q0.start();
      //phi = phino + slope*x + half*x*x*curvature
      BaseFab<Real>  coarslope(coarinterpbox, SpaceDim);
      BaseFab<Real>  coarcurva(coarinterpbox, SpaceDim);
      BaseFab<Real>  coarmixed(coarinterpbox, 1);

      coarslope.setVal(0.);

      //first find extended value phistar. get slopes of coarse solution
      IVSIterator coar_ivsit(coarsl_ivs);
      for(coar_ivsit.begin(); coar_ivsit.ok(); ++coar_ivsit) {
        IntVect coariv = coar_ivsit();

        coarslope(coariv, a_idir) = 0.0;
        coarcurva(coariv, a_idir) = 0.0;

        coarslope(coariv, you1) =
          a_qcfs.computeFirstDerivative (a_phic, you1, ivar, coariv, dxc);
        coarcurva(coariv, you1) =
          a_qcfs.computeSecondDerivative(a_phic, you1, ivar, coariv, dxc);

        coarslope(coariv, you2) =
          a_qcfs.computeFirstDerivative (a_phic, you2, ivar, coariv, dxc);
        coarcurva(coariv, you2) =
          a_qcfs.computeSecondDerivative(a_phic, you2, ivar, coariv, dxc);

#if(CH_SPACEDIM == 3)
        coarmixed(coariv) =
          a_qcfs.computeMixedDerivative(a_phic, ivar, coariv, dxc);
#endif
      } //end loop over coarse intvects

      //q0.stop();
      //q1.start();

      IntVect ivf, ivc, ivstar;
      int jf, jc;
      Real xf, xc, x1, x2;
      IVSIterator fine_ivsit(interp_ivs);
      for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit) {
        ivf = fine_ivsit();
        ivc = coarsen(ivf, m_refRatio);
        ivstar = ivf;
        ivstar.shift(a_idir, ihilo);
        a_phistar(ivstar, ivar)  = a_phic(ivc,ivar);

        jf = ivf[you1];
        jc = ivc[you1];
        xf = (jf+0.5)*dxf;
        xc = (jc+0.5)*dxc;
        x1 = xf-xc;
        a_phistar(ivstar, ivar) +=
          x1*coarslope(ivc, you1) + 0.5*x1*x1*coarcurva(ivc, you1);

        jf = ivf[you2];
        jc = ivc[you2];
        xf = (jf+0.5)*dxf;
        xc = (jc+0.5)*dxc;
        x2 = xf-xc;
        a_phistar(ivstar, ivar) +=
          x2*coarslope(ivc, you2) + 0.5*x2*x2*coarcurva(ivc, you2);

#if(CH_SPACEDIM==3)
        //add in mixed derivative component
        a_phistar(ivstar, ivar) += x1*x2*coarmixed(ivc);
#endif
      } //end loop over fine intvects
      //q1.stop();
    } //end loop over variables
  } //end if(level>0 && !hocfs.isempty())
  TQCF_getPhiStar.stop();
} //end function getphistar

void
QuadCFInterp::myinterpOnIVS(BaseFab<Real> & a_phif,
                            const BaseFab<Real> & a_phistar,
                            const QuadCFStencil& a_qcfs,
                            const Side::LoHiSide a_hiorlo,
                            const int a_idir,
                            const Interval& a_variables) const
{
  TQCF_IVS.start();
  assert(isDefined());
  assert(a_qcfs.isDefined());

  //const CFIVS* cfivs_ptr = NULL;
  //if(a_hiorlo == Side::Lo)
  //cfivs_ptr = &m_loCFIVS[a_idir][a_datInd];
  //else
  //cfivs_ptr = &m_hiCFIVS[a_idir][a_datInd];

  //if cfsten is empty, nothing to interpolate.
  if(!a_qcfs.isEmpty()) {
    //if there IS something to interpolate, the level ident
    //had better be greater than zero.  Otherwise a null
    //was sent in as coarse grids on construction
    assert(m_level > 0);
    const IntVectSet& interp_ivs = a_qcfs.getFineIVS();
    IVSIterator fine_ivsit(interp_ivs);
    assert(a_phistar.nComp() == a_phif.nComp());
    assert(a_phistar.nComp() == m_nComp);

    int nref = m_refRatio;
    Real h = m_dxFine;
    Real KH = 2.0/(h*h);
    Real KINVH = 1.0/h;
    Real KX = 2.0*h;
    Real K1 = nref+1.0;
    Real K3 = nref+3.0;
    Real KD = (nref*nref + 4*nref + 3.0);
    int ihilo = sign(a_hiorlo);
    Real pa, pb, ps, a, b, c, x;
    IntVect ivf;
    for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit) {
      ivf = fine_ivsit();
      // quadratic interpolation
      for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++) {
        pa =    a_phif(ivf -2*ihilo*BASISV(a_idir), ivar);
        pb =    a_phif(ivf -  ihilo*BASISV(a_idir), ivar);
        ps = a_phistar(ivf +  ihilo*BASISV(a_idir), ivar);
        //phi = ax**2 + bx + c, x = 0 at pa

        a = KH*(2.0*ps + pa*K1 - pb*K3)/KD;
        b = (pb-pa)/h - a*h;
        c = pa;
        a_phif(ivf,ivar) = a*KX*KX + b*KX + c;
      } //end loop over components
    } //end loop over fine intvects
  } //end if(level>0 && !oscfs.isempty())
  TQCF_IVS.stop();
} //end function interponivs

#endif
