/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _ARRAYVIEWDATA_H_
#define _ARRAYVIEWDATA_H_

#include "REAL.H"
#include "BaseFab.H"
#include "FArrayBox.H"
#include "LayoutData.H"

/// Special LayoutData type for use with ArrayView
/** ArrayViewData is a wrapper around LayoutData for BaseFab<Real> or
 *  FArrayBox for use in ArrayView.  It provides random access to the
 *  individual FABs and the ability to directly set the data in them.  It is
 *  intended to work only in ArrayView.  Any other attept to use this class
 *  is unsupported.
 */

class ArrayViewData
{
public:
  ArrayViewData()
    :
    m_layoutdata_ptr(NULL)
  {}

  ~ArrayViewData()
  {}

  ArrayViewData(LayoutData<BaseFab<Real> >* a_layoutdata_ptr)
    :
    m_layoutdata_ptr(a_layoutdata_ptr)
  {}

  ArrayViewData(LayoutData<FArrayBox>* a_layoutdata_ptr)
    :
    m_layoutdata_ptr((LayoutData<BaseFab<Real> >*) a_layoutdata_ptr)
  {}

  /// Access a single FAB in the LayoutData using its internal index
  BaseFab<Real>& operator[](int a_elem)
  {
    return *(m_layoutdata_ptr->m_vector[a_elem]);
  }

  /// Substitute a different FAB into the LayoutData
  /** The caller is responsible for managing the memory used by the FAB that
   * is being replaced.
   */
  void set(int            a_elem,
           BaseFab<Real>* a_fab_ptr)
  {
    m_layoutdata_ptr->m_vector[a_elem] = a_fab_ptr;
  }


private:
  LayoutData<BaseFab<Real> >* m_layoutdata_ptr;
};

#endif
