/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstring>

#include "BaseFab.H"

Real BaseFabRealSetVal = BASEFAB_REAL_SETVAL;

template < > void BaseFab<Real>::define()
{
  assert(m_nvar > 0);
  assert(m_dptr == 0);
  assert(m_numpts > 0);
  assert(!m_aliased);
  // assert(!(The_FAB_Arena == 0)); // not a sufficient test!!!

#ifdef ENABLE_MEMORY_TRACKING
  if (s_Arena == NULL)
  {
    s_Arena = new BArena(name().c_str());
  }
#else
  if (s_Arena == NULL)
  {
    s_Arena = new BArena("");
  }
#endif

  // if (s_Arena == NULL)
  // {
  //   MayDay::Error("malloc in basefab failed");
  // }

  m_truesize = m_nvar * m_numpts;
  m_dptr     = static_cast<Real*>(s_Arena->alloc(m_truesize * sizeof(Real)));

#ifdef ENABLE_MEMORY_TRACKING
  s_Arena->bytes += m_truesize * sizeof(Real) + sizeof(BaseFab<Real>);

  if (s_Arena->bytes > s_Arena->peak)
  {
    s_Arena->peak = s_Arena->bytes;
  }
#endif

#ifdef CH_USE_SETVAL
  setVal(BaseFabRealSetVal);
#endif
}

template < > void BaseFab<int>::define()
{
  assert(m_nvar > 0);
  assert(m_dptr == 0);
  assert(m_numpts > 0);
  assert(!m_aliased);
  // assert(!(The_FAB_Arena == 0)); // not a sufficient test!!!

#ifdef ENABLE_MEMORY_TRACKING
  if (s_Arena == NULL)
  {
    s_Arena = new BArena(name().c_str());
  }
#else
  if (s_Arena == NULL)
  {
    s_Arena = new BArena("");
  }
#endif

  // if(s_Arena == NULL)
  // {
  //   MayDay::Error("malloc in basefab failed");
  // }

  m_truesize = m_nvar * m_numpts;
  m_dptr     = static_cast<int*>(s_Arena->alloc(m_truesize * sizeof(int)));

#ifdef ENABLE_MEMORY_TRACKING
  s_Arena->bytes += m_truesize * sizeof(int) + sizeof(BaseFab<int>);

  if (s_Arena->bytes > s_Arena->peak)
  {
    s_Arena->peak = s_Arena->bytes;
  }
#endif
}

template < > void BaseFab<Real>::undefine()
{
  if (m_aliased)
  {
    m_dptr = 0;
    return;
  }

  if (m_dptr == 0)
  {
    return;
  }

  s_Arena->free(m_dptr);

#ifdef ENABLE_MEMORY_TRACKING
  s_Arena->bytes -= m_truesize * sizeof(Real) + sizeof(BaseFab<Real>);
#endif

  m_dptr = 0;
}

template < > void BaseFab<int>::undefine()
{
  if (m_aliased)
  {
    m_dptr = 0;
    return;
  }

  if (m_dptr == 0)
  {
    return;
  }

  s_Arena->free(m_dptr);

#ifdef ENABLE_MEMORY_TRACKING
  s_Arena->bytes -= m_truesize * sizeof(int) + sizeof(BaseFab<int>);
#endif

  m_dptr = 0;
}

template < > void BaseFab<Real>::setVal(Real a_val)
{
  if (a_val == 0)
  {
    memset(m_dptr, 0, m_truesize*sizeof(Real));
  }
  else
  {
    Real* end = m_dptr + m_truesize;
    for (Real* v = m_dptr; v<end; v++)
    {
      *v = a_val;
    }
  }
}
