/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifdef OPT_MYPERFORMCOPY

#include "BaseFab.H"
#include "BaseFabImplem.H"
//#include "Timer.H"
//extern Timer Everything;
//Timer TFAB_performSetVal  ("performSetVal",     Everything, 6);
//Timer TFAB_linearOut      ("linearOut",         Everything, 6);
//Timer TFAB_linearIn       ("linearIn",          Everything, 6);

template class BaseFab<double>;
template class BaseFab<int>;
//template class BaseFab<DerivStencil>;
template class BaseFab<bool>;

//#define NQ 1000000
//long unsigned int qiter=0;
//unsigned int qsubbox_len0[NQ];
//unsigned int qsubbox_len1[NQ];
//unsigned int qsubbox_len2[NQ];
//unsigned int qNN[NQ];

template <class T>
inline
void
BaseFab<T>::myperformCopy (const BaseFab<T>& src,
               const Box&        srcbox,
               int               srccomp,
               const Box&        destbox,
               int               destcomp,
               int               numcomp)
  // if (srcbox == destbox) -- try this
{
  if(!(( destbox ).isEmpty())) {
    const int *th_plo = loVect();
    const int *th_plen = size();
    const int *x_plo = src.loVect();
    const int *x_plen = src.size();
    const int *subbox_lo = destbox.loVect();
    const int *subbox_len = destbox.size().getVect();
    const int *bx_lo = srcbox.loVect();

    T * th_p = dataPtr(destcomp);
    const T * x_p = src.dataPtr(srccomp);

    const int aa0 = subbox_lo[0] - th_plo[0];
    const int aa1 = subbox_lo[1] - th_plo[1];
    const int aa2 = subbox_lo[2] - th_plo[2];
    const int bb0 = bx_lo[0] - x_plo[0];
    const int bb1 = bx_lo[1] - x_plo[1];
    const int bb2 = bx_lo[2] - x_plo[2];
    int na, nx, ka, kx;

    for(int n = 0; n < numcomp; ++n) {
      na = aa2 + n*th_plen[2];
      nx = bb2 + n*x_plen[2];
      for(int k = 0; k < subbox_len[2]; ++k) {
        ka = aa1 + th_plen[1]*(k + na);
        kx = bb1 + x_plen[1]*(k + nx);
        for(int j = 0; j < subbox_len[1]; ++j) {

          T *th_pp = th_p + aa0 + th_plen[0]*(j + ka);

          const T *x_pp = x_p + bb0 + x_plen[0]*(j + kx);

          for(int i = 0; i < subbox_len[0]; ++i, ++th_pp) {
            //T &thisR = * th_pp;
            //const T & srcR = x_pp[i];
            //thisR = srcR;
            (*th_pp) = x_pp[i];
          }
        }
      }
    }

    //qsubbox_len0[subbox_len[0]]++;
    //qsubbox_len1[subbox_len[1]]++;
    //qsubbox_len2[subbox_len[2]]++;
    //qNN += numcomp;
    //qsubbox_len0[qiter] = subbox_len[0];
    //qsubbox_len1[qiter] = subbox_len[1];
    //qsubbox_len2[qiter] = subbox_len[2];
    //qNN[qiter] = numcomp;
    //qiter ++;
  }
}

template <class T>
inline
void
BaseFab<T>::myperformCopyDiag (const BaseFab<T>& src,
                               const Box&        srcbox,
                               int               srccomp,
                               const Box&        destbox,
                               int               destcomp,
                               int               numcomp) {
#if 0
  cout << "---------------------------------------" << endl;
  cout << " loVect         " << loVect()[0] << " " << loVect()[1] << " " << loVect()[2] << endl;
  cout << " size           " << size()[0] << " " << size()[1] << " " << size()[2] << endl;

  cout << " src.loVect     " << src.loVect()[0] << " " << src.loVect()[1] << " " << src.loVect()[2] << endl;
  cout << " src.size       " << src.size()[0] << " " << src.size()[1] << " " << src.size()[2] << endl;

  cout << " destbox.loVect " << destbox.loVect()[0] << " " << destbox.loVect()[1] << " " << destbox.loVect()[2] << endl;
  cout << " destbox.size   " << destbox.size()[0] << " " << destbox.size()[1] << " " << destbox.size()[2] << endl;

  cout << " srcbox.loVect  " << srcbox.loVect()[0] << " " << srcbox.loVect()[1] << " " << srcbox.loVect()[2] << endl;
  cout << " srcbox.size    " << srcbox.size()[0] << " " << srcbox.size()[1] << " " << srcbox.size()[2] << endl;

  //if (destbox.size().getVect()[0] != destbox.size()[0] &&
  //  destbox.size().getVect()[1] != destbox.size()[1] &&
  //  destbox.size().getVect()[2] != destbox.size()[2] ) {
  //cout << " ldshfjslfdkjh" << endl;
  //}
#endif

  // if (srcbox == destbox) -- try this

  if(!(( destbox ).isEmpty())) {

    const int *th_plo = loVect();
    const int *th_plen = size();
    const int *x_plo = src.loVect();
    const int *x_plen = src.size();
    const int *subbox_lo = destbox.loVect();
    const int *subbox_len = destbox.size().getVect();  // getVect()?
    const int *bx_lo = srcbox.loVect();

    T * th_p = dataPtr(destcomp);
    const T * x_p = src.dataPtr(srccomp);

    const int aa0 = subbox_lo[0] - th_plo[0];
    const int aa1 = subbox_lo[1] - th_plo[1];
    const int aa2 = subbox_lo[2] - th_plo[2];
    const int bb0 = bx_lo[0] - x_plo[0];
    const int bb1 = bx_lo[1] - x_plo[1];
    const int bb2 = bx_lo[2] - x_plo[2];
    int na, nx, ka, kx;

    if (subbox_len[0] == 1) {

      for(int n = 0; n < numcomp; ++n) {
        na = aa2 + n*th_plen[2];
        nx = bb2 + n*x_plen[2];
        for(int k = 0; k < subbox_len[2]; ++k) {
          ka = aa1 + th_plen[1]*(k + na);
          kx = bb1 + x_plen[1]*(k + nx);
          for(int j = 0; j < subbox_len[1]; ++j) {
            //T *th_pp = th_p + aa0 + th_plen[0]*(j + ka);
            //(*th_pp) = x_p[bb0 + x_plen[0]*(j + kx)];
            th_p[aa0 + th_plen[0]*(j + ka)] = x_p[bb0 + x_plen[0]*(j + kx)];
          }
        }
      }

    } else if (subbox_len[1] == 1) {

      for(int n = 0; n < numcomp; ++n) {
        na = aa2 + n*th_plen[2];
        nx = bb2 + n*x_plen[2];
        for(int k = 0; k < subbox_len[2]; ++k) {
          ka = aa1 + th_plen[1]*(k + na);
          kx = bb1 + x_plen[1]*(k + nx);

          T *th_pp = th_p + aa0 + th_plen[0]*ka;
          const T *x_pp = x_p + bb0 + x_plen[0]*kx;

          for(int i = 0; i < subbox_len[0]; ++i, ++th_pp) {
            (*th_pp) = x_pp[i];
          }
        }
      }

    } else if (subbox_len[2] == 1) {

      for(int n = 0; n < numcomp; ++n) {
        na = aa2 + n*th_plen[2];
        nx = bb2 + n*x_plen[2];
        ka = aa1 + th_plen[1]*na;
        kx = bb1 + x_plen[1]*nx;
        for(int j = 0; j < subbox_len[1]; ++j) {

          T *th_pp = th_p + aa0 + th_plen[0]*(j + ka);
          const T *x_pp = x_p + bb0 + x_plen[0]*(j + kx);

          for(int i = 0; i < subbox_len[0]; ++i, ++th_pp) {
            (*th_pp) = x_pp[i];
          }
        }
      }

    } else {

      for(int n = 0; n < numcomp; ++n) {
        na = aa2 + n*th_plen[2];
        nx = bb2 + n*x_plen[2];
        for(int k = 0; k < subbox_len[2]; ++k) {
          ka = aa1 + th_plen[1]*(k + na);
          kx = bb1 + x_plen[1]*(k + nx);
          for(int j = 0; j < subbox_len[1]; ++j) {

            T *th_pp = th_p + aa0 + th_plen[0]*(j + ka);

            const T *x_pp = x_p + bb0 + x_plen[0]*(j + kx);

            for(int i = 0; i < subbox_len[0]; ++i, ++th_pp) {
              (*th_pp) = x_pp[i];
            }
          }
        }
      }
    }

  }
}
#endif
