/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <csignal>

#include "SPMD.H"

void AttachDebugger(int a_sig)
{
  if (a_sig == 0)
  {
    return;
  }

  char buf[300];
  char* display;
  FILE *f;
  int proc = getpid();
  char title[75];

  char binaryName[101];
  memset(binaryName, 0, 101); // readlink doesn't insert NULL at end

  char linkName[30];

  // voodoo specific to linux /proc system
  sprintf(linkName,"/proc/%d/exe",proc);
  printf("%s\n",linkName);
  ssize_t s = readlink(linkName, binaryName, 100);

  int a = 50;

  if (s > 100)
  {
    //filename too darn big
    signal(a_sig, SIG_DFL);
    raise(a_sig);
  }

  if (s < a)
  {
    a = s;
  }

  // try to find a display to send this debugger to.
  display = getenv("DISPLAY");

  if (display != NULL)
  {
    sprintf(title, "'%s %d'", binaryName+s-a, procID());
    sprintf(buf,"emacs -title %s -display %s --eval '(progn (gdb \" gdb -q %s --pid %d\") )'",
            title, display, binaryName, proc);
    printf("%s\n",buf);
    f = popen(buf,"w");
  }
  else
  {
    sprintf(buf,"gdb -q  %s %d", binaryName, proc);
    f=popen(buf, "w");
    fputs("where\ndetach\nquit\n", f);
  }

  pclose(f);

  // sleep(20);
  signal(a_sig, SIG_DFL);
  raise(a_sig);
}

// when running the code inside an actual debugger, the debugger signal handler
// will come in before this one, so the debugger should function correctly.
void  registerDebugger()
{
#ifndef CH_IRIX64  // for some reason IRIX64 doesn't like ANSI signals.
  signal(SIGABRT, AttachDebugger);
#endif
}
