/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _CH_COMPLEX_H_
#define _CH_COMPLEX_H_

#ifdef CH_USE_COMPLEX

#include <complex>
#include "REAL.H"
using std::complex ;

/// Complex numbers suitable for use in FABs
/** Chombo has its own complex number class so it can:
 *  1) provide the same class name regardless of the current precision
 *     (just like the Chombo "Real" type)
 *  2) provide write access to the real and imaginary parts of the
 *     complex number without going through a constructor (this is not
 *     allowed by the C++ 1999 standard).
 */

class Complex : public complex<Real>
{
public:
  /// return a non-const reference to the real part of the complex number
  inline Real & re() ;
  /// return a copy of the real part of the complex number
  inline Real   re() const { return real(); };
  /// return a non-const reference to the imaginary part of the complex number
  inline Real & im() ;
  /// return a copy of the imaginary part of the complex number
  inline Real   im() const { return imag(); };
};

//XXX dont check the compiler -- just assume this will work
//XXX#if defined(__GNUC__) || defined(__INTEL_COMPILER__) || defined(__IBMCPP__) || defined(__HP_aCC) || defined(__DECCXX)

// The compilers we've seen implement complex as a 2-array of
// numbers of the appropriate type, with the real part first.
// Run the test program "Chombo/lib/test/BoxTools/test_complex.cpp" to check.
// This is the most likely implementation, and the only real problems are
// whether the compiler puts the data at the beginning of the class's
// memory, and whether it will allow the cast.
inline Real & Complex::re() {
  return (((Real*)this)[0]) ;
}

inline Real & Complex::im() {
  return (((Real*)this)[1]) ;
}

//XXX#else
//XXX#error Complex type is not implemented for this compiler; contact the Chombo developers.
//XXX#endif

#endif //CH_USE_COMPLEX
#endif //_CH_COMPLEX_H_
