#ifdef CH_LANG_CC
/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//
#endif

#ifndef _CONSTANTS_H_
#define _CONSTANTS_H_

#if defined(CH_LANG_FORT) || defined(CH_CRAY)

/* use single-precision constants on the T3E regardless of whether
 * USE_FLOAT or USE_DOUBLE is defined
*/
#if defined(CH_USE_FLOAT)

#define zero     (0.0e0)
#define one      (1.0e0)
#define two      (2.0e0)
#define three    (3.0e0)
#define four     (4.0e0)
#define five     (5.0e0)
#define six      (6.0e0)
#define seven    (7.0e0)
#define eight    (8.0e0)
#define nine     (9.0e0)
#define ten     (10.0e0)
#define twelve  (12.0e0)
#define fifteen (15.0e0)
#define sixteen (16.0e0)
#define twenty  (20.0e0)
#define tenth   (0.100e0)
#define eighth  (0.125e0)
#define sixth   (1.000e0 /  6.000e0)
#define fourth  (0.250e0)
#define third   (1.000e0 /  3.000e0)
#define half    (0.500e0)
#define two3rd  (2.000e0 /  3.000e0)
#define Pi      (3.14159265358979323846e0)

#define ZERO     (0.0E0)
#define ONE      (1.0E0)
#define TWO      (2.0E0)
#define THREE    (3.0E0)
#define FOUR     (4.0E0)
#define FIVE     (5.0E0)
#define SIX      (6.0E0)
#define SEVEN    (7.0E0)
#define EIGHT    (8.0E0)
#define NINE     (9.0E0)
#define TEN     (10.0E0)
#define TWELVE  (12.0E0)
#define FIFTEEN (15.0E0)
#define SIXTEEN (16.0E0)
#define TWENTY  (20.0E0)
#define TENTH   (0.100E0)
#define EIGHTH  (0.125E0)
#define SIXTH   (1.000E0 / 6.000E0)
#define FOURTH  (0.250E0)
#define THIRD   (1.000E0 / 3.000E0)
#define HALF    (0.500E0)
#define TWO3RD  (2.000E0 / 3.000E0)
#define PI      (3.14159265358979323846E0)

#elif defined(CH_USE_DOUBLE)

#define zero     (0.0d0)
#define one      (1.0d0)
#define two      (2.0d0)
#define three    (3.0d0)
#define four     (4.0d0)
#define five     (5.0d0)
#define six      (6.0d0)
#define seven    (7.0d0)
#define eight    (8.0d0)
#define nine     (9.0d0)
#define ten     (10.0d0)
#define twelve  (12.0d0)
#define fifteen (15.0d0)
#define sixteen (16.0d0)
#define twenty  (20.0d0)
#define tenth   (0.100d0)
#define eighth  (0.125d0)
#define sixth   (1.000d0 / 6.000d0)
#define fourth  (0.250d0)
#define third   (1.000d0 / 3.000d0)
#define half    (0.500d0)
#define two3rd  (2.000d0 / 3.000d0)
#define Pi      (3.14159265358979323846264338327950288d0)

#define ZERO    (0.0D0)
#define ONE     (1.0D0)
#define TWO     (2.0D0)
#define THREE   (3.0D0)
#define FOUR    (4.0D0)
#define FIVE    (5.0D0)
#define SIX     (6.0D0)
#define SEVEN   (7.0D0)
#define EIGHT   (8.0D0)
#define NINE    (9.0D0)
#define TEN     (10.0D0)
#define TWELVE  (12.0D0)
#define FIFTEEN (15.0D0)
#define SIXTEEN (16.0D0)
#define TWENTY  (20.0D0)
#define TENTH   (0.100D0)
#define EIGHTH  (0.125D0)
#define SIXTH   (1.000D0 / 6.000D0)
#define FOURTH  (0.250D0)
#define THIRD   (1.000D0 / 3.000D0)
#define HALF    (0.500D0)
#define TWO3RD  (2.000D0 / 3.000D0)
#define PI      (3.14159265358979323846264338327950288D0)

#else

#error One of CH_USE_{FLOAT,DOUBLE} must be defined

#endif

#endif /*CH_LANG_FORT*/

#endif /*CH_CONSTANTS_H*/
