/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "Counter.H"
#include <cstdio>
#include "parstream.H"  // for pout()'s

int IterationFlavor;
int BIterationFlavor;
int GlobalLevel=0;

CounterStruct NoelsSooperIterationCounter[MAX_ITERATION_COUNTERS];

void writeOutNoelsSooperIterationCounterStuff(void) {

  // horrible hack!
  sprintf(NoelsSooperIterationCounter[0].Name, "Nothing");

  for(int i=0; i<_MAX_LEVELS; i++) {
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_init].Name,
            "levelSolve init level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_phi].Name,
            "levelSolve Phi level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_pi].Name,
            "levelSolve Pi level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_viscous1].Name,
            "levelSolve viscous1 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_viscous2].Name,
            "levelSolve viscous2 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_ustarinit1].Name,
            "levelSolve ustarinit1 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_ustarinit2].Name,
            "levelSolve ustarinit2 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_ustar1].Name,
            "levelSolve ustar1 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_ustar2].Name,
            "levelSolve ustar2 level%d", i);

    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_TGA1].Name,
            "levelSolve TGA1 level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolve_TGA2].Name,
            "levelSolve TGA2 level%d", i);

    sprintf(NoelsSooperIterationCounter[i+COUNTER_levelSolveH].Name,
            "levelSolveH level%d", i);

    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_INSinit].Name,
            "solveAMR INSinit level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_smoothVelField].Name,
            "solveAMR smoothVelField level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_smoothVelFieldScalars].Name,
            "solveAMR smoothVelFieldScalars level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_diffusion].Name,
            "solveAMR diffusion level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_viscous].Name,
            "solveAMR viscous level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_syncProjection].Name,
            "solveAMR syncProjection level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_VDCorrection].Name,
            "solveAMR VDCorrection level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_initVelProj].Name,
            "solveAMR initVelProj level%d", i);

    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_MacProject].Name,
            "solveAMR MacProject level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_CCProject].Name,
            "solveAMR CCProject level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_initData].Name,
            "solveAMR initData level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_updatePPos].Name,
            "solveAMR updatePPos level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_initPressure].Name,
            "solveAMR initPressure level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_solveAMR_projectedForce].Name,
            "solveAMR projectedForce level%d", i);

    sprintf(NoelsSooperIterationCounter[i+COUNTER_bottomSmoother_Helmholtz].Name,
            "bottomSmoother Helmholtz level%d", i);
    sprintf(NoelsSooperIterationCounter[i+COUNTER_bottomSmoother_Poisson].Name,
            "bottomSmoother Poisson level%d", i);

  }

  char tempstring[180];

  sprintf(tempstring, "Manual Counters  _MAX_LEVELS=%d MAX_ITERATION_COUNTERS=%d\n",
          _MAX_LEVELS, MAX_ITERATION_COUNTERS);
  pout() << tempstring;

  // do the first one explicitly
  int i=0;
  sprintf(tempstring, "Counter%03d %38s N=%9d Val=%9d Nmax=%6d wc=%12.4f\n", i,
          NoelsSooperIterationCounter[i].Name,
          NoelsSooperIterationCounter[i].Count,
          NoelsSooperIterationCounter[i].Value,
          NoelsSooperIterationCounter[i].NmaxIteration,
          NoelsSooperIterationCounter[i].wc);
  pout() << tempstring;

  // do the rest on a per-level basis
  for (int i=1; i < MAX_ITERATION_COUNTERS-1; i+=_MAX_LEVELS) {
    int countTotal=0;
    for (int l=0; l<_MAX_LEVELS; l++) {
      countTotal += NoelsSooperIterationCounter[i+l].Count;
    }
    if (countTotal!=0) {
      for (int l=0; l<_MAX_LEVELS; l++) {
        sprintf(tempstring, "Counter%03d %38s N=%9d Val=%9d Nmax=%6d wc=%12.4f\n", i+l,
                NoelsSooperIterationCounter[i+l].Name,
                NoelsSooperIterationCounter[i+l].Count,
                NoelsSooperIterationCounter[i+l].Value,
                NoelsSooperIterationCounter[i+l].NmaxIteration,
                NoelsSooperIterationCounter[i+l].wc);
        pout() << tempstring;
      }
    }
  }
}

//#define NQ 1000000
//extern unsigned long int qiter;
//extern unsigned int qsubbox_len0[NQ];
//extern unsigned int qsubbox_len1[NQ];
//extern unsigned int qsubbox_len2[NQ];
//extern unsigned int qNN[NQ];
