/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "DebugOut.H"
#include "LayoutIterator.H"
#include "IntVectSet.H"

static const char* indent2 = "      " ;

void dumpLDF(const LevelData<FArrayBox>* a_ldfabPtr)
{
  const LevelData<FArrayBox>& ldfab = *a_ldfabPtr;
  Vector<Box> boxes;

  const DisjointBoxLayout& memDBL = ldfab.getBoxes();
  LayoutIterator lit  = memDBL.layoutIterator();
  for (lit.reset(); lit.ok(); ++lit)
  {
    boxes.push_back(memDBL.get(lit()));
  }

  Vector<int> assign(boxes.size(), 0);

  DisjointBoxLayout locDBL(boxes, assign);
  locDBL.close();

  LevelData<FArrayBox> locLDF(locDBL,
                              ldfab.nComp(),
                              ldfab.ghostVect());

  const Interval& interv = ldfab.interval();
  ldfab.copyTo(interv,
               locLDF,
               interv);

  DataIterator dit = locLDF.dataIterator();
  for (dit.reset(); dit.ok(); ++dit)
  {
    const FArrayBox& fab = locLDF[dit()];
    BoxIterator bit(fab.box());
    for (bit.reset(); bit.ok(); ++bit)
    {
      pout() << "\t" << bit() << "\t" << fab(bit(),0) << endl;;
    }
  }
}

void dumpIVSFAB(const IVSFAB<Real>* a_ivsfabPtr)
{
  const IVSFAB<Real>& ivsfab = *a_ivsfabPtr;
  const IntVectSet& ivs = ivsfab.getIVS();
  for (IVSIterator ivsit(ivs); ivsit.ok(); ++ivsit)
  {
    const IntVect& iv = ivsit();

    for(int icomp = 0; icomp < ivsfab.nComp(); icomp++)
    {
      pout() << ivsfab(iv, icomp) << "  " << endl;
    }
  }
}

void dumpDBL(const DisjointBoxLayout* a_dblInPtr)
{
  pout() << indent2 <<"DisjointBoxLayout ";

  if (a_dblInPtr->isClosed())
  {
    pout() << "closed\n";
  }
  else
  {
    pout() << "open\n";
  }

  pout() << *a_dblInPtr << endl;
}

void dumpBL(const BoxLayout* a_dblInPtr)
{
  pout() << indent2 <<"BoxLayout ";

  if (a_dblInPtr->isClosed())
  {
    pout() << "closed\n";
  }
  else
  {
    pout() << "open\n";
  }

  pout() << *a_dblInPtr << endl;
}

void dumpIVS(const IntVectSet* a_ivsPtr)
{
  const IntVectSet& ivs = *a_ivsPtr;
  IVSIterator it(ivs);

  pout() << indent2 << ": IntVects in the IVS are:" << endl;
  pout() << indent2;

  for (it.begin(); it.ok(); ++it)
  {
    pout() << it() << "  ";
  }

  pout() << endl;
}

void dumpBox(const Box* a_boxPtr)
{
  pout() << indent2 << *a_boxPtr << endl;
}

void dumpVBox(const Vector<Box>* a_vectPtr)
{
  Vector<Box> vect = *a_vectPtr;
  for (int ivec = 0; ivec < vect.size(); ivec++)
  {
    pout() << indent2 << vect[ivec] << endl;
  }
}

void dumpVVBox(const Vector<Vector<Box> >* a_vectPtr)
{
  Vector<Vector<Box> > vect = *a_vectPtr;
  for (int iveco = 0; iveco < vect.size(); iveco++)
  {
    pout() << indent2;

    for (int iveci = 0; iveci < vect[iveco].size(); iveci++)
    {
      pout() <<  vect[iveco][iveci] << "   ";
    }

    pout() << endl;
  }
}

void dumpLDDBL(const LevelData<FArrayBox>* a_ldfabPtr)
{
  const DisjointBoxLayout& dbl = a_ldfabPtr->disjointBoxLayout();
  dumpDBL(&dbl);
}
