/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <cfloat>
#include <cmath>
#include <cstring>
#include <cassert>
using std::cin;
using std::cout;
using std::cerr;
using std::setw;
using std::setprecision;
using std::ios;
using std::pow;
using std::sqrt;

#include "Misc.H"
#include "FArrayBox.H"
#include "MayDay.H"

#include "Timer.H"

extern Timer Everything;
//Timer TFAB_norm   ("norm",     Everything, 6);
//Timer TFAB_sumPow ("sumPow",     Everything, 6);
//Timer TFAB_min    ("min",     Everything, 6);
//Timer TFAB_min2   ("min2",     Everything, 6);
//Timer TFAB_max    ("max",     Everything, 6);
//Timer TFAB_max2   ("max2",     Everything, 6);
//Timer TFAB_minIndex ("minIndex", Everything, 6);
//Timer TFAB_maxIndex ("maxIndex", Everything, 6);
//Timer TFAB_maxIndex2 ("maxIndex2", Everything, 6);
//Timer TFAB_maskLT ("maskLT", Everything, 6);
//Timer TFAB_maskLE ("maskLE", Everything, 6);
//Timer TFAB_maskEQ ("maskEQ", Everything, 6);
//Timer TFAB_maskGT ("maskGT", Everything, 6);
//Timer TFAB_maskGE ("maskGE", Everything, 6);
//Timer TFAB_abs ("abs", Everything, 6);
//Timer TFAB_abs2 ("abs2", Everything, 6);
//Timer TFAB_abs3 ("abs3", Everything, 6);
//Timer TFAB_sum ("sum", Everything, 6);
//Timer TFAB_sum2 ("sum2", Everything, 6);
//Timer TFAB_negate ("negate", Everything, 6);
//Timer TFAB_negate2 ("negate2", Everything, 6);
//Timer TFAB_negate3 ("negate3", Everything, 6);
//Timer TFAB_inver ("inver", Everything, 6);
//Timer TFAB_invert2 ("invert2", Everything, 6);
//Timer TFAB_invert3 ("invert3", Everything, 6);
//Timer TFAB_plusequal ("plusequal", Everything, 6);
//Timer TFAB_plus ("plus", Everything, 6);
//Timer TFAB_plus2 ("plus2", Everything, 6);
//Timer TFAB_plus3 ("plus3", Everything, 6);
//Timer TFAB_plus4 ("plus4", Everything, 6);
//Timer TFAB_plusequal2 ("plusequal2", Everything, 6);
//Timer TFAB_plus5 ("plus5", Everything, 6);
//Timer TFAB_plus6 ("plus6", Everything, 6);
//Timer TFAB_plus7 ("plus7", Everything, 6);
//Timer TFAB_minusequal ("minusequal", Everything, 6);
//Timer TFAB_minus ("minus", Everything, 6);
//Timer TFAB_minus2 ("minus2", Everything, 6);
//Timer TFAB_minus3 ("minus3", Everything, 6);
//Timer TFAB_timesequal ("timesequal", Everything, 6);
//Timer TFAB_mult ("mult", Everything, 6);
//Timer TFAB_mult2 ("mult2", Everything, 6);
//Timer TFAB_timesequal2 ("timesequal2", Everything, 6);
//Timer TFAB_mult3 ("mult3", Everything, 6);
//Timer TFAB_mult4 ("mult4", Everything, 6);
//Timer TFAB_mult5 ("mult5", Everything, 6);
//Timer TFAB_divequal ("divequal", Everything, 6);
//Timer TFAB_divide ("divide", Everything, 6);
//Timer TFAB_divide2 ("divide2", Everything, 6);
//Timer TFAB_divequal2 ("divequal2", Everything, 6);
//Timer TFAB_divide3 ("divide3", Everything, 6);
//Timer TFAB_divide4 ("divide4", Everything, 6);
//Timer TFAB_divide5 ("divide5", Everything, 6);
//Timer TFAB_performCopy ("performCopy", Everything, 6);

FArrayBox::FArrayBox()
  :
  BaseFab<Real>()
{
}

FArrayBox::FArrayBox(const Box& a_box,
                     int        a_n,
                     Real*      a_alias)
  :
  BaseFab<Real>(a_box,a_n,a_alias)
{
  // Note: All work is done in BaseFab<Real> constructor
}

FArrayBox::~FArrayBox() {}

Real FArrayBox::norm(const Box& a_subbox,
                     int        a_p,
                     int        a_comp,
                     int        a_numcomp) const
{
  assert(a_comp >= 0 && a_comp + a_numcomp <= nComp());
  assert(a_numcomp > 0);
  assert(a_p >= 0);
  assert(m_domain.contains(a_subbox));

  //timercomment//TFAB_norm.start();
  Real* tmp = 0;
  int tmplen = 0;
  Real nrm = 0;

  if (a_p == 0)
  {
    // here begins a normed fab function piece
    ForAllThisCPencil(Real,a_subbox,a_comp,a_numcomp)
    {
      const Real* row = &thisR;
      if (tmp == 0)
      {
        tmp = new Real[thisLen];
        tmplen = thisLen;
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] = Abs(Real(row[i]));
        }
      }
      else
      {
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] = Max(tmp[i],Real(Abs(row[i])));
        }
      }
    } EndForPencil

    nrm = tmp[0];
    for (int i = 1; i < tmplen; i++)
    {
      nrm = Max(nrm, tmp[i]);
    }
    // here it ends
  }
  else if (a_p == 1)
  {
    // here begins a normed fab function piece
    ForAllThisCPencil(Real,a_subbox,a_comp,a_numcomp)
    {
      const Real* row = &thisR;
      if (tmp == 0)
      {
        tmp = new Real[thisLen];
        tmplen = thisLen;
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] = Abs(Real(row[i]));
        }
      }
      else
      {
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] += Abs(Real(row[i]));
        }
      }
    } EndForPencil

    nrm = tmp[0];
    for (int i = 1; i < tmplen; i++)
    {
      nrm += tmp[i];
    }
    // here it ends
  }
  else if (a_p == 2)
  {
    nrm = sqrt(sumPow(a_subbox, 2, a_comp, a_numcomp));
  }
  else
  {
    // so standard norms weren't good enough for you?
    Real invpwr = 1.0/a_p;
    nrm = pow(sumPow(a_subbox, a_p, a_comp, a_numcomp),invpwr);
  }

  delete [] tmp;

  //timercomment//TFAB_norm.stop();
  return nrm;
}

Real FArrayBox::norm(int a_p,
                     int a_comp,
                     int a_numcomp) const
{
  return norm(m_domain,a_p,a_comp,a_numcomp);
}

// utility function used in norms and such only works for a_p >= 2
Real FArrayBox::sumPow(const Box& a_subbox,
                       int        a_p,
                       int        a_comp,
                       int        a_numcomp) const
{
  assert(a_p >= 2);
  assert(a_numcomp > 0);
  assert(m_domain.contains(a_subbox));

  //timercomment//TFAB_sumPow.start();
  Real sum = 0;
  Real* tmp = NULL;
  int tmplen = 0;

  if (a_p == 2)
  {
    ForAllThisCPencil(Real,a_subbox,a_comp,a_numcomp)
    {
      const Real* row = &thisR;
      if (tmp == 0)
      {
        tmp = new Real[thisLen];
        tmplen = thisLen;
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] = row[i]*row[i];
        }
      }
      else
      {
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] += row[i]*row[i];
        }
      }
    } EndForPencil

    sum = tmp[0];
    for (int i = 1; i < tmplen; i++)
    {
      sum += tmp[i];
    }
  }
  else
  {
    // so standard norms weren't good enough for you?
    Real pwr = Real(a_p);

    ForAllThisCPencil(Real,a_subbox,a_comp,a_numcomp)
    {
      const Real* row = &thisR;
      if (tmp == 0)
      {
        tmp = new Real[thisLen];
        tmplen = thisLen;
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] = pow(row[i],pwr);
        }
      }
      else
      {
        for (int i = 0; i < thisLen; i++)
        {
          tmp[i] += pow(row[i],pwr);
        }
      }
    } EndForPencil

    sum = tmp[0];
    for (int i = 1; i < tmplen; i++)
    {
      sum += tmp[i];
    }
  }

  delete [] tmp;
  //timercomment//TFAB_sumPow.stop();
  return sum;
}

// Take the dot product of "this" and "a_fab2" over their common box and
// all components.
Real FArrayBox::dotProduct(const FArrayBox& a_fab2) const
{
  Real dot = 0.0;
  const FArrayBox& fab1 = *this;
  Box commonBox = fab1.box() & a_fab2.box();
  int startcomp = 0;
  int endcomp = fab1.nComp()-1;
  int numcomp = endcomp+1;

  assert(fab1.nComp() == a_fab2.nComp());

  if (!commonBox.isEmpty())
  {
    // FORT_DOTPRODUCT(CHF_REAL(dot),
    //                 CHF_CONST_FRA(fab1),
    //                 CHF_CONST_FRA(a_fab2),
    //                 CHF_BOX(commonBox),
    //                 CHF_CONST_INT(startcomp),
    //                 CHF_CONST_INT(endcomp));

    ForAllThisCBNNXC(Real, commonBox, startcomp, numcomp, a_fab2, startcomp)
    {
      dot += thisR * a_fab2R;
    } EndForTX
  }

  return dot;
}

//here is where the orderedfab functions start

Real FArrayBox::min(int a_comp) const
{
  //timercomment//TFAB_min.start();
  Real *_min_row = 0;
  int _X_len = 0;

  ForAllThisCPencil(Real,m_domain,a_comp,1)
  {
    const Real* _row = &thisR;
    if (_min_row == 0)
    {
      _min_row = new Real[thisLen];
      _X_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _min_row[i] = _row[i];
      }
    }
    else
    {
      for (int i = 0; i < thisLen; i++)
      {
        _min_row[i] = Min(_row[i],_min_row[i]);
      }
    }
  } EndForPencil;

  Real _min = _min_row[0];
  for (int i = 1; i < _X_len; i++)
  {
    _min = Min(_min,_min_row[i]);
  }

  delete [] _min_row;
  //timercomment//TFAB_min.stop();
  return _min;
}

Real FArrayBox::min(const Box& a_subbox,
                    int        a_comp) const
{
  assert(m_domain.contains(a_subbox));

  //timercomment//TFAB_min2.start();
  Real *_min_row = 0;
  int _X_len = 0;

  ForAllThisCPencil(Real,a_subbox,a_comp,1)
  {
    const Real* _row = &thisR;
    if (_min_row == 0)
    {
      _min_row = new Real[thisLen];
      _X_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _min_row[i] = _row[i];
      }
    }
    else
    {
      for (int i = 0; i < thisLen; i++)
      {
        _min_row[i] = Min(_row[i],_min_row[i]);
      }
    }
  } EndForPencil;

  Real _min = _min_row[0];
  for (int i = 1; i < _X_len; i++)
  {
    _min = Min(_min,_min_row[i]);
  }

  delete [] _min_row;
  //timercomment//TFAB_min2.stop();
  return _min;
}

Real FArrayBox::max(int a_comp) const
{
  //timercomment//TFAB_max.start();
  Real *_max_row = 0;
  int _X_len = 0;

  ForAllThisCPencil(Real,m_domain,a_comp,1)
  {
    const Real* _row = &thisR;
    if (_max_row== 0)
    {
      _max_row = new Real[thisLen];
      _X_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _max_row[i] = _row[i];
      }
    }
    else
    {
      for (int i = 0; i < thisLen; i++)
      {
        _max_row[i] = Max(_row[i],_max_row[i]);
      }
    }
  } EndForPencil;

  Real _max = _max_row[0];
  for (int i = 1; i < _X_len; i++)
  {
    _max = Max(_max,_max_row[i]);
  }

  delete [] _max_row;
  //timercomment//TFAB_max.stop();
  return _max;
}

Real FArrayBox::max(const Box& a_subbox,
                    int        a_comp) const
{
  assert(m_domain.contains(a_subbox));
  //timercomment//TFAB_max2.start();
  Real *_max_row = 0;
  int _X_len = 0;

  ForAllThisCPencil(Real,a_subbox,a_comp,1)
  {
    const Real* _row = &thisR;
    if (_max_row == 0)
    {
      _max_row = new Real[thisLen];
      _X_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _max_row[i] = _row[i];
      }
    }
    else
    {
      for (int i = 0; i < thisLen; i++)
      {
        _max_row[i] = Max(_row[i],_max_row[i]);
      }
    }
  } EndForPencil;

  Real _max = _max_row[0];
  for (int i = 1; i < _X_len; i++)
  {
    _max = Max(_max,_max_row[i]);
  }

  delete [] _max_row;
  //timercomment//TFAB_max2.stop();
  return _max;
}

IntVect FArrayBox::minIndex(int a_comp) const
{
  IntVect _min_loc(m_domain.smallEnd());
  Real _min_val = (*this).operator()(_min_loc,a_comp);

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;

    if (thisR < _min_val)
    {
      _min_val = thisR;
      D_EXPR(_min_loc[0] = iR,
             _min_loc[1] = jR,
             _min_loc[2] = kR);
    }
  } EndFor

  return _min_loc;
}

IntVect FArrayBox::minIndex(const Box& a_subbox,
                            int        a_comp) const
{
  assert(m_domain.contains(a_subbox));

  //timercomment//TFAB_minIndex.start();
  IntVect _min_loc(a_subbox.smallEnd());
  Real _min_val = (*this).operator()(_min_loc,a_comp);

  ForAllThisCBNN(Real,a_subbox,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;

    if (thisR < _min_val)
    {
      _min_val = thisR;
      D_EXPR(_min_loc[0] = iR,
             _min_loc[1] = jR,
             _min_loc[2] = kR);
    }
  } EndFor
  //timercomment//TFAB_minIndex.stop();
  return _min_loc;
}

IntVect FArrayBox::maxIndex(int a_comp) const
{
  //timercomment//TFAB_maxIndex.start();
  IntVect _max_loc(m_domain.smallEnd());
  Real _max_val = (*this).operator()(_max_loc,a_comp);

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;

    if (thisR > _max_val)
    {
      _max_val = thisR;
      D_EXPR(_max_loc[0] = iR,
             _max_loc[1] = jR,
             _max_loc[2] = kR);
    }
  } EndFor

  //timercomment//TFAB_maxIndex.stop();
  return _max_loc;
}

IntVect FArrayBox::maxIndex(const Box& a_subbox,
                            int        a_comp) const
{
  assert(m_domain.contains(a_subbox));

  IntVect _max_loc(a_subbox.smallEnd());
  Real _max_val = (*this).operator()(_max_loc,a_comp);

  ForAllThisCBNN(Real,a_subbox,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;

    if (thisR > _max_val)
    {
      _max_val = thisR;
      D_EXPR(_max_loc[0] = iR,
             _max_loc[1] = jR,
             _max_loc[2] = kR);
    }
  } EndFor

  return _max_loc;
}

int FArrayBox::maskLT(BaseFab<int>& a_mask,
                      Real          a_val,
                      int           a_comp) const
{
  a_mask.resize(m_domain,1);
  a_mask.setVal(0);

  int *mptr = a_mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;
    int ir = iR;
    int jr = jR;
    ir = jr;
    jr = ir;
    int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);

    if (thisR < a_val)
    {
      mptr[ix] = 1;
      cnt++;
    }
  } EndFor

  return cnt;
}

int FArrayBox::maskLE(BaseFab<int>& a_mask,
                      Real          a_val,
                      int           a_comp) const
{
  a_mask.resize(m_domain,1);
  a_mask.setVal(0);

  int *mptr = a_mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;
    int ir = iR;
    int jr = jR;
    ir = jr;
    jr = ir;
    int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);

    if (thisR <= a_val)
    {
      mptr[ix] = 1;
      cnt++;
    }
  } EndFor

  return cnt;
}

int FArrayBox::maskEQ(BaseFab<int>& a_mask,
                      Real          a_val,
                      int           a_comp) const
{
  a_mask.resize(m_domain,1);
  a_mask.setVal(0);

  int *mptr = a_mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;
    int ir = iR;
    int jr = jR;
    ir = jr;
    jr = ir;
    int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);

    if (thisR == a_val)
    {
      mptr[ix] = 1;
      cnt++;
    }
  } EndFor

  return cnt;
}

int FArrayBox::maskGT(BaseFab<int>& a_mask,
                      Real          a_val,
                      int           a_comp) const
{
  a_mask.resize(m_domain,1);
  a_mask.setVal(0);

  int *mptr = a_mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;
    int ir = iR;
    int jr = jR;
    ir = jr;
    jr = ir;
    int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);

    if (thisR > a_val)
    {
      mptr[ix] = 1;
      cnt++;
    }
  } EndFor

  return cnt;
}

int FArrayBox::maskGE(BaseFab<int>& a_mask,
                      Real          a_val,
                      int           a_comp) const
{
  a_mask.resize(m_domain,1);
  a_mask.setVal(0);

  int *mptr = a_mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,m_domain,a_comp,1)
  {
    int nr = nR;
    int nr2 = nr;
    nr = nr2;
    int ir = iR;
    int jr = jR;
    ir = jr;
    jr = ir;
    int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);

    if (thisR >= a_val)
    {
      mptr[ix] = 1;
      cnt++;
    }
  } EndFor

  return cnt;
}

//here is where the orderedfab functions end

//here is where the normedfab functions begin

void FArrayBox::abs()
{
  ForAllThis(Real)
  {
    thisR = Abs(thisR);
  } EndFor
}

void FArrayBox::abs(int a_comp,
                    int a_numcomp)
{
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR = Abs(thisR);
  } EndFor
}

void FArrayBox::abs(const Box& a_subbox,
                    int        a_comp,
                    int        a_numcomp)
{
  assert(m_domain.contains(a_subbox));

  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR = Abs(thisR);
  } EndFor
}

//here is where the normedfab functions end

//here is where the arithfab functions begin

Real FArrayBox::sum(int a_comp,
                    int a_numcomp) const
{
  Real *_sum_row = 0;
  int _sum_len = 0;

  ForAllThisCPencil(Real,m_domain,a_comp,a_numcomp)
  {
    const Real* _row = &thisR;
    if (_sum_row == 0)
    {
      _sum_row = new Real[thisLen];
      _sum_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _sum_row[i] = _row[i];
      }
    }
    else
    {
      for (int i = 0; i < thisLen; i++)
      {
        _sum_row[i] += _row[i];
      }
    }
  } EndForPencil;

  Real _sum = _sum_row[0];
  for (int i = 1; i < _sum_len; i++)
  {
    _sum += _sum_row[i];
  }

  delete [] _sum_row;

  return _sum;
}

Real FArrayBox::sum(const Box& a_subbox,
                    int        a_comp,
                    int        a_numcomp) const
{
  assert(m_domain.contains(a_subbox));

  //timercomment//TFAB_sum2.start();
  Real *_sum_row = 0;
  int _sum_len = 0;

  ForAllThisCPencil(Real,a_subbox,a_comp,a_numcomp)
  {
    const Real* _row = &thisR;
    if (_sum_row == 0)
    {
      _sum_row = new Real[thisLen];
      _sum_len = thisLen;
      for (int i = 0; i < thisLen; i++)
      {
        _sum_row[i] = _row[i];
      }
    }
    else
      {
        for (int i = 0; i < thisLen; i++)
        {
          _sum_row[i] += _row[i];
        }
      }
  } EndForPencil;

  Real _sum = _sum_row[0];
  for (int i = 1; i < _sum_len; i++)
  {
    _sum += _sum_row[i];
  }

  delete [] _sum_row;
  //timercomment//TFAB_sum2.stop();
  return _sum;
}

FArrayBox& FArrayBox::invert(Real a_r)
{
  ForAllThis(Real)
  {
    thisR = a_r / thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::invert(Real a_r,
                             int  a_comp,
                             int  a_numcomp)
{
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR = a_r / thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::invert(Real       a_r,
                             const Box& a_subbox,
                             int        a_comp,
                             int        a_numcomp)
{
  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR = a_r / thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::negate(const Box& a_subbox,
                             int        a_comp,
                             int        a_numcomp)
{
  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR = - thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::negate(int a_comp,
                             int a_numcomp)
{
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR = - thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::negate()
{
  ForAllThis(Real)
  {
    thisR = - thisR;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::plus(Real       a_r,
                           const Box& a_subbox,
                           int        a_comp,
                           int        a_numcomp)
{
  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR += a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::plus(Real a_r,
                           int  a_comp,
                           int  a_numcomp)
{
  //timercomment//TFAB_plus.start();
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR += a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::operator += (Real a_r)
{
  ForAllThis(Real)
  {
    thisR += a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::operator += (const FArrayBox& a_x)
{
  ForAllThisXC(Real,a_x)
  {
    thisR += a_xR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(Real a_r)
{
  return operator += (a_r);
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_x)
{
  return operator += (a_x);
}

// added bvs Tue May 18, PDT 1999
FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           const Real&      a_scale)
{
  ForAllThisBNNXC(Real, a_src.box() & m_domain, 0, m_nvar, a_src, 0)
  {
    thisR += a_srcR * a_scale;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           const Real&      a_scale,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXC(Real, a_src.box() & m_domain, a_destcomp, a_numcomp, a_src, a_srccomp)
  {
    thisR += a_srcR * a_scale;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXC(Real,m_domain,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR += a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           const Box&       a_subbox,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  assert(m_domain.contains(a_subbox));

  ForAllThisBNNXC(Real,a_subbox,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR += a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           const Box&       a_srcbox,
                           const Box&       a_destbox,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXCBN(Real,a_destbox,a_destcomp,a_numcomp,a_src,a_srcbox,a_srccomp)
  {
    thisR += a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::plus(const FArrayBox& a_src,
                           const Box&       a_srcbox,
                           const Box&       a_destbox,
                           const Real&      a_scale,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXCBN(Real,a_destbox,a_destcomp,a_numcomp,a_src,a_srcbox,a_srccomp)
  {
    thisR += a_srcR * a_scale;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::operator -= (Real a_r)
{
  return operator += (-a_r);
}

FArrayBox& FArrayBox::operator -= (const FArrayBox& a_x)
{
  ForAllThisXC(Real,a_x)
  {
    thisR -= a_xR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::minus(const FArrayBox& a_x)
{
  return operator -= (a_x);
}

FArrayBox& FArrayBox::minus(const FArrayBox& a_src,
                           int               a_srccomp,
                           int               a_destcomp,
                           int               a_numcomp)
{
  ForAllThisBNNXC(Real,m_domain,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR -= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::minus(const FArrayBox& a_src,
                            const Box&       a_subbox,
                            int              a_srccomp,
                            int              a_destcomp,
                            int              a_numcomp)
{
  assert(m_domain.contains(a_subbox));

  ForAllThisBNNXC(Real,a_subbox,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR -= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::minus(const FArrayBox& a_src,
                            const Box&       a_srcbox,
                            const Box&       a_destbox,
                            int              a_srccomp,
                            int              a_destcomp,
                            int              a_numcomp)
{
  ForAllThisBNNXCBN(Real,a_destbox,a_destcomp,a_numcomp,a_src,a_srcbox,a_srccomp)
  {
    thisR -= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::operator *= (Real a_r)
{
  ForAllThis(Real)
  {
    thisR *= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::mult(Real a_r)
{
  return operator *= (a_r);
}

FArrayBox& FArrayBox::mult(Real a_r,
                           int  a_comp,
                           int  a_numcomp)
{
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR *= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::mult(Real       a_r,
                           const Box& a_subbox,
                           int        a_comp,
                           int        a_numcomp)
{
  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR *= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::operator *= (const FArrayBox &a_x)
{
  ForAllThisXC(Real,a_x)
  {
    thisR *= a_xR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::mult(const FArrayBox& a_x)
{
  return operator *= (a_x);
}

FArrayBox& FArrayBox::mult(const FArrayBox& a_src,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXC(Real,m_domain,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR *= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::mult(const FArrayBox& a_src,
                           const Box&       a_subbox,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  assert(m_domain.contains(a_subbox));

  ForAllThisBNNXC(Real,a_subbox,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR *= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::mult(const FArrayBox& a_src,
                           const Box&       a_srcbox,
                           const Box&       a_destbox,
                           int              a_srccomp,
                           int              a_destcomp,
                           int              a_numcomp)
{
  ForAllThisBNNXCBN(Real,a_destbox,a_destcomp,a_numcomp,a_src,a_srcbox,a_srccomp)
  {
    thisR *= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::operator /= (Real a_r)
{
  ForAllThis(Real)
  {
    thisR /= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::divide(Real a_r)
{
  return operator /= (a_r);
}

FArrayBox& FArrayBox::divide(Real a_r,
                             int  a_comp,
                             int  a_numcomp)
{
  ForAllThisNN(Real,a_comp,a_numcomp)
  {
    thisR /= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::divide(Real       a_r,
                             const Box& a_subbox,
                             int        a_comp,
                             int        a_numcomp)
{
  ForAllThisBNN(Real,a_subbox,a_comp,a_numcomp)
  {
    thisR /= a_r;
  } EndFor

  return *this;
}

FArrayBox& FArrayBox::operator /= (const FArrayBox &a_x)
{
  ForAllThisXC(Real,a_x)
  {
    thisR /= a_xR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::divide(const FArrayBox& a_x)
{
  return operator /= (a_x);
}

FArrayBox& FArrayBox::divide(const FArrayBox& a_src,
                             int              a_srccomp,
                             int              a_destcomp,
                             int              a_numcomp)
{
  ForAllThisBNNXC(Real,m_domain,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR /= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::divide(const FArrayBox& a_src,
                             const Box&       a_subbox,
                             int              a_srccomp,
                             int              a_destcomp,
                             int              a_numcomp)
{
  assert(m_domain.contains(a_subbox));
  ForAllThisBNNXC(Real,a_subbox,a_destcomp,a_numcomp,a_src,a_srccomp)
  {
    thisR /= a_srcR;
  } EndForTX

  return *this;
}

FArrayBox& FArrayBox::divide(const FArrayBox& a_src,
                             const Box&       a_srcbox,
                             const Box&       a_destbox,
                             int              a_srccomp,
                             int              a_destcomp,
                             int              a_numcomp)
{
  ForAllThisBNNXCBN(Real,a_destbox,a_destcomp,a_numcomp,a_src,a_srcbox,a_srccomp)
  {
    thisR /= a_srcR;
  } EndForTX

  return *this;
}

void FArrayBox::performCopy(const BaseFab<Real>& a_src,
                            const Box&           a_srcbox,
                            int                  a_srccomp,
                            const Box&           a_destbox,
                            int                  a_destcomp,
                            int                  a_numcomp)
{
  //TFAB_performCopy.start();
#ifdef OPT_MYPERFORMCOPY
  BaseFab<Real>::myperformCopy(a_src, a_srcbox, a_srccomp, a_destbox, a_destcomp, a_numcomp);
  //BaseFab<Real>::myperformCopyDiag(a_src, a_srcbox, a_srccomp, a_destbox, a_destcomp, a_numcomp);
#else
  BaseFab<Real>::performCopy(a_src, a_srcbox, a_srccomp, a_destbox, a_destcomp, a_numcomp);
#endif
  //TFAB_performCopy.stop();

  // FArrayBox& dest = *this;
  // FORT_PERFORMCOPY(CHF_FRA(dest),
  //                  CHF_CONST_FRA(a_src),
  //                  CHF_BOX(a_destbox),
  //                  CHF_BOX(a_srcbox),
  //                  CHF_CONST_INT(a_srccomp),
  //                  CHF_CONST_INT(a_destcomp),
  //                  CHF_CONST_INT(a_numcomp));
}
