/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifdef USE_MYEXCHANGE_IN_POISSONOP

#include "LevelData.H"
#include "parstream.H"
#include <cstdio>

#ifdef MPI
#include <mpi.h>
#endif

#include "Timer.H"
extern Timer Everything;
#ifdef COUNT_PEX
extern Timer TPlevelGSRB_ex;
#endif
//Timer TEX1("PendingSends1",    Everything, 9);
//Timer TEX2("allocateBuffers",  Everything, 9);
//Timer TEX3("writeSendData",    Everything, 9);
//Timer TEX4("postReceivesToMe", Everything, 9);
//Timer TEX5("postSendsFromMe",  Everything, 9);
//Timer TEX6("local copy",       Everything, 9);
//Timer TEX7("unpackRecvToMe",   Everything, 9);

////int NN=0;
#ifdef DEBUG_STUFF
char stuff[90]; //temporarily here, obviously!
#endif

//Timer TEX_postSendsFromMe_nonMPI   ("nonMPI",    TEX5);
//Timer TEX_postSendsFromMe_MPI_Isend("MPI_Isend", TEX5);

//Timer TEX_unpackReceivesToMe_compPendSends("PendingSends2", TEX7);
//Timer TEX_unpackReceivesToMe_MPI_Waitall  ("MPI_Waitall",   TEX7);
//Timer TEX_unpackReceivesToMe_linearIn     ("linearIn",      TEX7);

#ifdef COUNT_PEX
extern unsigned int pex_Nsends[NPEX];
#endif

template<class T>
void LevelData<T>::myexchange(const Interval& comps,
                              const Copier& copier) {

  mymakeItSo(comps, *this, *this, comps, copier);
}

template<class T>
void LevelData<T>::mymakeItSo(const Interval&   a_srcComps,
                              const LevelData<T>& a_src,
                              BoxLayoutData<T>& a_dest,
                              const Interval&   a_destComps,
                              const Copier&     a_copier,
                              const LDOperator<T>& a_op) const {

  //numSends = m_toMe.size();
  //numReceives = m_fromMe.size();
  //if (numSends != m_toMe.size())
  //pout() << "sends noteq    numSends="
  //     << numSends <<    " m_toMe.size  =" << m_toMe.size() << endl;
  //if (numReceives != m_fromMe.size())
  //pout() << "recvs noteq numReceives="
  //     << numReceives << " m_fromMe.size=" << m_fromMe.size() << endl;

#ifdef MPI
#ifdef DEBUG_STUFF
  int M=0;
  for(CopyIterator it(a_copier, CopyIterator::LOCAL); it.ok(); ++it) M++;
  pout() << " ----------------------------------------" << endl;
  //sprintf(stuff, "exchange rank%03d numSends=%4d(%4d) numReceives=%4d(%4d) last_numSends=%d numLocalCopies=%5d\n",
  //    procID(), m_toMe.size(),numSends, m_fromMe.size(), numReceives, last_numSends, M); pout() << stuff;
  sprintf(stuff, "exchange rank%03d numSends=%4d(%4d) numReceives=%4d(%4d) numLocalCopies=%5d\n",
    procID(), m_toMe.size(), numSends, m_fromMe.size(), numReceives, M); pout() << stuff;
  //pout() << " last_numSends = " << last_numSends << " NN=" << NN << endl;
#endif

  //calledTooMuch TEX2.start();
  static Copier* lastCopier=NULL;
  if(T::preAllocatable() == 2 || !a_copier.bufferAllocated() ||
     (m_fromMe.size() + m_toMe.size() == 0) ||lastCopier != &a_copier){
    allocateBuffers(a_src, a_srcComps, a_dest, a_destComps, a_copier, a_op);
    a_copier.setBufferAllocated(true);
  }
  lastCopier = (Copier*)(&a_copier);
  //calledTooMuch TEX2.stop();
#endif

#ifdef DEBUG_STUFF
#ifdef MPI
  sprintf(stuff, "1xchange rank%03d numSends=%4d(%4d) numReceives=%4d(%4d) numLocalCopies=%5d\n",
    procID(), m_toMe.size(), numSends, m_fromMe.size(), numReceives, M); pout() << stuff;
  //pout() << " last_numSends = " << last_numSends << " NN=" << NN << endl;
#endif
#endif

  //calledTooMuch TEX3.start();
  writeSendDataFromMeIntoBuffers(a_src, a_srcComps, a_op);
  //calledTooMuch TEX3.stop();

  //sprintf(stuff, "2xchange rank%03d numSends=%4d(%4d) numReceives=%4d(%4d) numLocalCopies=%5d\n",
  //    procID(), m_toMe.size(),numSends, m_fromMe.size(), numReceives,M); pout() << stuff;

#ifdef MPI
  if (m_toMe.size() > 0) {
    //calledTooMuch TEX4.start();
    mypostReceivesToMe(); // all non-blocking
    //calledTooMuch TEX4.stop();
  }

  if (m_fromMe.size() > 0) {
    //calledTooMuch TEX5.start();
    mypostSendsFromMe();  // all non-blocking
    //calledTooMuch TEX5.stop();
  }
#endif

  //calledTooMuch TEX6.start();
  // perform local copy
  for(CopyIterator it(a_copier, CopyIterator::LOCAL); it.ok(); ++it) {

    const MotionItem& item = it();
    // i was wondering if there were any copies happening twice...
    //char crap[90];
    //sprintf(crap, "fromRegion %5d
    //pout() << " fromRegion " << item.fromRegion
    //     << " toRegion " << item.toRegion
    //     << endl;

    a_op.op(a_dest[item.toIndex], item.fromRegion,
            a_destComps,
            item.toRegion,
            a_src[item.fromIndex],
            a_srcComps);
  }
  //pout() << " qqqqqqqqqqqqqqqqqqqqqqqqqq" << endl;
  //calledTooMuch TEX6.stop();

  //calledTooMuch TEX1.start();
  mycompletePendingSends();
  //calledTooMuch TEX1.stop();

  //calledTooMuch TEX7.start();
  myunpackReceivesToMe(a_dest, a_destComps, a_op); // nullOp in uniprocessor mode
  //calledTooMuch TEX7.stop();
}

#ifdef MPI
template<class T>
void LevelData<T>::mycompletePendingSends() const {

  //pout() << " mycompletePendingSends" << endl;

  if(numSends > 0){
#ifdef DEBUG_STUFF
    sprintf(stuff, "rank%03d completePend numSends=%5d delete sends\n",
            procID(), numSends); pout() << stuff;
#endif
    MPI_Waitall(numSends, m_sendRequests, m_sendStatus);
    delete[] m_sendRequests;
    delete[] m_sendStatus;
  }
  numSends = 0;
}

template<class T>
void LevelData<T>::mypostReceivesToMe() const {

  numReceives = m_toMe.size();
#ifdef DEBUG_STUFF
  pout() << " mypostReceivesToMe numSends = " << numSends << endl;
#endif

  if(numReceives > 1) {
    for(unsigned int i=m_toMe.size()-1; i>0; --i)  {
#ifdef DEBUG_STUFF
      sprintf(stuff, "i%4d: ", i); pout() << stuff;
#endif
      if(m_toMe[i].procID == m_toMe[i-1].procID) {
        numReceives--;
        m_toMe[i-1].size+=m_toMe[i].size;
#ifdef DEBUG_STUFF
        sprintf(stuff, " numReceives=%4d  prev m_toMe size=%4d  m_toMe size=%4d",
                numReceives, m_toMe[i-1].size, m_toMe[i].size);
        pout() << stuff;
#endif
        m_toMe[i].size = 0;
      }
#ifdef DEBUG_STUFF
      pout() << endl;
#endif
    }
  }
  m_receiveRequests = new MPI_Request[numReceives];
  m_receiveStatus = new MPI_Status[numReceives];
#ifdef DEBUG_STUFF
  sprintf(stuff, " %4d new recvs\n", numReceives); pout() << stuff;
#endif

  unsigned int next=0;
  for(int i=0; i<numReceives; ++i) {

    const bufEntry& entry = m_toMe[next];
#ifdef DEBUG_STUFF
    sprintf(stuff, "rank%03d recv mesg of size %7d from rank%03d i=%5d next=%6d\n",
            procID(), entry.size, entry.procID, i, next); pout() << stuff;
#endif

    MPI_Irecv(entry.bufPtr, entry.size, MPI_BYTE, entry.procID,
              0, Chombo_MPI::comm, m_receiveRequests+i);
    ++next;
    while(next < m_toMe.size() && m_toMe[next].size == 0) ++next;
  }

}

template<class T>
void LevelData<T>::mypostSendsFromMe() const {

  //pout() << " mypostSendsFromMe" << endl;
  //TEX_postSendsFromMe_nonMPI.start();
  numSends = m_fromMe.size();
#ifdef DEBUG_STUFF
  pout() << " mypostSendsFromMe numSends = " << numSends << endl;
#endif
  if(numSends > 1) {
    for(unsigned int i=m_fromMe.size()-1; i>0; --i) {
#ifdef DEBUG_STUFF
      sprintf(stuff, "i%4d: ", i); pout() << stuff;
#endif
      if(m_fromMe[i].procID == m_fromMe[i-1].procID) {
        numSends--;
        m_fromMe[i-1].size+=m_fromMe[i].size;
#ifdef DEBUG_STUFF
        sprintf(stuff, " numSends=%4d  prev m_fromMe size=%4d  m_fromMe size=%4d",
                numSends, m_fromMe[i-1].size, m_fromMe[i].size);
        pout() << stuff;
#endif
        m_fromMe[i].size = 0;
      }
#ifdef DEBUG_STUFF
      pout() << endl;
#endif
    }
  }

  m_sendRequests = new MPI_Request[numSends];
  m_sendStatus = new MPI_Status[numSends];

  //TEX_postSendsFromMe_nonMPI.stop();
#ifdef DEBUG_STUFF
  sprintf(stuff, " new sends numSends=%d\n", numSends); pout() << stuff;
#endif

  //TEX_postSendsFromMe_MPI_Isend.start();
  unsigned int next=0;
  for(int i=0; i<numSends; ++i)  {

    const bufEntry& entry = m_fromMe[next];
#ifdef DEBUG_STUFF
    sprintf(stuff, "rank%03d send mesg of size %7d  to  rank%03d i=%5d next=%6d\n",
            procID(), entry.size, entry.procID, i, next); pout() << stuff;
#endif
    MPI_Isend(entry.bufPtr, entry.size, MPI_BYTE, entry.procID,
              0, Chombo_MPI::comm, m_sendRequests+i);
    ++next;
    while(next < m_fromMe.size() && m_fromMe[next].size == 0) ++next;
  }
  //TEX_postSendsFromMe_MPI_Isend.stop();

#ifdef COUNT_PEX
  pex_Nsends[TPlevelGSRB_ex.Count()] = numSends;
#endif
  //pout() << " numSends=" << numSends; //noel

}

template<class T>
void LevelData<T>::myunpackReceivesToMe(BoxLayoutData<T>& a_dest,
                                        const Interval&   a_destComps,
                                        const LDOperator<T>& a_op) const {

   if(numReceives > 0){

#ifdef DEBUG_STUFF
    sprintf(stuff, "rank%03d unpacking numReceives=%5d delete recvs\n",
            procID(), numReceives); pout() << stuff;
#endif

    MPI_Waitall(numReceives, m_receiveRequests, m_receiveStatus);

    for(unsigned int i=0; i<m_toMe.size(); ++i) {
      const bufEntry& entry = m_toMe[i];
      a_op.linearIn(a_dest[entry.item->toIndex], entry.bufPtr, entry.item->toRegion, a_destComps);
    }

    delete[] m_receiveRequests;
    delete[] m_receiveStatus;
  }
  numReceives = 0;
}

#else //MPI

// uniprocessor version of all these nullop functions.
template<class T>
void LevelData<T>::mycompletePendingSends() const
{;}

template<class T>
void LevelData<T>::mypostSendsFromMe() const
{;}

template<class T>
void LevelData<T>::mypostReceivesToMe() const
{;}

template<class T>
void LevelData<T>::myunpackReceivesToMe(BoxLayoutData<T>& a_dest,
                                        const Interval&   a_destComps,
                                        const LDOperator<T>& a_op) const
{;}

#endif //MPI

#include "FluxBox.H"
#include "FArrayBox.H"
template class LevelData<BaseFab<int> >;
template class LevelData<FluxBox>;
template class LevelData<FArrayBox>;

//#define PARTICLE_CODE
//#ifdef PARTICLE_CODE
//#include "BaseFab.H"
//#include "../ParticleTools/BinFab.H"
//#include "../ParticleTools/BinItem.H"
//#include "../../../../TimeddMartin/particle-IAMR/particles/DragParticle.H"
//template class LevelData<BinFab<DragParticle> >;
//template class LevelData<BinFab<BinItem> >;
//#endif

#endif // USE_MYEXCHANGE_IN_POISSONOP
