/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _LOHISIDE_H_
#define _LOHISIDE_H_

/// Encapsulation of high or low side.
/**
   Encapsulation of high or low side.
*/
class Side
{
public:
  ///
  /**
     Lo is the low side.<br>
     Hi is the high side.
  */
  enum LoHiSide {Invalid=-1, Lo=0, Hi, NUMSIDES};

  /**
     Constructor.
   */
  Side()
  {}

  /**
     Destructor.
   */
  ~Side()
  {}

  ///
  /**
     Returns the other side.  That is, swap Hi for Lo and vice versa.
  */
  friend Side::LoHiSide flip(Side::LoHiSide a_side);

  ///
  static Side::LoHiSide flip(Side::LoHiSide a_side);

  ///
  /**
     Returns -1 for Lo Side, +1 for Hi Side.
  */
  friend
  int sign(Side::LoHiSide a_side);
};

/// Iterator for low and high side.
/**
   SideIterator iterates over the values Side::Lo and Side::Hi.
   There are no adjustable parameters whatsoever.
*/
class SideIterator
{
public:
  ///
  /**
      Default, and only, constructor.
  */
  SideIterator();

  ///
  /**
     Default destructor.
  */
  ~SideIterator()
  {}

  /**
     \name Modification functions
  */
  /*@{*/

  ///
  /**
     Set the iterator to the first side.
  */
  void begin();

  ///
  /**
     Set the iterator to the first side.
  */
  void reset()
  {
    begin();
  }

  ///
  /**
     Advance the iterator to the next face.
  */
  void next();

  ///
  /**
     Advance the iterator to the next face.
  */
  void operator ++ ();

  /*@}*/

  /**
     \name Access functions
  */
  /*@{*/

  ///
  /**
     Returns the value of the side for the current iteration.
  */
  Side::LoHiSide operator () () const;

  ///
  /**
     Returns <tt>true</tt> if there is a next side, <tt>false</tt> if not.
  */
  bool ok() const;


private:
  int m_current;
};

#endif
