/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <iostream>
#include <list>
#include <set>
using std::cout;

#include "parstream.H"

#include "DataIterator.H"
#include "Misc.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "LayoutIterator.H"

#include "memtrack.H"
#include <cstdio>
#include "Timer.H"
#include "memusage.H"
extern Timer Everything;
//Timer TimeLoadBalanceAll("LoadBalance", Everything, 342);
// Timer lb1("lb1", TimeLoadBalanceAll);
// Timer lb2("lb2", TimeLoadBalanceAll);
// Timer lb3("lb3", TimeLoadBalanceAll);

// Timer lb3a("lb3a", lb3);
// Timer lb3b("lb3b", lb3);
// Timer lb3c("lb3c", lb3);
// Timer lb3d("lb3d", lb3);
// Timer lb3e("lb3e", lb3);
// Timer lb3f("lb3f", lb3);
extern int GlobalLevel;

static int LBCounter=0;
int globalLoad;

// local prototypes
int
min_element( const Vector<long>& Vect );
void
min_max_elements( int& imin ,int& imax ,const Vector<long>& Vect );

void
min_max_elements( int& imin ,int& imax ,const Vector<long long>& Vect );

// Local class definition (needed to sort loads)
class Load
{
public:
  Load():load(0) ,grid_index(0){}
  bool operator < (const Load& rhs) const {return load < rhs.load;}

  long load;      //actual load on this box
  int grid_index; //link to Grids[]
};

// Code:

///
// This version takes a Vector<BoxLayout> and builds a matching Vector of
// Vector<Box> and uses it to call the full version.
///
int
LoadBalance( Vector<BoxLayout>&    Grids            //in-out: input grids to balance
             ,Real&                 effRatio         //output: ratio of min load
             ,const Vector<Vector<long> >&  ComputeLoads     //input: computational cost
             ,const Vector<int>&           RefRatios        //input: refinement ratio
             ,int nProc
             )
{
  Vector<Vector<Box> > boxes( Grids.size() );

  for( int i = 0; i < Grids.size(); ++i )
    {
      LayoutIterator j = Grids[i].layoutIterator();
      int p;
      for( j.reset() ,p = 0; j.ok(); ++j ,++p )
        {
          boxes[i][p] = Grids[i][j()];
        }
    }

  Vector<Vector<int> > procIDs( Grids.size() );
  int status = LoadBalance( procIDs ,effRatio
                            ,boxes ,ComputeLoads ,RefRatios, nProc );

  if( status < 0 ) return status;  //LoadBalance() failed

  for( int i = 0; i < Grids.size(); ++i )
    {
      LayoutIterator j = Grids[i].layoutIterator();
      int p;
      for( j.reset() ,p = 0; j.ok(); ++j ,++p )
        {
          Grids[i].setProcID( j() ,procIDs[i][p] );
        }
    }

  return status;
}

// one of my loadbalance's  when u want to play with ProblemDomain
int LoadBalance(Vector<int>& procs, const Vector<Box>& a_boxes,
                const ProblemDomain& pdom,
                int a_numProc) {

  //pout() << " my LB " << endl;

  // modify boxes first.  for example:
  // grow all boxes by one before balancing...
  Vector<Box> boxes(a_boxes.size());
  for(int i = 0; i < a_boxes.size(); ++i) {
    //pout() << i << " a_boxes[i]=" << a_boxes[i] << endl;
    //boxes[i] = grow(a_boxes[i], 1);
    //boxes[i] = grow(a_boxes[i], 4);
    //boxes[i] &= pdom;
    boxes[i] = a_boxes[i];
  }

  LoadBalance(procs, boxes, a_numProc);
  return 1;
}

///
// This version takes a single BoxLayout (i.e. one level)
// and uses the box volumes to construct the compute loads
// and calls the full version.
///
int LoadBalance(Vector<int>& procs, const Vector<Box>& a_boxes,
                int a_numProc)
{
  Vector<long> computeLoads(a_boxes.size());

  for (int i = 0; i < a_boxes.size(); ++i)
    {
      computeLoads[i] = a_boxes[i].numPts();
    }

  int status;
  status = LoadBalance(procs, computeLoads, a_boxes, a_numProc);

  // save this older commented code
  //   Vector<Vector<Box> > layouts(1, boxes);
  //   Vector<Vector<long> > computeLoads(1, Vector<long>(boxes.size()));
  //   Vector<int> refRatios(1,1);
  //   Vector<Vector<int> > assignments(1,Vector<int>(boxes.size(), -1));
  //   Real effRatio;

  //   for(int index = 0; index < boxes.size(); ++index)
  //     {
  //       computeLoads[0][index] = layouts[0][index].numPts();
  //     }

  //   int ret = LoadBalance(assignments, effRatio, layouts, computeLoads, refRatios);

  //   if(ret == 0)
  //     procs = assignments[0];

  return status;
}

///
// This version takes compute loads directly and does the "simple"
// load balance algorithm (modified knapsack algorithm).
// Boxes are also included, but only used if SWAP_SAME_SIZES_BOXES
// is enabled
///
int LoadBalance(Vector<int>&        procs,
                const Vector<long>& a_computeLoads,
                const Vector<Box>&  a_boxes,
                const int           a_numProc)
{
  pout() << "  LoadBalance  sizes: a_computeLoads=" << a_computeLoads.size()
         << " a_boxes=" << a_boxes.size()
         << " a_numProc=" << a_numProc
         << endl;
  //TimeLoadBalanceAll.start();
  //lb1.start();
#ifdef OPT_BVS_LOADBALANCE_NEW

  // Phase 1  modified knapsack algorithm.  this one doesn't use
  // load sorting followed by round robin.  This one does bin packing
  // first by finding vector divisors, then does regular knapsack
  // optimization.

  int status = 0;
  procs.resize(0);
  procs.resize(a_computeLoads.size(), 0);

  int Nprocs = procs.size();

  if(a_numProc == 1 || Nprocs == 1) {
    //lb1.stop();
    //TimeLoadBalanceAll.stop();
    return 0;
  }

  // first, compute 'total load'
  double totalLoad = 0; long long tload = 0;
  for(int index = 0; index < Nprocs; ++index)
    {
      totalLoad += a_computeLoads[index];
      tload += a_computeLoads[index];
      procs[index] = -index;
    }

  // figure out roughly single processor load goal
  double goal = totalLoad/a_numProc;

  // debugging prints
  pout() << "--- LoadBalance L" << GlobalLevel
         << " --- Total Number of Loads on this Level: "
         << Nprocs << std::endl;
  //pout() << " Box sizes: " << std::endl;
  //for(int i=0; i < Nprocs; i++) pout() <<" "<<a_computeLoads[i];
  //pout() << std::endl;

  // now, first bin assignments

  // change goal to dynamicgoal (ndk)
  // We desire equal amount of cells per processor, which
  // is the total number of cells divided by the number of processors: ie the goal.
  // But because the cells are grouped into boxes of sometimes very different sizes,
  // and because we assume that the boxes are in a locality-preferable vector,
  // using a constant goal can allow for several consecutive processors to be
  // assigned a number of cells _less_ than (or _more_ than) the goal.
  // This can lead to the very last processor receiving an abnormally large
  // (or small) workload.  Changing the algorithm to compute a new dynamic
  // goal after each processor-box assignment eliminates this problem.
  // However, there are still improvements that can be made when the
  // number of boxes to be farmed out is not much larger than the total number
  // of processors doing the work.  (ndk)

  Vector<long long> loads(a_numProc, 0);
  int bin = 0;
  double dynamicLoad = totalLoad;
  double dynamicgoal = goal;
  double load = 0;
  long long numP;
  //pout() << " first time thru: dynamicLoad =" << dynamicLoad
  //       << " dynamicgoal = " << dynamicgoal << std::endl;

  for(int i=0; i < Nprocs; i++)
    {
      numP = a_computeLoads[i];
      if(bin != a_numProc-1)  // NOT last rank
        {
          load += numP;
          if(load >= dynamicgoal)
            {
              if( ( load - dynamicgoal ) > numP/2 )
                {
                  procs[i] = bin+1;
                  loads[bin+1] += numP;
                  load = numP;
                }
              else
                {
                  procs[i] = bin;
                  loads[bin] += numP;
                  load = 0;
                }
#ifdef OPT_LOADBALANCE_DYNAMICGOAL
              dynamicLoad -= (double)loads[bin];
              dynamicgoal = dynamicLoad/(double)(a_numProc-(bin+1));
#endif
              bin++;
            }
          else
            {  // last rank
              loads[bin] += numP;
              procs[i] = bin;
            }
        }
      else
        {
          procs[i] = bin;
          loads[bin] += numP;
        }

    }
  globalLoad = loads[procID()];

  //int NminNumberBoxesToDiagPrint = a_numProc + a_numProc + a_numProc/2;
  //int NminNumberBoxesToDiagPrint = a_numProc + a_numProc/2;
  int NminNumberBoxesToDiagPrint = 0;
  //pout()<<"tload = "<<(double)tload<< std::endl;
  //pout() << " loads: ";
  //for(int i=0; i<numProc(); i++) pout()<<" "<<(long)loads[i];
  //pout()<<std::endl;

  //Phase 2 of loadbalance, actual knapsack swapping to improve efficiency
  // first , scan for highest load processor
  long long mload = loads[0];
  for(int i=1; i<loads.size(); ++i) mload = Max(mload, loads[i]);

  //parallel inefficiency: sum of idle time of processors waiting for max load processor
  // dividedd by total load
  double eff = 1 - ((double)(mload*a_numProc - totalLoad))/totalLoad;
  if (Nprocs > NminNumberBoxesToDiagPrint) {
    pout() << "loadbalance efficiency (before knapsack )= "<<eff
           << "  Nboxes=" << Nprocs
           <<std::endl;
  }

#define PRINT_EXTRA_LB_STUFF
#ifdef PRINT_EXTRA_LB_STUFF
  // extra stuff:
#ifdef MPI
  if (procID() == uniqueProc (SerialTask::compute) ) {
#endif

    // only print if there's something interesting to see...
    if (Nprocs > NminNumberBoxesToDiagPrint) {

      for(int i=0; i<a_numProc; i++) {

        int NboxCount=0;
        for(int j=0; j < Nprocs; j++) {
          if (procs[j] == i) NboxCount++;
        }

        int bigbox=0, Nboxes=0, totalpoints=0;
        for(int j=0; j < Nprocs; j++) {
          if (procs[j] == i) {
            Nboxes=a_computeLoads[j];
            totalpoints += Nboxes;
            if (Nboxes > bigbox) bigbox=Nboxes;
          }
        }
        bool allSameSize=false;
        // need to fix this -- cuz this is only per proc and
        // what i want is no printintg if every box on all procs
        // are the same size...
        //if (totalpoints/NboxCount == Nboxes) allSameSize=true;

        if (allSameSize) { // don't print
          pout() << NboxCount << "  boxes of same size" << Nboxes << endl;
        } else {
          char tempstring2[120];
          sprintf(tempstring2, "rank%03d load=%-8ld boxes(%2d) [%6d] ",
                  i, (long int)loads[i], NboxCount, bigbox);
          pout() << tempstring2;
          //pout() << "rank" << i << "  load=" << (long)loads[i] << " boxes("
          //     << NboxCount << ")=";
          for(int j=0; j < Nprocs; j++) {
            if (procs[j] == i) {
              Nboxes=a_computeLoads[j];
              pout() << Nboxes << " ";
            }
          }
          pout() << std::endl;
        }
      }
    }
#ifdef MPI
  }
#endif
#endif

  // for xmgrace
//   int N=0;
//   int currentRank = 0;
//   for(int j=0; j < Nprocs; j++) {
//     pout() << N << " " << boxes[j].numPts() << std::endl;
//     N++;
//     if (procs[j] == currentRank+1) {
//       pout() << N << " " << (long)loads[currentRank] << std::endl;
//       N++;
//       currentRank++;
//     }
//   }

#else
  Vector<Vector<Box> > layouts(1, boxes);
  Vector<Vector<long> > computeLoads(1, Vector<long>(Nprocs));
  Vector<int> refRatios(1,1);
  Vector<Vector<int> > assignments(1,Vector<int>(Nprocs, -1));
  Real effRatio;

  for(int index = 0; index < Nprocs; ++index)
    {
      computeLoads[0][index] = layouts[0][index].numPts();
    }

  int ret = LoadBalance(assignments, effRatio, layouts, computeLoads, refRatios);

  if(ret == 0)
    procs = assignments[0];
#endif
  //lb1.stop();

#ifdef OPT_BVS_LOADBALANCE_PHASE3
  NOEL!
  before using this, fix computationalLoad[i] conversion

  //lb2.start();
  // Phase 3 of loadbalance:  exhanges of identical sized boxes in order
  // to improve locality.  Need to handle periodicity somehow also, but
  // that will have to wait for loadbalance to be redesigned.

  bool sorted = true;
  for(int i=1; i<Nprocs; i++)
    {
      if(a_boxes[i] < a_boxes[i-1]) sorted = false;
    }
  Vector<std::list<int> > neighbours(Nprocs);

  for(int i=0; i<Nprocs; i++)
    {
      Box b = a_boxes[i];
      b.grow(1);
      std::list<int>& list = neighbours[i];
      for(int ii=0; ii<Nprocs; ii++)
        {
          if(ii == i); //don't bother with self intersect
          else
            {
              const Box& bb = a_boxes[ii];

              if(b.intersects(bb))
                {
                  list.push_front(ii);

                }
              if(sorted && bb.smallEnd(0) > b.bigEnd(0)) ii=Nprocs;
            }
        }
    }
  //lb2.stop();
  //lb3.start();
  int swapcount = 6;
  int diffBoxes;
  int sameBoxes;
  int iter = 0;
#ifdef REDUCE_MAXSWAPITERATIONS
  const int maxSwapIterations = 2;
#else
  const int maxSwapIterations = 20;
#endif
  while (iter < maxSwapIterations && swapcount > 0)
    {
      //lb3a.start();
      swapcount = 0;
      diffBoxes = 0;
      sameBoxes = 0;
      iter++;

      for(int i=0; i<neighbours.size(); i++)
        {
          //lb3b.start();
          std::list<int>& list = neighbours[i];
          for(int ii=i+1; ii<neighbours.size(); ii++)
            {
              //lb3c.start();
              int p = procs[i];
              int pii = procs[ii];
              if(a_boxes[i].numPts() != a_boxes[ii].numPts()) diffBoxes++; // don't swap
              else if(p == pii); //do nothing, no swaps on same proc
              else
                {
                  sameBoxes++;
                  std::list<int>& listii = neighbours[ii];

                  std::list<int>::iterator it;
                  // count : number of "local" exchanges before swap
                  // countafter: number of "local" exchanges after swap
                  int count = 0; int countafter = 0;
                  //lb3d.start();
                  for(it = list.begin();it != list.end(); it++)
                    {
                      if(procs[*it] == p)
                        count++;
                      if(procs[*it] == pii)
                        countafter++;
                    }
                  //lb3d.stop();
                  //lb3e.start();
                  for(it = listii.begin();it != listii.end(); it++)
                    {
                      if(procs[*it] == pii)
                        count++;
                      if(procs[*it] == p)
                        countafter++;
                    }
                  //lb3e.stop();
                  //lb3f.start();
                  if(countafter > count)
                    {
                      swapcount++;
                      procs[i] = pii;
                      procs[ii] = p;
                      //pout()<<"count : "<<count<<" countafter :"<<countafter
                      //    <<std::endl;
                    }
                  //lb3f.stop();
                }
              //lb3c.stop();
            }
          //lb3b.stop();
        }
      //pout()<<"Boxes: "<<Nprocs<<" same: "<<sameBoxes
      //    <<" swaps: "<<swapcount<<std::endl;
      //lb3a.stop();
    }
  //lb3.stop();
#endif

  //TimeLoadBalanceAll.stop();

#if defined(TIMER) && defined (ENABLE_MEMORY_TRACKING)
  Real memtrackCurrentMemory;
  Real memtrackPeakMemory;
  memtrackStamp(memtrackCurrentMemory, memtrackPeakMemory);
  char tempstring[100];
  sprintf(tempstring,
          " after LBal%03d  mem=%-10.2f  memtrack=%-10.2f  memtrack peak=%-10.2f\n",
          LBCounter, get_memory_usage_from_OS(), memtrackCurrentMemory, memtrackPeakMemory);
  pout() << tempstring;
  LBCounter++;
  //pout() << " LoadBalance mem= " << get_memory_usage_from_OS()
  //     << " mt current=" << memtrackCurrentMemory
  //     << " mt peak=" << memtrackPeakMemory
  //     << endl;
#endif
  return status;
}

///
// This version does the real work.
///
int
LoadBalance(Vector<int>&          a_procAssignments
            ,Real&                a_effRatio
            ,const Vector<Box>&   a_grids
            ,const Vector<long>&  a_computeLoads)
{
  Vector<Vector<Box> >   grids(1, a_grids);
  Vector<Vector<long> >  computeLoads(1, a_computeLoads);
  Vector<int>            refRatios(1,2);
  Vector<Vector<int> > procs;
  int nproc = numProc();
  int retval = LoadBalance(procs, a_effRatio, grids, computeLoads,
                           refRatios, nproc);

  a_procAssignments = procs[0];
  return retval;
}

int
LoadBalance(Vector<Vector<int> >& procAssignments  //output: processor number
            ,Real&                 effRatio         //output: ratio of min load
            ,const Vector<Vector<Box> >&  Grids            //input: meshes to balance
            ,const Vector<Vector<long> >& ComputeLoads     //input: computational cost
            ,const Vector<int>&           RefRatios        //input: refinement ratio
            ,int nProc                            // number of procs to assugn to
            )
{
  // local variables
  Real eff_ratio; // efficiency ratio on a level
  int status = 0; // return code

  // Validate inputs
  if( Grids.size() != ComputeLoads.size() )
    { return -1011; }
  if( Grids.size() != RefRatios.size() )
    { return -1013; }
  for( int lvl=0; lvl<Grids.size(); ++lvl )
    {
      if( Grids[lvl].size() != ComputeLoads[lvl].size() )
        { return -1012; }
    }

  // set the number of elements in the output vector to the number
  // of levels and the number of elements in each element to the
  // number of boxes on each level and set the value of each
  // element to zero
  procAssignments.resize( Grids.size() );
  for( int lvl=0; lvl<Grids.size(); ++lvl )
    {
      procAssignments[lvl].resize( Grids[lvl].size(),0 );
    }

  // check for special case of all loads on 1 processor
  if( nProc == 1 )
    {
      for( int lvl=0; lvl<Grids.size(); ++lvl )
        {
          for( int i=0; i<Grids[lvl].size(); ++i )
            {
              procAssignments[lvl][i] = 0;
            }
        }
      effRatio = 1.0;
      status = 0;
    }
  else
    {
      // general case: loads on more than one processor
      effRatio = 1.0;

      // balance each level separately
      for( int lvl=0; lvl<Grids.size(); ++lvl )
        {
          // first, build the load structure and sort by compute_load
          Vector<Load> loads( Grids[lvl].size() );
          for( int i=0; i<Grids[lvl].size(); ++i )
            {
              loads[i].load = ComputeLoads[lvl][i];
              loads[i].grid_index = i;
            }
          //          std::sort( loads.begin() ,loads.end() );
          loads.sort();
          // do the initial assignments by sequentially
          // `handing out' the loads from largest to smallest
          Vector<long> total_loads( nProc,0 ); //total load per processor
          Vector<Vector<Load> > proc_loads( nProc ); //loads per processor
          int iproc_minload = 0; //processor with lowest load
          // loads are sorted in increasing order, so work backwards through the vector
          for( int i=loads.size()-1; i>=0; --i )
            {
              // put the next load on the processor with the lowest total load
              proc_loads[iproc_minload].push_back( loads[i] );
              total_loads[iproc_minload] += loads[i].load;

              // recompute which processor has the lowest load
              //[NOTE: this would be faster if the loads were sorted]
              iproc_minload = min_element( total_loads );
            }
          // compute average load per processor, truncated to int
          long avg_load = 0;
          for( int i=0; i<total_loads.size(); ++i ) avg_load += total_loads[i];
          avg_load /= nProc;

          // optimize the assignments by swapping a load off the
          // processor with the max load onto another processor
          // such that the load balance is improved
          int iter_count = 0, swap_count = 0;
          int iproc_maxload;
          long max_change; //largest change in load balance
          int ibmax,jbmax,ipmax,jpmax;  //box and processor indices corresponding to max_change

          while( 1 )
            {
              max_change = 0;

              // find the processor that has the largest deviation from perfect load balance
              min_max_elements( iproc_minload ,iproc_maxload ,total_loads );
              if( iproc_minload == iproc_maxload )
                {
                  // load balance is perfect
                  // (this won't happen except in test cases)
                  break;
                }
              assert( total_loads[iproc_minload] <= avg_load &&
                      avg_load <= total_loads[iproc_maxload] );
              if( avg_load - total_loads[iproc_minload] > total_loads[iproc_maxload] - avg_load )
                ipmax = iproc_minload;
              else
                ipmax = iproc_maxload;

              //[NOTE: dont need this here, but it may be useful for debugging.]
              eff_ratio = (Real)total_loads[iproc_minload] / (Real)total_loads[iproc_maxload];

              // deviation from perfect load balance for this proc
              long devib = total_loads[ipmax] - avg_load;

              // search all the other processors for the swap that has the maximum
              // reduction in the total deviation from the perfect load balance
              for( int j=0; j<proc_loads.size(); ++j )
                {
                  if( j != ipmax )
                    {
                      long devjb = total_loads[j] - avg_load;

                      // loop over all boxes on both processors
                      for( int ibox=0; ibox<proc_loads[ipmax].size(); ++ibox )
                        {
                          for( int jbox=0; jbox<proc_loads[j].size(); ++jbox )
                            {
                              iter_count++;
                              // how much bigger is the ibox load than the jbox load?
                              long diff = proc_loads[ipmax][ibox].load
                                - proc_loads[  j  ][jbox].load;
                              // change in total deviation from swapping boxes
                              long change = Abs( devib ) + Abs( devjb )
                                - Abs( devib - diff ) - Abs( devjb + diff );
                              // remember this pair of boxes if the change is better
                              //[NOTE: max_change starts at 0, so this is always an improvement]
                              if( change > max_change )
                                {
                                  max_change = change;
                                  ibmax = ibox; jbmax = jbox; jpmax = j;
                                }
                            }
                        }
                    }
                }
              // if there is a swap that improves load balance, take it; else stop
              if( max_change > 0 )
                {
                  // adjust the total loads on each processor
                  long load_diff = proc_loads[ipmax][ibmax].load
                    - proc_loads[jpmax][jbmax].load;
                  assert( load_diff != 0 );
                  total_loads[ipmax] -= load_diff;
                  total_loads[jpmax] += load_diff;
                  // swap the loads
                  Load tmp = proc_loads[ipmax][ibmax];
                  proc_loads[ipmax][ibmax] = proc_loads[jpmax][jbmax];
                  proc_loads[jpmax][jbmax] = tmp;

                  swap_count++;
                }
              else
                {
                  break;
                }
            }

          // Done with this level.

          // Compute the final efficiency ratio and save it if appropriate.
          min_max_elements( iproc_minload ,iproc_maxload ,total_loads );
          eff_ratio = (Real)total_loads[iproc_minload] / (Real)total_loads[iproc_maxload];
          if( eff_ratio < effRatio ) effRatio = eff_ratio;

          // Assign boxes to processors for this level.
          for( int ip=0; ip<proc_loads.size(); ++ip )
            {
              for( int jb=0; jb<proc_loads[ip].size(); ++jb )
                {
                  procAssignments[lvl][proc_loads[ip][jb].grid_index] = ip;
                }
            }

#ifndef NDEBUG
          //           if( iter_count > 0 )
          //             {
          //               cout << "    debug: LoadBalance: level " << lvl << " used "
          //                    << iter_count << " iterations and "
          //                    << swap_count << " swaps to get efficiency ratio "
          //                    << eff_ratio << std::endl;
          //             }
#endif
        }

      // Done with all levels.
      // We could try to permute processors assignments between levels to
      // reduce communication, but it is probably not worth the effort
      // since it probably would have O(N^4) cost (N==#boxes).
    }

  return status;
}

////////////////////////////////////////////////////////////////
//                utility functions                           //
////////////////////////////////////////////////////////////////

//
// Find the index of the small value in a non-empty (long) Vector
//
int
min_element( const Vector<long>& Vect )
{
  assert( Vect.size() > 0 );
  int imin = 0;
  for( int i=1; i<Vect.size(); ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i;
    }
  return imin;
}

//
// Find the indices of the smallest and largest values in a non-empty (long) Vector
//
void
min_max_elements( int& imin ,int& imax ,const Vector<long>& Vect )
{
  assert( Vect.size() > 0 );
  imin = 0; imax = 0;
  for( int i=1; i<Vect.size(); ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i;
      if( Vect[i] > Vect[imax] ) imax = i;
    }
  return;
}
void
min_max_elements( int& imin ,int& imax ,const Vector<long long>& Vect )
{
  assert( Vect.size() > 0 );
  imin = 0; imax = 0;
  for( int i=1; i<Vect.size(); ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i;
      if( Vect[i] > Vect[imax] ) imax = i;
    }
  return;
}
