/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdio>
#include <fstream>

#include "BoxIterator.H"
#include "LayoutIterator.H"
#include "BRMeshRefine.H"
//#include "LevelOp.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "Vector.H"
//#include "AMRIO.H"
#include "IntVectSet.H"

#include "NoelsPrettyPrint.H"
#include "DebugOut.H"

// under construction -- not ready for primetime ndk

void NoelsLDFABgetUsefulInfo(const LevelData<FArrayBox>& LDFAB,
                             int& Nboxes, int& Npoints, int& NpointsFromDataIterator) {

  Nboxes=0; Npoints=0;
  const DisjointBoxLayout& dbl = LDFAB.getBoxes();
  LayoutIterator lit = dbl.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit) {
    Nboxes++;
    Npoints += dbl[lit()].numPts();
  }

  NpointsFromDataIterator=0;
  for (DataIterator dit = LDFAB.dataIterator(); dit.ok(); ++dit) {
    NpointsFromDataIterator += LDFAB[dit()].box().numPts();
  }
}

void NoelsLDFABgetUsefulInfo(const LevelData<FArrayBox>* LDFAB_ptr,
                             int& Nboxes, int& Npoints, int& NpointsFromDataIterator) {

  Nboxes=0; Npoints=0;
  const DisjointBoxLayout& dbl = LDFAB_ptr->getBoxes();
  LayoutIterator lit = dbl.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit) {
    Nboxes++;
    Npoints += dbl[lit()].numPts();
  }

  NpointsFromDataIterator=0;
  for (DataIterator dit = LDFAB_ptr->dataIterator(); dit.ok(); ++dit) {
    NpointsFromDataIterator += (*LDFAB_ptr)[dit()].box().numPts();
  }

}

void NoelsDumpInputFile(const char *input_filename) {
  // spew out the input file

  pout() << " write out input file" << endl;
  char line[102];
  if (procID() == uniqueProc(SerialTask::compute)) {

    // verify file exists
    ifstream TS;
    TS.open(input_filename, ios::in);
    TS.close();

    if(TS.fail()) {
      pout() << " can't find this file or doesn't exist!  " << endl;
      return;
    }

    pout() << "---- input used for this run was as follows: ---" << endl;
    TS.open(input_filename, ios::in);
    while (!TS.eof()) {
      //TS >> pout() << endl;
      //cout >> TS << endl;
      TS.getline(line, 102);
      pout() << line << endl;
    }
    TS.close();
    pout() << "---- end input ---" << endl;
  }
}

void NoelsCompare(const FArrayBox* fab1, const FArrayBox* fab2, Real &diff) {

  diff = 0.0;
  BoxIterator bit((*fab1).box());
  for(bit.reset(); bit.ok(); ++bit) {
    if ( abs((*fab1)(bit(),0) - (*fab2)(bit(),0)) > diff) {
      diff = abs((*fab1)(bit(),0) - (*fab2)(bit(),0));
    }
  }
}

//void NoelsPrettyPrint(const FArrayBox* fab_ptr) {
// const string blank;
//NoelsPrettyPrint(&fab_ptr, blank);
//}

void NoelsPrettyPrint(const FArrayBox* fab_ptr, const string label) {

  BoxIterator bit((*fab_ptr).box());
  int nn=0;
  char temp[80];
  for(int iComp=0; iComp<(*fab_ptr).nComp(); iComp++) {
    cout << " comp" << iComp << ": " << endl;
    for(bit.reset(), nn=1; bit.ok(); ++nn, ++bit) {
      //sprintf(temp, "%16.10e ", (*fab_ptr)(bit(),0));
      //cout << temp;
      //if (nn%7 == 0) { cout << endl; }
      sprintf(temp, "%20.13e ", (*fab_ptr)(bit(),iComp));
      cout << temp;
      if (nn%4 == 0) { cout << endl; }
    }
    cout << endl;
  }
}

// poor dood's search
bool findValue_(const LevelData<FArrayBox>& locLDFB,
           const int x, const int y, const int z, Real& value) {

  DataIterator dit = locLDFB.dataIterator();

  for(dit.reset(); dit.ok(); ++dit) {
    const FArrayBox& fab = locLDFB[dit()];
    BoxIterator bit(fab.box());
#if CH_SPACEDIM == 3
    for(bit.reset(); bit.ok(); ++bit) {
      int xx=bit().getVect()[0];
      int yy=bit().getVect()[1];
      int zz=bit().getVect()[2];
      if(x==xx && y==yy && z==zz) {
    value = fab(bit(),0);
    return true;
      }
    }
#else
    for(bit.reset(); bit.ok(); ++bit) {
      int xx=bit().getVect()[0];
      int yy=bit().getVect()[1];
      if(x==xx && y==yy) {
    value = fab(bit(),0);
    return true;
      }
    }
#endif
  }
  return false;
}

void NoelsPrettyPrint(const LevelData<FArrayBox> memLDFB,
                      const string& label) {
  NoelsPrettyPrint(&memLDFB, label);
}

void NoelsPrettyPrint(const LevelData<FArrayBox> memLDFB) {
  const string blank;
  NoelsPrettyPrint(&memLDFB, blank);
}

// void NoelsPrettyPrint(const LevelData<FArrayBox>& memLDFB,
//                       const string& label) {
//   NoelsPrettyPrint(memLDFB, label);
// }
// void NoelsPrettyPrint(const LevelData<FArrayBox>& memLDFB) {
//   const string blank;
//   NoelsPrettyPrint(memLDFB, blank);
// }

void NoelsPrettyPrint(const LevelData<FArrayBox>* memLDFB_Ptr) {
  const string blank;
  NoelsPrettyPrint(memLDFB_Ptr, blank);
}
void NoelsPrettyPrint(const LevelData<FArrayBox>* memLDFB_Ptr,
                      const string& label) {

  const LevelData<FArrayBox>& memLDFB = *memLDFB_Ptr;

  Vector<Box> boxes;
  const DisjointBoxLayout& memDBL = memLDFB.getBoxes();
  LayoutIterator lit = memDBL.layoutIterator();
  for(lit.reset(); lit.ok(); ++lit) {
    boxes.push_back(memDBL.get(lit()));
  }

  Vector<int> assign(boxes.size(), 0);
  DisjointBoxLayout locDBL(boxes, assign);
  locDBL.close();
  LevelData<FArrayBox> locLDFB(locDBL,
                              memLDFB.nComp(),
                              memLDFB.ghostVect());
  const Interval& interv = memLDFB.interval();
  memLDFB.copyTo(interv, locLDFB, interv);

  LayoutIterator it = memDBL.layoutIterator();
  Box bound = memDBL[it()]; ++it;
  for(;it.ok() ; ++it) {
    bound = minBox(bound, memDBL[it()]);
  }
  const int xmin =  bound.smallEnd(0);
  const int xmax =  bound.bigEnd(0);
  const int ymin =  bound.smallEnd(1);
  const int ymax =  bound.bigEnd(1);
  //#if CH_SPACEDIM == 3
  //const int zmin =  bound.smallEnd(2);
  //const int zmax =  bound.bigEnd(2);
  //#endif

  Real value;
  Real tempNorm_max = norm(memLDFB, memLDFB.interval(), 0);
  Real tempNorm_L1 = norm(memLDFB, memLDFB.interval(), 1);

  printf(" %-36s max norm=%11.4e  L1=%11.4e\n",
         label.c_str(), tempNorm_max, tempNorm_L1);
  int k=0;
  for(int i=xmin; i<=xmax; i++) {
    printf("%4d ", i);
    for(int j=ymin; j<=ymax; j++) {
      if(findValue_(locLDFB, i, j, k, value)) {
    printf("%10.3e ", value);
      } else {
    printf("     .   ");
      }
    }
    printf("\n");
  }

  //printf("%4d-%11.3e", x, fab(bit(),0));

}

void NoelsMatrixCreation(const LevelData<FArrayBox>* memLDFB_Ptr) {
  //stub
}
