#ifdef CH_LANG_CC
/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//
#endif

#ifndef _REAL_H_
#define _REAL_H_

#ifdef CH_USE_FLOAT /* precision */
#  undef  CH_USE_DOUBLE
#  undef  CH_USE_FLOAT
#  define CH_USE_FLOAT  1
#else   /* precision */
#  undef  CH_USE_FLOAT
#  undef  CH_USE_DOUBLE
#  define CH_USE_DOUBLE 1
#endif  /* precision */

#if !defined(CH_LANG_FORT) /* language */
#  ifdef CH_USE_FLOAT /* precision */
     typedef float Real;
#    define MPI_CH_REAL         MPI_FLOAT
#    define H5T_NATIVE_REAL     H5T_NATIVE_FLOAT
#    define BASEFAB_REAL_SETVAL 1.23456789e+30
#  else   /* precision */
     typedef double Real;
#    define MPI_CH_REAL         MPI_DOUBLE
#    define H5T_NATIVE_REAL     H5T_NATIVE_DOUBLE
#    define BASEFAB_REAL_SETVAL 1.23456789e+300
#  endif  /* precision */

#else /* language */

#  ifdef CH_USE_FLOAT /* precision */
#    define REAL_T REAL
#    define COMPLEX_T COMPLEX
#  else /* precision */
#    if defined(CH_CRAY)
#      define REAL_T REAL
#      define COMPLEX_T COMPLEX
#    else /*cray */
#      define REAL_T REAL*8
#      define COMPLEX_T COMPLEX*16
#    endif /*cray*/
#  endif /* precision */

#endif /* language */

#if defined(CH_LANG_FORT)
#  define Real_T REAL_T
#  define Real_t REAL_T
#  define real_t REAL_T
#  define Complex_T COMPLEX_T
#  define complex_T COMPLEX_T
#  define complex_t COMPLEX_T
#endif

#if !defined(CH_USE_FLOAT) && !defined(CH_USE_DOUBLE)
#error Exactly one of CH_USE_FLOAT or CH_USE_DOUBLE must be defined
#endif

#endif
