/*
 * Customizable MPI stubs/wrappers for profiling, automatically
 * generated using m4. Generates stubs to call PMPI_* interface
 * from MPI_* calls.
 *
 * usage: m4 < mpi_prfwrap.m4 > mpi_prfwrap.c
 */

divert(-1)
define(`LCASE', `abcdefghijklmnopqrstuvwxyz')
define(`UCASE', `ABCDEFGHIJKLMNOPQRSTUVWXYZ')
/*
 * ----------------------------------------------------------------------
 * Code fragments to insert before and after using
 * MPI library (useful for initializing profiling, setting 
 * arrays, and dumping profiling results).
 * ----------------------------------------------------------------------
 *
 */
define(`INIT_PROLOG', `mpi_prf_init();')
define(`INIT_EPILOG', `PMPI_Comm_rank(MPI_COMM_WORLD,&(mpi_prf_myrank));')
define(`FINAL_PROLOG', `;')
define(`FINAL_EPILOG', `mpi_prf_finalize();')
define(`INIT_FPROLOG', `;')
define(`INIT_FEPILOG', `;')
define(`FINAL_FPROLOG', `;')
define(`FINAL_FEPILOG', `;')
/*
 *
 * ----------------------------------------------------------------------
 * Code fragments to insert before (prolog) and 
 * after (epilog) *all* MPI calls. C and Fortran variants
 * ----------------------------------------------------------------------
 *
 */


define(`PROLOG', `struct timeval       tv0, tv1;

  ++(mpi_prf_cs[$1`_k'].callcnt);
  gettimeofday(&tv0,0);
')
define(`EPILOG', `
  gettimeofday(&tv1,0);
  mpi_prf_cs[$1`_k'].cum_usec += TV2USEC(tv1,tv0);')
define(`FPROLOG', `
  ;')
define(`FEPILOG', `
  ;')

#
# utility macros
#
define(`FPOSTPEND', `ifdef(`POSTPEND', ``$1_'', ``$1'')')
define(`FUPCASE', `ifdef(`UPCASE', translit(``$1'', LCASE, UCASE),``$1'')')
define(`MERGE_ARG_PAIRS', `ifelse($#,  0,  , $#, 1,  , $#, 2, ``$1 $2'', ``$1 $2', MERGE_ARG_PAIRS(shift(shift($@)))')')
define(`DEL_ODD_ARG', `ifelse($#,  0,  , $#, 1,  , $#, 2, ``$2'', ``$2', DEL_ODD_ARG(shift(shift($@)))')')
define(`PM_SUB', `P$1(DEL_ODD_ARG(shift($@)));')
define(`PMF_SUB', `FPOSTPEND(FUPCASE(p$1))(DEL_ODD_ARG(shift($@)));')
define(`CONCAT', `define(`concattmp',  $1 $2`_k' `+' 
)define(`$1',`concattmp')')dnl
define(`CONCAT2',`define(`concat2tmp', $1 $2`_k' `+' "$2" `+' `0' `+' `0L' `+'
)define(`$1',`concat2tmp')')dnl
#
#
/*
 * ----------------------------------------------------------------------
 * Rules to generate C stubs
 * ----------------------------------------------------------------------
 */
define(`ENUM_LIST',)
define(`LBL_LIST',)

ifdef(`NOWRAP', `define(`MPI_FN_C', )', 
`define(`MPI_FN_C', `int $1(MERGE_ARG_PAIRS(shift($@)))
{
  int		rval;
  PROLOG($@)dnl
  rval = PM_SUB($@)dnl
  EPILOG($@)
  return rval;
}
CONCAT(`ENUM_LIST', `$1')
CONCAT2(`LBL_LIST', `$1')'
)')


define(`MPI_FN_C0', `int $1(MERGE_ARG_PAIRS(shift($@)))
{
  int		rval;

  ifelse($1, `MPI_Init', INIT_PROLOG, $1, `MPI_Finalize', FINAL_PROLOG, PROLOG)
  rval = PM_SUB($@)
  ifelse($1, `MPI_Init', INIT_EPILOG, $1, `MPI_Finalize', FINAL_EPILOG, EPILOG)
  return rval;
}')


/*
 * ----------------------------------------------------------------------
 * MPI types in C
 * ----------------------------------------------------------------------
 */

define(`Cargc',     `int *, argc')
define(`Cargv',     `char ***, argv')
define(`Ccomm',     `MPI_Comm, comm')
define(`Crank',     `int *, rank')
define(`Csize',     `int *, size')
define(`Csbuf',     `void *, sbuf')
define(`Cscnt',     `int, scnt')
define(`Cstype',    `MPI_Datatype, stype')
define(`Cstag',     `int, stag')
define(`Cdestrank', `int, destrank')
define(`Crbuf',     `void *, rbuf')
define(`Crcnt',     `int, rcnt')
define(`Crtype',    `MPI_Datatype, rtype')
define(`Crtag',     `int, rtag')
define(`Csrcrank',  `int, srcrank')
define(`Cstatus',   `MPI_Status *, status')
define(`Creq',      `MPI_Request *, req')
define(`Cstatusv',  `MPI_Status *, status')
define(`Creqv',     `MPI_Request *, req')
define(`Croot',     `int, root')
define(`Cop',       `MPI_Op, op')
#
#
# C source starts here
#
divert
`#include' <stdio.h>
`#include' <mpi.h>
`#include' <inttypes.h>
`#include' <sys/time.h>



#define TV2USEC(x,y)	((x.tv_sec - y.tv_sec)*1000000 + (x.tv_usec - y.tv_usec))

struct mpi_prf_callstat {
  int                   idx;
  char			*label;
  int		    	callcnt;
  int64_t		cum_usec;
};


/*
 *
 * ----------------------------------------------------------------------
 * MPI C stubs
 * ----------------------------------------------------------------------
 *
 */

divert(9)
MPI_FN_C0( MPI_Init,         Cargc, Cargv)
MPI_FN_C0( MPI_Finalize,     )
MPI_FN_C( MPI_Comm_rank,     Ccomm, Crank)
MPI_FN_C( MPI_Comm_size,     Ccomm, Csize)
MPI_FN_C( MPI_Send,          Csbuf, Cscnt, Cstype, Cdestrank, Cstag, Ccomm )
MPI_FN_C( MPI_Recv,          Crbuf, Crcnt, Crtype, Csrcrank, Crtag, Ccomm, Cstatus)
MPI_FN_C( MPI_Irecv,         Crbuf, Crcnt, Crtype, Csrcrank, Crtag, Ccomm, Creq )
MPI_FN_C( MPI_Isend,         Csbuf, Cscnt, Cstype, Cdestrank, Cstag, Ccomm, Creq )
MPI_FN_C( MPI_Issend,        Csbuf, Cscnt, Cstype, Cdestrank, Cstag, Ccomm, Creq )
MPI_FN_C( MPI_Wait,          Creq,  Cstatus )
MPI_FN_C( MPI_Waitall,       Cscnt, Creqv,  Cstatusv )
MPI_FN_C( MPI_Barrier,       Ccomm)
MPI_FN_C( MPI_Bcast,         Csbuf, Cscnt, Cstype, Croot, Ccomm)
MPI_FN_C( MPI_Allreduce,     Csbuf, Crbuf, Cscnt, Cstype, Cop, Ccomm)
divert

static int mpi_prf_myrank = -1;

enum {
translit(ENUM_LIST, `+', `,')
MPI_enum_end
};


static struct mpi_prf_callstat mpi_prf_cs[] = {
translit(LBL_LIST, `+', `,')
};


void mpi_prf_init( void )
{
  int		i;
  for (i=0; i<(MPI_enum_end-1); ++i) {
    mpi_prf_cs[i].callcnt  = 0;
    mpi_prf_cs[i].cum_usec = 0L;
  }
}



void mpi_prf_finalize( void )
{
  int		i;
  for (i=0; i<(MPI_enum_end-1); ++i) {
    fprintf(stderr, "[rank%04d]  %-20s     %8d  calls    %20lld usecs\n", 
             mpi_prf_myrank, mpi_prf_cs[i].label, mpi_prf_cs[i].callcnt, 
            (long long)mpi_prf_cs[i].cum_usec);
  }
}


undivert(9)
divert(-1)
/*
 * ----------------------------------------------------------------------
 * Rules to generate Fortran stubs
 * ----------------------------------------------------------------------
 */

ifdef(`NOWRAP', `define(`MPI_FN_F', )', 
`define(`MPI_FN_F', `void FPOSTPEND(FUPCASE($1))(MERGE_ARG_PAIRS(shift($@)))
{
  FPROLOG
  PMF_SUB($@)
  FEPILOG
}')')

define(`MPI_FN_F0',`void FPOSTPEND(FUPCASE($1))(MERGE_ARG_PAIRS(shift($@)))
{
  ifelse($1, `mpi_init', INIT_FPROLOG, $1, `mpi_finalize', FINAL_FPROLOG, FPROLOG)
  PMF_SUB($@)
  ifelse($1, `mpi_init', INIT_FEPILOG, $1, `mpi_finalize', FINAL_FEPILOG, FEPILOG)
}')

/*
 * ----------------------------------------------------------------------
 * MPI types in Fortran
 * ----------------------------------------------------------------------
 */
/*
define(`Fierr',         `int *, ierr')
define(`Fsbuf',         `void *, sbuf')
define(`Fscnt',         `int *, scnt')
define(`Fstype',        `int *, stype')
define(`Fdestrank',     `int *, destrank')
define(`Fstag',         `int *, stag')
define(`Fcomm',         `int *, comm')
define(`Freq',          `int *, req')
define(`Frbuf',         `void *, rbuf')
define(`Frcnt',         `int *, rcnt')
define(`Frtype',        `int *, rtype')
define(`Frtag',         `int *, rtag')
define(`Fsrcrank',      `int *, srcrank')
define(`Fstatus',       `int *, status')
define(`Fflag',         `int *, flag')
define(`Frank',         `int *, rank')
define(`Fsize',         `int *, size')
define(`Fop',           `int *, op')
define(`Froot',         `int *, root')
define(`Fblen',         `int *, blen')
define(`Fdisp',         `int *, disp')
define(`Fcom2',         `int *, com2')
define(`Fcolor',        `int *, color')
define(`Fkey',          `int *, key')
*/

divert
/*
 * ----------------------------------------------------------------------
 * MPI FORTRAN stubs
 * ----------------------------------------------------------------------
 */
/*MPI_FN_F0( mpi_init,          Fierr)
MPI_FN_F0( mpi_finalize,      Fierr)
MPI_FN_F( mpi_barrier,       Fcomm, Fierr)
MPI_FN_F( mpi_send,          Fsbuf, Fscnt, Fstype, Fdestrank, Fstag, Fcomm, Fierr)
MPI_FN_F( mpi_rsend,         Fsbuf, Fscnt, Fstype, Fdestrank, Fstag, Fcomm, Fierr)
MPI_FN_F( mpi_isend,         Fsbuf, Fscnt, Fstype, Fdestrank, Fstag, Fcomm, Freq, Fierr)
MPI_FN_F( mpi_recv,          Frbuf, Frcnt, Frtype, Fsrcrank,  Frtag, Fcomm, Fstatus, Fierr)
MPI_FN_F( mpi_irecv,         Frbuf, Frcnt, Frtype, Fsrcrank,  Frtag, Fcomm, Freq, Fierr)
MPI_FN_F( mpi_wait,          Freq,  Fstatus, Fierr)
MPI_FN_F( mpi_waitall,       Fscnt, Freq, Fstatus, Fierr)
MPI_FN_F( mpi_testall,       Fscnt, Freq, Fflag, Fstatus, Fierr)
MPI_FN_F( mpi_comm_rank,     Fcomm, Frank, Fierr)
MPI_FN_F( mpi_comm_size,     Fcomm, Fsize, Fierr)
MPI_FN_F( mpi_comm_dup,      Fcomm, Fcom2, Fierr)
MPI_FN_F( mpi_comm_free,     Fcomm, Fierr)
MPI_FN_F( mpi_comm_split,    Fcomm, Fcolor, Fkey, Fcom2, Fierr)
MPI_FN_F( mpi_sendrecv,      Fsbuf, Fscnt, Fstype, Fdestrank, Fstag, Frbuf, Frcnt, Frtype, Fsrcrank, Frtag, Fcomm, Fstatus, Fierr )
MPI_FN_F( mpi_reduce,        Fsbuf, Frbuf, Fscnt, Fstype, Fop, Froot, Fcomm, Fierr)
MPI_FN_F( mpi_allreduce,     Fsbuf, Frbuf, Fscnt, Fstype, Fop, Fcomm, Fierr)
MPI_FN_F( mpi_gather,        Fsbuf, Fscnt, Fstype, Frbuf, Frcnt, Frtype, Froot, Fcomm, Fierr)
MPI_FN_F( mpi_scatter,       Fsbuf, Fscnt, Fstype, Frbuf, Frcnt, Frtype, Froot, Fcomm, Fierr)
MPI_FN_F( mpi_allgather,     Fsbuf, Fscnt, Fstype, Frbuf, Frcnt, Frtype, Fcomm, Fierr)
MPI_FN_F( mpi_gatherv,       Fsbuf, Fscnt, Fstype, Frbuf, Frcnt, Fdisp, Frtype, Froot, Fcomm, Fierr)
MPI_FN_F( mpi_alltoall,      Fsbuf, Fscnt, Fstype, Frbuf, Frcnt, Frtype, Fcomm, Fierr)
MPI_FN_F( mpi_bcast,         Fsbuf, Fscnt, Fstype, Froot, Fcomm, Fierr)
MPI_FN_F( mpi_type_commit,   Fstype, Fierr)
MPI_FN_F( mpi_type_free,     Fstype, Fierr)
MPI_FN_F( mpi_type_indexed,  Fscnt, Fblen, Fdisp, Fstype, Frtype, Fierr)

*/