/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cmath>
#include "Misc.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BoxIterator.H"
#include "LayoutIterator.H"

#include "Timer.H"
Timer Everything("Everything");
Timer TMPI_Poi_levelGSRB_exchange      ("PlevelGSRB ex",      Everything, DIAG_TABLE3);
/// Global variables for test output:

static const char *pgmname = "copyTest" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

void parseTestOptions( int argc ,char* argv[] ) ;
extern int copyTest(void);

int main(int argc, char* argv[]) {

  bool passed = true;
  int icode = 0;

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = copyTest();

  return 0;
}

int copyTest() {

  DisjointBoxLayout loLayoutFine, loLayoutCoar;

#if(CH_SPACEDIM == 2)
  Box b1(IntVect(7,6), IntVect(7,9));
  Box b2(IntVect(3,0), IntVect(3,3));
  Box b3(IntVect(9,4), IntVect(9,5));
  Box b4(IntVect(-1,4), IntVect(-1,12));
#elif(CH_SPACEDIM == 3)
  Box b1(IntVect(7,6,0), IntVect(7,9,0));
  Box b2(IntVect(3,0,0), IntVect(3,3,0));
  Box b3(IntVect(9,4,0), IntVect(9,5,0));
  Box b4(IntVect(-1,4,0), IntVect(-1,12,0));
#else
  bogus spacedim;
#endif
  loLayoutCoar.addBox(b1,0);
  loLayoutCoar.addBox(b2,0);
  loLayoutCoar.addBox(b3,0);

  loLayoutFine.addBox(b1,0);
  loLayoutFine.addBox(b2,0);
  loLayoutFine.addBox(b3,0);
  loLayoutFine.addBox(b4,0);

  loLayoutCoar.close();
  loLayoutFine.close();

  LevelData<FArrayBox> loFabCoar(loLayoutCoar,1);
  LevelData<FArrayBox> loFabFine(loLayoutFine,1);

  DataIterator ditCoar =loFabCoar.dataIterator();

  Real fluxval = 4.77;
  int icode = 0;
  for(ditCoar.reset(); ditCoar.ok(); ++ditCoar) {
    FArrayBox& fab = loFabCoar[ditCoar()];
    Real val;
    if(loLayoutCoar.get(ditCoar()) == b2)
      val = -fluxval;
    else
      val = 0.0;

    fab.setVal(val);
  }

  DataIterator ditFine =loFabFine.dataIterator();
  for(ditFine.reset(); ditFine.ok(); ++ditFine)  {
    FArrayBox& fab = loFabFine[ditFine()];
    Real val = -fluxval/2.0;
    fab.setVal(val);
  }

  if(verbose)  {
    cout << indent2 << "LoFabCoar:" << endl;
    for(ditCoar.reset(); ditCoar.ok(); ++ditCoar)  {
      FArrayBox& fab = loFabCoar[ditCoar()];
      BoxIterator bit(fab.box());
      for(bit.reset(); bit.ok(); ++bit) {
    cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
      }
    }
    cout << indent2 << "LoFabFine:" << endl;
    for(ditFine.reset(); ditFine.ok(); ++ditFine) {
      FArrayBox& fab = loFabFine[ditFine()];
      BoxIterator bit(fab.box());
      for(bit.reset(); bit.ok(); ++bit) {
    cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
      }
    }
  }

  //loFabCoar.copyTo( loFabCoar.interval(),
  //        loFabFine,
  //        loFabFine.interval());

  if(&loFabCoar == &loFabFine){
    MayDay::Error("src == dest in copyTo function. Perhaps you want exchange ?");
  }

  if(loFabCoar.boxLayout() == loFabFine.boxLayout() ) {
    cout << " boxLayout's are identical" << endl;
  }

  if (loFabFine.ghostVect() == IntVect::Zero) {
    cout << " ghosts are zero" << endl;
  }

   if(loFabCoar.boxLayout() == loFabFine.boxLayout()  && loFabFine.ghostVect() == IntVect::Zero)  {
     cout << "parallel direct copy here, no communication issues" << endl;
     // parallel direct copy here, no communication issues
     for(DataIterator it(loFabCoar.dataIterator()); it.ok(); ++it) {
       loFabFine[it()].copy(loFabCoar.box(it()),
                loFabFine.interval(),
                loFabCoar.box(it()),
                loFabCoar.operator[](it()),
                loFabCoar.interval());
     }

   } else {

     cout << "make Copier" << endl;
     Copier copier(loFabCoar.disjointBoxLayout(), loFabFine.getBoxes(), loFabFine.ghostVect());
     loFabCoar.copyTo(loFabCoar.interval(), loFabFine, loFabFine.interval(), copier);
   }

  Real eps = 1.0e-10;
  for(ditFine.reset(); ditFine.ok(); ++ditFine) {
      FArrayBox& fab = loFabFine[ditFine()];
      Real rmax = fab.max(0);
      Real rmin = fab.min(0);
      Real valcomp;

      if(loLayoutFine.get(ditFine()) == b2)
        valcomp = -fluxval;
      else if(loLayoutFine.get(ditFine()) == b4)
        valcomp = -fluxval/2.0;
      else
        valcomp = 0.0;

      if(Abs(rmax - valcomp) > eps) {
          if(verbose)
            cout << indent2 << "max is wrong" << endl;
          icode =-1;
        }
      if(Abs(rmin - valcomp) > eps)
        {
          if(verbose)
            cout << indent2 << "min is wrong" << endl;
          icode =-2;
        }
    }
  if(verbose)
    {
      cout << indent2 << "LoFabCoar:" << endl;
      for(ditCoar.reset(); ditCoar.ok(); ++ditCoar)
        {
          FArrayBox& fab = loFabCoar[ditCoar()];
          BoxIterator bit(fab.box());
          for(bit.reset(); bit.ok(); ++bit)
            {
              cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
            }
        }
      cout << indent2 << "LoFabFine:" << endl;
      for(ditFine.reset(); ditFine.ok(); ++ditFine)
        {
          FArrayBox& fab = loFabFine[ditFine()];
          BoxIterator bit(fab.box());
          for(bit.reset(); bit.ok(); ++bit)
            {
              cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
            }
        }
    }
  return(icode);
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
