/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

//
// seq2ascii
//
// converts colormap file from seq format, produced by icol, to ascii rgb
// triplet format.  there must be num_colorslots color entries.
//
#include <fstream>
#include <string>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

int
ReadSeqPalette (std::vector <unsigned char>& red,
                std::vector <unsigned char>& green,
                std::vector <unsigned char>& blue,
                const std::string& fileName
  );

void
WriteASCIIPalette (const std::vector <unsigned char>& red,
                   const std::vector <unsigned char>& green,
                   const std::vector <unsigned char>& blue,
                   const std::string& fileName
  );

int num_colorslots;

// -------------------------------------------------------------------
int
main(int argc, char** argv)
{
  if (argc != 3)
  {
    cerr << "usage: seq2ascii seqfile asciifile" << endl;
    return (-1);
  }
  num_colorslots = 256;

  std::vector <unsigned char> red   (num_colorslots);
  std::vector <unsigned char> green (num_colorslots);
  std::vector <unsigned char> blue  (num_colorslots);

  std::string seqfile (argv[1]);
  ReadSeqPalette (red, green, blue, seqfile);

  std::string asciifile (argv[2]);
  WriteASCIIPalette (red, green, blue, asciifile);
}

// -------------------------------------------------------------------
int
ReadSeqPalette (std::vector <unsigned char>& red,
                std::vector <unsigned char>& green,
                std::vector <unsigned char>& blue,
                const std::string& fileName
  )
{
  unsigned char rbuff[num_colorslots];
  unsigned char gbuff[num_colorslots];
  unsigned char bbuff[num_colorslots];
  int   i, fd;          /* file descriptor */

  if((fd = open(fileName.c_str(), O_RDONLY, NULL)) < 0) {
    cout << "Can't open colormap file:  " << fileName << endl;
    return (-1);
  }

  if(read(fd, rbuff, num_colorslots) != num_colorslots) {
    cout << "palette is not a seq colormap." << endl;
    return(-2);
  }
  if(read(fd, gbuff, num_colorslots) != num_colorslots) {
    cout << "file is not a seq colormap." << endl;
    return(-2);
  }
  if(read(fd, bbuff, num_colorslots) != num_colorslots) {
    cout << "palette is not a seq colormap." << endl;
    return(-2);
  }
  (void) close(fd);

  for (int i = 0; i < num_colorslots; ++i)
  {
    red[i]   = rbuff[i];
    green[i] = gbuff[i];
    blue[i]  = bbuff[i];
  }
  return(0);

}  // end ReadSeqPalette()

// -------------------------------------------------------------------
void
WriteASCIIPalette (const std::vector <unsigned char>& red,
                   const std::vector <unsigned char>& green,
                   const std::vector <unsigned char>& blue,
                   const std::string& fileName
  )
{
  assert (red.size () == green.size () && red.size () == blue.size ());
  ofstream strm (fileName.c_str ());
  for (int i = 0; i < red.size (); ++i)
  {
    strm << int(red[i]) << "  " << int(green[i]) << "  " << int(blue[i]) << endl;
  }
  strm.close ();
}
