/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DATATYPE_H_
#define _DATATYPE_H_
#
/* This structure contains information relevant to your program.
 * You should fill it in with information that you need.
 *
 */
#include <X11/Intrinsic.h>
#include "Vector.H"
#include "Box.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "IntVectSet.H"
#include "PiecewiseLinearFillPatch.H"
#include <string>
#include "Tuple.H"
using std::string;
/* define's */

/*
#define X_SIZE 512
#define Y_SIZE 512
#define INFOB 125
#define NUMCONT 30
*/

typedef LevelData<FArrayBox> MultiFab;
enum FileType { AMRBinary, EBAMRASCII, SingleFAB, XDangFAB, AMRASCII};
enum PlotType { ContourPlot, ColorMap };

typedef struct datatype
{
  Vector<Box> vect_box;
  Vector<int> vect_ratio;
  Vector<LevelData<FArrayBox>* > vect_mf;
  Vector<string> vect_char;
  Vector<DisjointBoxLayout> vect_dbl;
  int cur_var;
  int xdraw;
  int ydraw;
  int zdraw;
  char filename[250];
  PlotType plottype;
  int width;
  int height;
  int infob;
  Real max;
  Real min;
  Real mag;
  Real eps;
  bool drawboxes;
  int inormal;
  int idepth;
  int numcont;
  char* filein[1000];
  Tuple<int, 2> axisdir;
  FileType filetype;
  Widget window;
  Widget drawwidget;
  int num_windows;
  int nfiles;
  Box lev0subbox;
  Box mousebox;
  bool doingsubregion;
  IntVect mouse_start_pix;
  IntVect mouse_end_pix;
  unsigned char rcol[256];
  unsigned char gcol[256];
  unsigned char bcol[256];
  char filecolors[250];
  bool hascolorfile;
  //eb addenda
  bool ebflag;
  Vector<IntVectSet> coveredCells;
  Vector<IntVectSet> multiValuedCells;
} datatype;

#endif
