/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// NoOpBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, July 12, 2004

#ifndef _NOOPBC_H_
#define _NOOPBC_H_

#include "GhostBC.H"

/// class to do nothing at a boundary
/** This class functions as a placeholder -- if you don't want
    to change the values on this edge, but it's a physical boundary,
    it's the clas you want! */
class NoOpBC : public BoxGhostBC
{
public:
  
  /// 
  NoOpBC();

  ///
  NoOpBC(int a_dir, Side::LoHiSide a_sd);

  ///
  NoOpBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  ///
  virtual ~NoOpBC();

protected:

  /// virtual constructor workaround
  virtual BoxGhostBC* new_boxghostbc() const;

  /// sets interval to be (0,0)
  void define(int a_dir, Side::LoHiSide a_sd);

  /// 
  void define(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  ///
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state,
			const ProblemDomain& a_domain,
			Real a_dx) const;

  /// deprecated interface using Box instead of ProblemDomain
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state,
			const Box& a_domain,
			Real a_dx) const;
  ///
  virtual void
  applyHomogeneousBCs(FArrayBox& a_state, 
		      const ProblemDomain& a_domain,
		      Real a_dx) const;

  /// deprecated interface using Box instead of ProblemDomain
  virtual void
  applyHomogeneousBCs(FArrayBox& a_state, 
		      const Box& a_domain,
		      Real a_dx) const;

  
  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const ProblemDomain& a_domain) const;


  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const Box& a_domain) const;

private:
  NoOpBC(const NoOpBC&) {;}
  virtual void operator =(const NoOpBC&) {;}


};
#endif



