/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// NoOpBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, July 12, 2004


/*****************/
/*****************/


#include "NoOpBC.H"


NoOpBC::NoOpBC() :BoxGhostBC()
{
}


NoOpBC::NoOpBC(int a_dir, Side::LoHiSide a_sd) :BoxGhostBC(a_dir, a_sd)
{
}

NoOpBC::NoOpBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps)
  : BoxGhostBC(a_dir, a_sd, a_comps)
{
}

NoOpBC::~NoOpBC() 
{
}

BoxGhostBC*
NoOpBC::new_boxghostbc() const
{
  NoOpBC* newop = new NoOpBC();
  if (newop == NULL) {
    MayDay::Error("Out of memory in edgeDirichletBC::new_boxghostBC");
  } 

  return static_cast<BoxGhostBC*> (newop);
}


void 
NoOpBC::define(int a_dir, Side::LoHiSide a_sd) 
{
  BoxGhostBC::define(a_dir, a_sd);
}

void 
NoOpBC::define(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps) 
{
  BoxGhostBC::define(a_dir, a_sd, a_comps);
}

void
NoOpBC::applyInhomogeneousBCs(FArrayBox& a_state,
			      const Box& a_domain,
			      Real a_dx) const
{
  // do nothing here
}

void
NoOpBC::applyHomogeneousBCs(FArrayBox& a_state, 
		      const Box& a_domain,
		      Real a_dx) const
{
  // true to its name, this class does nothing here
}


void
NoOpBC::fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const Box& a_domain) const
{
  // this also doesn't do anything...
}



void
NoOpBC::applyInhomogeneousBCs(FArrayBox& a_state,
			      const ProblemDomain& a_domain,
			      Real a_dx) const
{
  // do nothing here
}

void
NoOpBC::applyHomogeneousBCs(FArrayBox& a_state, 
		      const ProblemDomain& a_domain,
		      Real a_dx) const
{
  // true to its name, this class does nothing here
}


void
NoOpBC::fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const ProblemDomain& a_domain) const
{
  // this also doesn't do anything...
}

