/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// ReflectODDBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#include "MayDay.H"
#include "ReflectOddBC.H"
#include "BCF_F.H"
#include "GhostBCF_F.H"

#include "Timer.H"
extern Timer Everything;
Timer TF_REFLECTGHOSTBC2("REFLECTGHOSTBC2",   Everything, DTIMEFORT, true);

// -------------------------------------------------------------
ReflectOddBC::ReflectOddBC() : BoxGhostBC()
{}

// -------------------------------------------------------------
ReflectOddBC::~ReflectOddBC() 
{
}


// -------------------------------------------------------------
ReflectOddBC::ReflectOddBC(int dir, Side::LoHiSide sd,
		   const Interval& a_comps) :BoxGhostBC(dir,sd,a_comps)
{
}

// -------------------------------------------------------------
ReflectOddBC::ReflectOddBC(int dir, Side::LoHiSide sd)
  :BoxGhostBC(dir,sd)
{
}

// -------------------------------------------------------------
void 
ReflectOddBC::fillBCValues(FArrayBox& a_neumfac,
		       FArrayBox& a_dircfac,
		       FArrayBox& a_inhmval,
		       Real a_dx,
		       const Box& a_domain) const
{

  // this doesn't do anything because we don't need these arrays...
  // this function is just needed to complete the instantiation.
}

// -------------------------------------------------------------
void 
ReflectOddBC::fillBCValues(FArrayBox& a_neumfac,
		       FArrayBox& a_dircfac,
		       FArrayBox& a_inhmval,
		       Real a_dx,
		       const ProblemDomain& a_domain) const
{

  // this doesn't do anything because we don't need these arrays...
  // this function is just needed to complete the instantiation.
}




// -------------------------------------------------------------
void
ReflectOddBC::applyHomogeneousBCs(FArrayBox& a_state,
			      const Box& a_domain,
			      Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyReflectOddBCs(a_state, physdomain, a_dx);
}


// -------------------------------------------------------------
void
ReflectOddBC::applyHomogeneousBCs(FArrayBox& a_state,
			      const ProblemDomain& a_domain,
			      Real a_dx) const 
{
  applyReflectOddBCs(a_state, a_domain, a_dx);
}

// -------------------------------------------------------------
void
ReflectOddBC::applyInhomogeneousBCs(FArrayBox& a_state,
				const Box& a_domain,
				Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyReflectOddBCs(a_state, physdomain, a_dx);
}


// -------------------------------------------------------------
void
ReflectOddBC::applyInhomogeneousBCs(FArrayBox& a_state,
				const ProblemDomain& a_domain,
				Real a_dx) const 
{
  applyReflectOddBCs(a_state, a_domain, a_dx);
}



// -------------------------------------------------------------
void
ReflectOddBC::applyReflectOddBCs(FArrayBox& a_state, 
                                 const ProblemDomain& a_domain, 
                                 Real a_dx) const 
{
  
  // do nothing in periodic case
  if (!a_domain.isPeriodic(m_direction))
    {
      // first construct BC box  
      Box bx= a_state.box();
      bx.grow(-1);
      Box bc_box;
      bool isbc;
      int idir = m_direction;
      if(m_side == Side::Lo)
        {
          isbc = (bx.smallEnd(idir) <= a_domain.domainBox().smallEnd(idir));
          if(isbc)
            {
              int ichop = a_domain.domainBox().smallEnd(m_direction);
              bc_box = a_state.box();
              bc_box.chop(m_direction, ichop);
            }
        }
      else if(m_side == Side::Hi)
        {
          isbc = (bx.bigEnd(idir) >= a_domain.domainBox().bigEnd(idir));
          if(isbc)
            {
              int ichop = a_domain.domainBox().bigEnd(m_direction)+1;
              Box chop_box = a_state.box();
              bc_box = chop_box.chop(m_direction,ichop);
            }
        }
      else 
        {
          cerr << "DomainGhostBC::applyghostbc: bogus side" << endl;
          abort();
        }
      
      if (isbc) {
        int iSide = sign(m_side);
        int iDir = m_direction;
        
#ifndef NDEBUG
        assert((m_side == Side::Hi)||(m_side == Side::Lo));
        Box biggerbox = bc_box;
        //state has to contain one cell on side of ghost cell
        if(m_side == Side::Lo)
          biggerbox.growHi(iDir, 1);
        else 
          biggerbox.growLo(iDir, 1);
        
        
        assert(a_state.box().contains(biggerbox));
        
        assert(m_components.begin() >= 0);
        assert(m_components.end() < a_state.nComp());
#endif
        
        
        int startcomp = m_components.begin();
        int endcomp = m_components.end();
        
        // this is reflect-odd, so scale = -1
        Real scale = -1.0;
        TF_REFLECTGHOSTBC2.start();
        FORT_REFLECTGHOSTBC(CHF_FRA(a_state),
                            CHF_BOX(bc_box),
                            CHF_CONST_INT(iDir),
                            CHF_CONST_INT(iSide),
                            CHF_CONST_REAL(a_dx),
                            CHF_CONST_INT(startcomp),
                            CHF_CONST_INT(endcomp),
                            CHF_CONST_REAL(scale));
        TF_REFLECTGHOSTBC2.stop();
      } // end if this is a bc
      
    }
}
		       

// -------------------------------------------------------------
BoxGhostBC*
ReflectOddBC::new_boxghostbc() const
{
  ReflectOddBC* newop = new ReflectOddBC();
  if (newop == NULL) {
    MayDay::Error("Out of memory in ReflectOddBC::new_boxghostbc");
  }
  
  return static_cast<BoxGhostBC*>(newop);
  
}
