/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// basicScalBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _basicScalBC_H_
#define _basicScalBC_H_

#include "FArrayBox.H"
#include "REAL.H"
#include "LevelData.H"
#include "GhostBC.H"


/// this is a basic physical BC class designed to handle scalars
class basicScalBC
{
public:
  
  ///
  basicScalBC();
  
  /// 
  virtual ~basicScalBC();

  /// DomainGhostBC is set 
  virtual void setBCType(const DomainGhostBC& a_ghostBC);

  /// or the derived class can do it for you
  virtual void setAllBCTypes() = 0;

  ///
  basicScalBC();
  
  /// 
  virtual void applyBCs(FArrayBox& a_state, 
                        const ProblemDomain& a_domain,
			const Real a_dx) const;

  /// 
  virtual void applyBCs(LevelData<FArrayBox>& a_stateMF, 
			const ProblemDomain& a_domain, 
                        const Real a_dx) const;


  /// deprecated interface
  virtual void applyBCs(FArrayBox& a_state, const Box& a_domain,
			const Real a_dx) const;

  /// deprecated interface
  virtual void applyBCs(LevelData<FArrayBox>& a_stateMF, 
			const Box& a_domain, const Real a_dx) const;



protected:
  DomainGhostBC* m_componentBC;

private:

};


#endif


