/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// basicScalBC.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "basicScalBC.H"

// -------------------------------------------------------------
basicScalBC::basicScalBC()
{
  // explicitly set pointers to NULL
  m_componentBC = NULL;
    
}

// --------------------------------------------------------------
basicScalBC::~basicScalBC() 
{
  if (m_componentBC != NULL) {
    delete m_componentBC;
    m_componentBC = NULL;
  }
}


// --------------------------------------------------------------
void
basicScalBC::setBCType(const DomainGhostBC& a_ghostBC) 
{

  if (m_componentBC == NULL) {
    m_componentBC = new DomainGhostBC;
    DomainGhostBC& thisBC = *m_componentBC;
    thisBC = a_ghostBC;    
  } else {
    DomainGhostBC& thisBC = *m_componentBC;
    thisBC = a_ghostBC;
  }
}



// -----------------------------------------------------------------
void
basicScalBC::applyBCs(LevelData<FArrayBox>& a_stateMF, 
		      const Box& a_domain, const Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_stateMF, a_domain, a_dx);
}


// -----------------------------------------------------------------
void
basicScalBC::applyBCs(LevelData<FArrayBox>& a_stateMF, 
		      const ProblemDomain& a_domain, 
                      const Real a_dx) const 
{
  
  // just loop through boxes and call single-grid version
  DataIterator dit=a_stateMF.iterator();

  for (dit.reset(); dit.ok(); ++dit) {
    applyBCs(a_stateMF[dit()], a_domain, a_dx);
  }
}


// ------------------------------------------------------------
void
basicScalBC::applyBCs(FArrayBox& a_state, const Box& a_domain,
		      const Real a_dx) const 
{
  ProblemDomain physdomain(a_domain);
  applyBCs(a_state, a_domain, a_dx);
}

// ------------------------------------------------------------
void
basicScalBC::applyBCs(FArrayBox& a_state, 
                      const ProblemDomain& a_domain,
		      const Real a_dx) const 
{
  
  // basic assumption here that all scalar BC's are homogeneous
  // (or that it doesn't matter) -- can obviously be overridden 
  // in a derived class

  // loop over velocity components and apply appropriate BC's
  DomainGhostBC& thisBC = *(m_componentBC);
  thisBC.applyHomogeneousBCs(a_state, a_domain, a_dx);
  
}


