/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// boxVelBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _boxVelBC_H_
#define _boxVelBC_H_

#include "basicVelBC.H"
#include "FluxBox.H"
#include "ProblemDomain.H"

/// implements cell-centered velocity BC's for flow in a box!
class boxVelBC: public basicVelBC
{
  
public:
  
  /// 
  boxVelBC(bool isViscous=false);

  /// 
  virtual ~boxVelBC() {};

  /// 
  virtual void setAllBCTypes();

  /// access function
  bool isViscous();

  /// 
  void isViscous(bool a_isViscous);

protected:
  ///
  bool m_isViscous;

private:

};


/// implements edge-centered velocity BC's for flow in a box!
class boxEdgeVelBC: public basicVelBC
{
  
public:
  
  /// 
  boxEdgeVelBC(bool isViscous=false);

  /// 
  virtual ~boxEdgeVelBC() {};

  /// 
  virtual void setAllBCTypes();

  /// access function
  bool isViscous();

  /// 
  void isViscous(bool a_isViscous);

  /// also need to implement edge-centered version of applyBCs
  virtual void applyBCs(LevelData<FluxBox>& a_edgeStateMF,
			const ProblemDomain& a_domain, 
                        const Real a_dx) const;

  /// also need to implement edge-centered version of applyBCs (old interface)
  virtual void applyBCs(LevelData<FluxBox>& a_edgeStateMF,
			const Box& a_domain, const Real a_dx) const;


  /// also need to implement edge-centered version of applyBCs
  virtual void applyBCs(FluxBox& a_edgeStateMF,
			const ProblemDomain& a_domain, 
                        const Real a_dx) const;

  /// also need to implement edge-centered version of applyBCs (old interface)
  virtual void applyBCs(FluxBox& a_edgeStateMF,
			const Box& a_domain, const Real a_dx) const;
  
  /// since the other two functions hide the applyBCs function...
  virtual void applyBCs(FArrayBox& a_state, 
                        const ProblemDomain& a_domain, 
			const Real a_dx) const;


  /// since the other two functions hide the applyBCs function (old interface)
  virtual void applyBCs(FArrayBox& a_state, const Box& a_domain, 
			const Real a_dx) const;

  /// this one assumes that each edge-centered FAB is one component
  virtual void applyBCs(FArrayBox& a_state, const ProblemDomain& a_domain,
			const Real a_dx, int a_velComp) const;


  /// this one assumes each edge-centered FAB is one component (old interface)
  virtual void applyBCs(FArrayBox& a_state, const Box& a_domain,
			const Real a_dx, int a_velComp) const;

protected:
  ///
  bool m_isViscous;

private:
};

#endif
