/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// edgeGhostBC.H

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000

#ifndef _edgeGhostBC_H_
#define _edgeGhostBC_H_

#include "GhostBC.H"
// include NoOpBC for backward compatibility (used to be in this file)
#include "NoOpBC.H"

/// class to do edge-centered physical boundary conditions
/** BoxEdgeBC is designed to do physical BC's on edge centered
    data.  It is derived from BoxGhostBC so that it can be used
    seamlessly with BoxGhostBCs.  */

class BoxEdgeBC: public BoxGhostBC
{
public:
  /// 
  virtual ~BoxEdgeBC();
  
  ///
  BoxEdgeBC();

  ///  sets interval to be (0,0)
  BoxEdgeBC(int a_dir, Side::LoHiSide a_sd);

  ///
  BoxEdgeBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  /// 
  BoxEdgeBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps, 
	    IndexType& a_ixtype);

  /// returns the type of centering of this edge-centered BC
  const IndexType& ixType() const;
  
  ///
  void ixType(const IndexType& a_ixType);


protected:

  /// virtual constructor workaround
  virtual BoxGhostBC* new_boxghostbc() const;

  /// sets interval to be (0,0)
  void define(int a_dir, Side::LoHiSide a_sd);

  /// 
  void define(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  /// 
  void define(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps,
	      IndexType& a_ixtype);

  ///
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state,
			const ProblemDomain& a_domain,
			Real a_dx) const;
  ///
  virtual void
  applyHomogeneousBCs(FArrayBox& a_state, 
		      const ProblemDomain& a_domain,
		      Real a_dx) const;

  ///
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const ProblemDomain& a_domain) const;


  /// deprecated interface
  virtual void
  applyInhomogeneousBCs(FArrayBox& a_state,
			const Box& a_domain,
			Real a_dx) const;
  /// deprecated interface
  virtual void
  applyHomogeneousBCs(FArrayBox& a_state, 
		      const Box& a_domain,
		      Real a_dx) const;

  /// deprecated interface
  virtual void
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const Box& a_domain) const;

  ///
  virtual void 
  applyBCs(const Box& a_bcbox, 
           FArrayBox& a_state,
           const FArrayBox& a_neumfac,
           const FArrayBox& a_dircfac,
           const FArrayBox& a_inhmval,
           Real a_dx) const;


private:
  BoxEdgeBC(const BoxEdgeBC&) {;}
  virtual void operator=(const BoxEdgeBC&) {;}
  
  IndexType m_ix_type;
  
};


/// class to do Dirichlet edge-centered BC's 
/** this class implements the edge-centered instantiation
    of the edge-centered homogeneous BC's */
class edgeDirichletBC : public BoxEdgeBC
{
public:
  friend class DomainGhostBC;

  /// 
  ~edgeDirichletBC();
  
  ///
  edgeDirichletBC();

  /// sets interval to be (0,0)
  edgeDirichletBC(int a_dir, Side::LoHiSide a_sd);

  /// 
  edgeDirichletBC(int a_dir, Side::LoHiSide a_sd, const Interval& a_comps);

  /// set edge-BC value (single-valued)
  void setBCVal(const Real a_bcVal);

  ///
  Real BCVal() const;

protected:
  virtual void 
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const ProblemDomain& a_domain) const;


  virtual void 
  fillBCValues(FArrayBox& a_neumfac,
	       FArrayBox& a_dircfac,
	       FArrayBox& a_inhmval,
	       Real a_dx,
	       const Box& a_domain) const;

  ///
  virtual  BoxGhostBC* new_boxghostbc() const;

  /// 
  Real m_bcVal;


private:
  edgeDirichletBC(const edgeDirichletBC& dirin) {;}
  void operator=(const edgeDirichletBC& dirin) {;}
};
    

#endif



