/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokes.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


/*****************/
/*****************/

#ifndef _AMRNavierStokeHeaders_H_
#define _AMRNavierStokeHeaders_H_

#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
#include <cmath>
#include <cstdio>

#include "parstream.H"
#include "LoadBalance.H"
#include "LayoutIterator.H"
#include "FineInterp.H"
#include "ParmParse.H"
#include "LoHiSide.H"
#include "BoxIterator.H"
#include "ProblemDomain.H"
#include "AMRIO.H"

#ifdef HDF5
#include "CH_HDF5.H"
#endif

#include "AMRNavierStokes.H"
#include "GodunovTrace.H"
#include "basicVelBC.H"
#include "ExtrapBC.H"
#include "HOExtrapBC.H"
#include "PoissonBC.H"
#include "scalarDirichletBC.H"
#include "viscousBC.H"
#include "Divergence.H"
#include "CellToEdge.H"
#include "EdgeToCell.H"
#include "computeNorm.H"
#include "computeSum.H"
#include "IAMRGodunovF_F.H"
#include "PROB_F_F.H"
#include "AMRNS_F_F.H"
#include "timeInterp.H"
#include "Gradient.H"

#include "Timer.H"

#ifdef DEBUG
#include "computeSum.H"
#endif

// this is necessary because we call library fortran directly
#include "PoissonOpF_F.H"

// #define TEST_TAGGING  

#ifdef NEW_ADVECTION
#define ADVECT_GROW 4
#else
#define ADVECT_GROW 2
#endif

#ifdef CH_USE_DOUBLE
#define TIME_EPS 1.0e-10
#else
#define TIME_EPS 1.0e-5
#endif

// small parameter used in defining TGA constants for viscous solves
#define TGA_EPS 1.0e-8

//#include "AMRNavierStokesTimers.H"
extern Timer Everything;

extern Timer TimeNSadvance;           
extern Timer TimeNSadvanceIntro;
extern Timer TimeNSadvanceCompAV;
extern Timer TimeNSadvanceAdvectLambda;
extern Timer TimeNSadvanceAdvectDiffScalar;
extern Timer TimeNSadvancePredictVelocities;
extern Timer TimeNSadvanceCompUStar;
extern Timer TimeNSadvanceLevelProject;

extern Timer TimeUStar_notViscous;
extern Timer TimeUStarMetricNorm;
extern Timer TimeUStarDefines;
extern Timer TimeUStarComputeLap;   
extern Timer TimeUStarCompSource;
extern Timer TimeUStarSetupForSolves;
extern Timer TimeUStarlevelSolve1;  
extern Timer TimeUStarlevelSolve2;
extern Timer TimeUStarConcludeSolve;

extern Timer TimeNSpost;
extern Timer TimeNSpostPart1;
extern Timer TimeNSpostPart2;
extern Timer TimeNSpostImplicitReflux;
extern Timer TimeNSpostSyncOperations;
extern Timer TimeNSpostDiagnostic;
extern Timer TimeNSpostCreateBC;

extern Timer TimeNSpostRegrid;
extern Timer TimeinitialVelocityProject;
extern Timer TimedoPostRegridOps;
extern Timer TimeinitializeGlobalPressure;

extern Timer TimecomputeAdvectionVelocities;
extern Timer TimeCAVfillVelocity;
extern Timer Time_computeLapVel;
extern Timer TimeTraceState;
extern Timer TimelevelMacProject;

extern Timer TTS_setupPatchGodunov;
extern Timer TTS_tempStorage;
extern Timer TTS_computeFluxes;
extern Timer TTS_copyflux;

extern Timer TimeSA_diffusionSolver;
extern Timer TimeSA_viscousSolver;
extern Timer TimeSA_initialData;
extern Timer TimeSA_velocity;
extern Timer TimeSA_scalar;

extern Timer TimeLS_advectScalar1;
extern Timer TimeLS_advectScalar2;
extern Timer TimeLS_compUstar1;
extern Timer TimeLS_compUstar2;

extern Timer TimeLS_compUstarInit1; //noel
extern Timer TimeLS_compUstarInit2; //noel


// AMRNavierStokesInit.cpp
extern Timer TF_INITVORT;
extern Timer TF_COMPUTEVORT;
extern Timer TF_INITVEL;
extern Timer TF_INITSCAL;

// AMRNavierStokesVel.cpp
extern Timer TF_OPERATORLAP_ustar;
extern Timer TF_OPERATORLAP_ustarInit;
extern Timer TF_UDELS_predictVelocities;

// AMRNavierStokesScal.cpp
extern Timer TF_OPERATORLAP_advectDiffuseScalar;

// AMRNavierStokesRegrid.cpp
extern Timer TF_MAGVECT_Regrid;

#endif
