/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// AMRNavierStokesTimers.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


/*****************/
/*****************/

#include "AMRNavierStokesHeaders.H"

extern Timer Everything;

Timer TimeNSadvance                 ("advance", 5);
Timer TimeNSadvanceIntro            ("Intro",            TimeNSadvance);
Timer TimeNSadvanceCompAV           ("CompAV",           TimeNSadvance);
Timer TimeNSadvanceAdvectLambda     ("AdvectLambda",     TimeNSadvance);
Timer TimeNSadvanceAdvectDiffScalar ("AdvectDiffScalar", TimeNSadvance);
Timer TimeNSadvancePredictVelocities("Predict Velocity", TimeNSadvance);
Timer TimeNSadvanceCompUStar        ("Comp UStar",       TimeNSadvance, 30);
Timer TimeNSadvanceLevelProject     ("LevelProject",     TimeNSadvance);


Timer TimeUStar_notViscous   ("not viscous",      TimeNSadvanceCompUStar);
Timer TimeUStarMetricNorm    ("norm for metric",  TimeNSadvanceCompUStar);
Timer TimeUStarDefines       ("defines",          TimeNSadvanceCompUStar);
Timer TimeUStarComputeLap    ("comp laplacian",   TimeNSadvanceCompUStar);
Timer TimeUStarCompSource    ("comp source term", TimeNSadvanceCompUStar);
Timer TimeUStarSetupForSolves("setup for solves", TimeNSadvanceCompUStar);
Timer TimeUStarlevelSolve1   ("levelSolve1",      TimeNSadvanceCompUStar);
Timer TimeUStarlevelSolve2   ("levelSolve2",      TimeNSadvanceCompUStar);
Timer TimeUStarConcludeSolve ("conclude solves",  TimeNSadvanceCompUStar);

Timer TimeNSpost              ("PostTimeStep", 7);
Timer TimeNSpostPart1         ("Part1",          TimeNSpost);
Timer TimeNSpostPart2         ("Part2",          TimeNSpost);
Timer TimeNSpostImplicitReflux("ImplicitReflux", TimeNSpost);
Timer TimeNSpostSyncOperations("SyncOperations", TimeNSpost);
Timer TimeNSpostDiagnostic    ("Diagnostic",     TimeNSpost);
Timer TimeNSpostCreateBC      ("CreateBC",       TimeNSpost);

Timer TimeNSpostRegrid("PostRegrid", 15);
Timer TimeinitialVelocityProject  ("initVelProject",     TimeNSpostRegrid);
Timer TimedoPostRegridOps         ("postRegridOps",      TimeNSpostRegrid);
Timer TimeinitializeGlobalPressure("initGlobalPress",    TimeNSpostRegrid);

Timer TimecomputeAdvectionVelocities("comp Adv Vel", 20);
Timer TimeCAVfillVelocity           ("fillVelocity",  TimecomputeAdvectionVelocities);
Timer Time_computeLapVel            ("computeLapVel", TimecomputeAdvectionVelocities);
Timer TimeTraceState                ("Trace State",   TimecomputeAdvectionVelocities, 21);
Timer TimelevelMacProject           ("levelMacProj",  TimecomputeAdvectionVelocities);

Timer TTS_setupPatchGodunov("setup PatchG", TimeTraceState);
Timer TTS_computeFluxes    ("computeFluxes",TimeTraceState);
Timer TTS_tempStorage      ("tempStorage",  TimeTraceState);
Timer TTS_copyflux         ("copy flux",    TimeTraceState);

Timer TimeSA_diffusionSolver      ("SA diffusion",   Everything, DTIMESAMR, true);
Timer TimeSA_viscousSolver        ("SA viscous",     Everything, DTIMESAMR, true);
Timer TimeSA_initialData          ("SA initData",    Everything, DTIMESAMR, true);
Timer TimeSA_velocity             ("SA velocity",    Everything, DTIMESAMR, true);
Timer TimeSA_scalar               ("SA scalar",      Everything, DTIMESAMR, true);

Timer TimeLS_advectScalar1        ("LS advScalar1",  Everything, DTIMELS, true);
Timer TimeLS_advectScalar2        ("LS advScalar2",  Everything, DTIMELS, true);
Timer TimeLS_compUstar1           ("LS compUstar1",  Everything, DTIMELS, true);
Timer TimeLS_compUstar2           ("LS compUstar2",  Everything, DTIMELS, true);
Timer TimeLS_compUstarInit1       ("LS compUstarI1", Everything, DTIMELS, true);
Timer TimeLS_compUstarInit2       ("LS compUstarI2", Everything, DTIMELS, true);

// AMRNavierStokeInit.cpp
Timer TF_INITVORT    ("INITVORT",     Everything, DTIMEFORT, true);
Timer TF_COMPUTEVORT ("COMPUTEVORT",  Everything, DTIMEFORT, true);
Timer TF_INITVEL     ("INITVEL",      Everything, DTIMEFORT, true);
Timer TF_INITSCAL    ("INITSCAL",     Everything, DTIMEFORT, true);

// AMRNavierStokeVel.cpp
Timer TF_OPERATORLAP_ustar      ("ustar OLAP",  Everything, DTIMEFORT, true);
Timer TF_OPERATORLAP_ustarInit  ("ustarI OLAP", Everything, DTIMEFORT, true); //noel
Timer TF_UDELS_predictVelocities("UDELS pV",    Everything, DTIMEFORT, true);

// AMRNavierStokeScal.cpp
Timer TF_OPERATORLAP_advectDiffuseScalar("advect OLAP", Everything, DTIMEFORT, true);


// AMRNavierStokesRegrid.cpp
Timer TF_MAGVECT_Regrid("regrid magvect", Everything, DTIMEFORT, true);
