
#include "AMRNavierStokesHeaders.H"

#include "Timer.H"
extern Timer Everything;
Timer TMPI_computeUStarInit_cp      ("compUstarI cp", Everything, DTIMECOMM, true);
Timer TMPI_computeUStarInit_exchange("compUstarI ex", Everything, DTIMECOMM, true);

//#include "memusage.H"
//char crap[40];

// noel
// copy of above function to facilitate timing only!
// Call computeUStarInit during Setup, otherwise call computeUStar.
// make sure to delete and clean up when done using this.
// Should be exactly the same function above, except with Timers
// removed.
void
AMRNavierStokes::computeUStarInit(LevelData<FArrayBox>& a_uStar)
{

  bool isViscous = (s_nu > 0.0);
  Interval velComps(0,SpaceDim-1);

  if (!isViscous) {
    /// compute uStar = oldVel - dt*(u-dot-del-u)
    DataIterator dit = a_uStar.dataIterator();
    for (dit.reset(); dit.ok(); ++dit) {
      Real mult = -1.0*m_dt;
      a_uStar[dit()] *= mult;
      a_uStar[dit()] += oldVel()[dit()];
    }
  } else { 
    // viscous case is more elaborate -- need to do viscous solves.

    const DisjointBoxLayout& levelGrids = a_uStar.getBoxes();

    // these are parameters for TGA viscous solves
    Real a = 2.0 - sqrt(2.0) - TGA_EPS;
    Real discr = sqrt(a*a - 4.0*a + 2.0);
    Real r1 = (2.0*a - 1.0)/(a + discr);
    Real r2 = (2.0*a - 1.0)/(a - discr);  

    // set convergence metric to be norm of velocity field
    Interval allVelComps(0,SpaceDim-1);
    int normType = 0;
    Real velNorm = norm(oldVel(), allVelComps, normType);
    m_viscous_solver.setConvergenceMetric(velNorm, 0);
    
    // m_time is new time
    Real old_time = m_time - m_dt;

    // define temporary storage
    ////sprintf(crap, "compUstar1 mem=%-10.3f\n", get_memory_usage()); pout() << crap;
    LevelData<FArrayBox> gradPi(levelGrids, SpaceDim);
    LevelData<FArrayBox>* crseDataPtr = NULL;

    //LevelData<FluxBox> viscousFlux(levelGrids, SpaceDim);
    DataIterator dit = a_uStar.dataIterator();

    // these are necessary because solver can only handle one component
    // at a time.
    LevelData<FArrayBox> solverRHS(levelGrids, 1, 
                                   IntVect::TheUnitVector());
    LevelData<FArrayBox> intermediateVel(levelGrids,1,
                                         IntVect::TheUnitVector());

    LevelData<FArrayBox>* crseSolverDataPtr = NULL;

    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        //viscousSource[dit()].setVal(s_bogus_value);
        //tempStorage[dit()].setVal(s_bogus_value);
        solverRHS[dit()].setVal(s_bogus_value);
        intermediateVel[dit()].setVal(s_bogus_value);
        for (int dir=0; dir<SpaceDim; ++dir) {
          //viscousFlux[dit()][dir].setVal(s_bogus_value);
          //tempFlux[dit()][dir].setVal(s_bogus_value);
        }
      }
    }   

    // will need single-component QuadCFInterp to do fluxes,
    // and Laplacians
    QuadCFInterp singleCompCFInterp;

    if (m_level > 0) {
      const DisjointBoxLayout& crseBoxes = crseNSPtr()->newVel().getBoxes();
      crseDataPtr = new LevelData<FArrayBox>(crseBoxes, SpaceDim);
      crseSolverDataPtr = new LevelData<FArrayBox>(crseBoxes, 1);
      int nRefCrse = crseNSPtr()->refRatio();
      singleCompCFInterp.define(levelGrids, &crseBoxes, m_dx, 
                                nRefCrse, 1, m_problem_domain);
    }

    //Real viscousBeta; //not used?  (ndk)
    if (s_viscous_solver_type == TGA) {
      //viscousBeta = -r2*s_nu*m_dt; //not used? (ndk)
    } else {
      MayDay::Error("undefined viscous solver type");
    }

    // first get gradPi
    m_projection.gradPi(gradPi);
    
    // "f" is -u-del-u -gradPi +source
    for (dit.reset(); dit.ok(); ++dit) {
      a_uStar[dit()] += gradPi[dit()];
      a_uStar[dit()] *= -1.0;
    }

    LevelData<FArrayBox> tempStorage(levelGrids, SpaceDim,
                                     IntVect::TheUnitVector());

    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        tempStorage[dit()].setVal(s_bogus_value); 
      }
    }   

    // now compute diffused version of this
    if (crseDataPtr != NULL) {
      DataIterator crseDit = crseDataPtr->dataIterator();
      for (crseDit.reset(); crseDit.ok(); ++crseDit) {
        ((*crseDataPtr)[crseDit()]).setVal(0.0);
      }
    }

#ifdef MATCH_OLDCODE
    // this is a kluge to make new code agree with old code
    // C/F BC's are different -- here we can force them
    // to be extrap by being clever
    if (m_level > 0) {
      // this BC object will contain only extrapBC's 
      // and will be used at C/F interfaces
      DomainGhostBC bogusBC;
      // loop over directions
      for (int dir=0; dir<SpaceDim; dir++) {
        // loop over hi-lo
        SideIterator sit;
        for (sit.begin(); sit.ok(); ++sit) {
          HOExtrapBC thisBC(dir,sit(), velComps);
          bogusBC.setBoxGhostBC(thisBC);
        }
      }
    
      // now loop over grids
      for (dit.reset(); dit.ok(); ++dit) {
        Box thisGridBox = levelGrids[dit()];
        FArrayBox& thisData = a_uStar[dit()];
        bogusBC.applyHomogeneousBCs(thisData,thisGridBox,m_dx);
      }
      
    } // at this point, ALL ghost cells are set by extrap       
    
    // now set physical BC's on (-uDelU-GradPi)
    bool isViscous = (s_nu > 0.0);
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC(isViscous);

#ifdef MATCH_OLDCODE
    //velBC.isViscous(false);
#endif

    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);    
    TMPI_computeUStarInit_exchange.start();
    a_uStar.exchange(velComps);
    TMPI_computeUStarInit_exchange.stop();
    delete velBCPtr;

    // this computes L(-uDelU-GradPi) and puts it in tempStorage

    TF_OPERATORLAP_ustarInit.start();
    for (dit.reset(); dit.ok(); ++dit) {
      FORT_OLDOPERATORLAP(CHF_FRA(tempStorage[dit()]),
                          CHF_FRA(a_uStar[dit()]),
                          CHF_BOX(levelGrids.get(dit())),
                          CHF_CONST_REAL(m_dx));
    }
    TF_OPERATORLAP_ustarInit.stop();
#if 0
    {
      // put these in their own localized scope
      viscousBC viscBC;
      viscBC.applyBCs(tempStorage,m_problem_domain, m_dx);
      TMPI_computeUStarInit_exchange.start();
      tempStorage.exchange(tempStorage.interval());
      TMPI_computeUStarInit_exchange.stop();
    }
#endif

    LevelData<FluxBox> viscousFlux(levelGrids, SpaceDim);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        for (int dir=0; dir<SpaceDim; ++dir) {
          viscousFlux[dit()][dir].setVal(s_bogus_value); 
        }
      }
    }   

    // before we add in stuff, compute flux due to antidiffusive part
    // in this case, use the extrap C/F BC we've already set.
    // (this version of the levelGradientMac fn doesn't do C/F BC's)
    levelGradientMAC(viscousFlux, a_uStar, m_dx);

#else    

    // set physical BC's on (-uDelU-GradPi)
    basicVelBC* velBCPtr = m_physBCPtr->uDelUBC(isViscous);
    velBCPtr->applyBCs(a_uStar,m_problem_domain,m_dx);
    //a_uStar.exchange(velComps);
    delete velBCPtr;

    // this computes L(-uDelU-GradPi) and puts it in tempStorage
    computeLapVel(tempStorage, a_uStar, crseDataPtr);

    // before we add in stuff, compute flux due to antidiffusive part
    levelGradientMAC(viscousFlux, a_uStar, crseDataPtr, m_dx, 
                     m_velCFInterp);
#endif
    
    // multiply grad(forcing term) by mult, then initialize
    // and increment viscousFlux (will increment flux registers
    // all at once)
    // sign convention will be so that flux will have the opposite
    // sign when incrementing flux registers as the fluxes
    // computed in earlier case.  (same sign convention as in 
    // old code).

    Real mult = (0.5-a)*m_dt*s_nu;
    // this is experimental, for debugging purposes
    //mult = 0.0;
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisFlux = viscousFlux[dit()];
      for (int dir=0; dir<SpaceDim; dir++) {
        thisFlux[dir] *= mult;
      }
    }
        
    // source term we will use is (I + (half-a)*dt*nu*L)(-uDelU-gradPi)
    // this will be stored in a_uStar
    for (dit.reset(); dit.ok(); ++dit) {
      tempStorage[dit()] *= mult;
      a_uStar[dit()] += tempStorage[dit()];
    }
    
    // now need to get old-time velocity
    ///pout() << " fillVelocity vel6" << std::endl;
    fillVelocity(tempStorage, old_time);
    if (crseDataPtr != NULL) {
      crseNSPtr()->fillVelocity(*crseDataPtr, old_time);
    }
    
    velBCPtr = m_physBCPtr->viscousVelBC();
    velBCPtr->applyBCs(tempStorage, m_problem_domain, m_dx);
    delete velBCPtr;

    mult = -(1.0-a)*s_nu;
    // this is experimental, for debugging
    //mult = 0.0;

    // compute viscous flux due to old-time velocity
    {
      // need a second multicomponent LevelData<FluxBox>
      // to hold these fluxes -- goes out of scope and 
      // is deleted when we're done
      LevelData<FluxBox> oldViscFlux(levelGrids, SpaceDim);
      levelGradientMAC(oldViscFlux, tempStorage, crseDataPtr, 
                       m_dx, m_velCFInterp);
            
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = oldViscFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int dir = 0; dir<SpaceDim; dir++) {
          thisFlux[dir].mult(mult);
          thisViscFlux[dir] -= thisFlux[dir];
        }
      }
    } // temporary flux container goes out of scope here.
    
    LevelData<FArrayBox> viscousSource(levelGrids, SpaceDim);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        viscousSource[dit()].setVal(s_bogus_value);
      }
    }   

    // compute L(old_velocity)
    computeLapVel(viscousSource, tempStorage, crseDataPtr);   

    // now loop over boxes, multiply viscousSource by mult,
    // then add old-time velocity and viscousSource to 
    // the RHS for the first solve, which will all be in
    // a_uStar, and which is 
    // RHS = (I + (1-a)*nu*dt*L)u_n +dt*f,
    // where f = (I +(half-a)*dt*nu*L)(-uDelU - gradPi)

    mult = (1.0-a)*m_dt*s_nu;
    for (dit.reset(); dit.ok(); ++dit) {
      viscousSource[dit()] *= mult;
      a_uStar[dit()] *= m_dt;
      
      a_uStar[dit()] += viscousSource[dit()];
      a_uStar[dit()] += tempStorage[dit()];
    }

    // set up coarse-level BC
    Real intermediateTime = old_time + (1-r1)*m_dt;
    if (crseDataPtr != NULL) {
      ///pout() << " fillVelocity vel8" << std::endl;
      crseNSPtr()->fillVelocity(*crseDataPtr, intermediateTime);
    }


    // finally ready to do intermediate solve!
    // actually, don't need this one
    //a_uStar.exchange(velComps);

    LevelData<FluxBox> tempFlux(levelGrids, 1);
    if (s_set_bogus_values) {
      for (dit.reset(); dit.ok(); ++dit) {
        for (int dir=0; dir<SpaceDim; ++dir) {
          tempFlux[dit()][dir].setVal(s_bogus_value);
        }
      }
    }   

    // loop over directions
    for (int dir=0; dir<SpaceDim; dir++) {

      // copy RHS and coarse BC into single-component holder
      LevelData<FArrayBox>& crseBC = *crseSolverDataPtr;
      LevelData<FArrayBox>& crseVel = *crseDataPtr;
      for (dit.reset(); dit.ok(); ++dit) {
        solverRHS[dit()].copy(a_uStar[dit()],dir,0,1);
        // initialize soln to be RHS
        intermediateVel[dit()].copy(a_uStar[dit()],dir,0,1);
      }
      if (crseDataPtr != NULL) {
        DataIterator crseDit = crseVel.dataIterator();
        for (crseDit.reset(); crseDit.ok(); ++crseDit) {
          // copy coarse vel BC into single-component holder
          crseBC[crseDit()].copy(crseVel[crseDit()],dir,0,1);
        }
      }
      
      mult = -r2*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);            

      DomainGhostBC* thisVelBCPtr = m_physBCPtr->viscousSolveBC(dir);
      m_viscous_solver.setBC(*thisVelBCPtr);
      
      TimeLS_compUstarInit1.start();
      IterationFlavor = COUNTER_levelSolve_ustarinit1 + m_level;
      m_viscous_solver.levelSolve(intermediateVel, crseSolverDataPtr,
                                  solverRHS, false);    
      TimeLS_compUstarInit1.stop();

      // increment fluxes with flux due to intermediate velocity
      // set physical BC's -- this also relies on the fact that
      // we use the same velocity components for all directions
      // (however, need to use the x-direction DomainGhostBC
      // because of the single-component nature of intermediateVel
      
      //const DomainGhostBC& thisVelBC = velBC.DomGhostBC(dir);
      for (dit.reset(); dit.ok(); ++dit) {
        thisVelBCPtr->applyHomogeneousBCs(intermediateVel[dit()], 
                                          m_problem_domain, m_dx);
      }
    
      levelGradientMAC(tempFlux, intermediateVel, crseSolverDataPtr, 
                       m_dx, singleCompCFInterp);
      
      mult = -r2*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = tempFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
          thisFlux[fluxDir].mult(mult);
          thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
        }
      }
      
      // now ready to do second solve
      mult = -r1*m_dt*s_nu;
      m_viscous_solver.setBeta(mult);
      
      
      // fill coarse-level boundary condition...
      if (crseDataPtr != NULL) {
        ///pout() << " fillVelocity vel9" << std::endl;
        crseNSPtr()->fillVelocity(*crseSolverDataPtr, m_time,
                                  dir, 0, 1);
      }
      
      // initialize initial guess 
      for (dit.reset(); dit.ok(); ++dit) {
        solverRHS[dit()].copy(intermediateVel[dit()]);
      }

      // and solve away
      TimeLS_compUstarInit2.start();
      IterationFlavor = COUNTER_levelSolve_ustarinit2 + m_level;
      m_viscous_solver.levelSolve(solverRHS, crseSolverDataPtr, intermediateVel,
                                  false);      
      TimeLS_compUstarInit2.stop();

      // set physical BC's
      for (dit.reset(); dit.ok(); ++dit) {
        thisVelBCPtr->applyHomogeneousBCs(solverRHS[dit()], 
                                          m_problem_domain, m_dx);
      }
      
      // compute flux due to uStar
      levelGradientMAC(tempFlux, solverRHS, crseSolverDataPtr, 
                       m_dx, singleCompCFInterp);
      
      mult = -r1*s_nu;
      // experimental debugging change
      //mult = 0.0;
      for (dit.reset(); dit.ok(); ++dit) {
        FluxBox& thisFlux = tempFlux[dit()];
        FluxBox& thisViscFlux = viscousFlux[dit()];
        for (int fluxDir=0; fluxDir<SpaceDim; fluxDir++) {
          thisFlux[fluxDir].mult(mult);
          thisViscFlux[fluxDir].minus(thisFlux[fluxDir],0,dir,1);
        }

        // need to copy computed uStar into a_uStar
        a_uStar[dit()].copy(solverRHS[dit()],0,dir,1);
        
      } // end loop over grids

      delete thisVelBCPtr;

    } // end loop over directions

   
    // now increment flux registers
    for (dit.reset(); dit.ok(); ++dit) {
      FluxBox& thisViscFlux = viscousFlux[dit()];
      if (!finestLevel()) {
        assert(m_flux_register.isDefined());
        for (int dir=0; dir<SpaceDim; dir++) {
          // this is to make things look the same as old code
          //thisViscFlux[dir] *= m_dx;
          // use same multipliers as momentum fluxes, but
          // opposite sign convention
          Real fluxMult = m_dt;
          m_flux_register.incrementCoarse(thisViscFlux[dir],
                                          fluxMult, dit(), velComps,
                                          velComps, dir);
        }
      }
      
      if (m_level > 0) {
        LevelFluxRegister& crseFR = crseNSPtr()->m_flux_register;
        assert(crseFR.isDefined());
        // use same multipliers as momemtum fluxes, but opposite
        // sign convention
        Real fluxMult = m_dt;
        for (int dir=0; dir<SpaceDim; dir++) {
          // this is to make things look the same as old code
          //thisViscFlux[dir] *= m_dx;

          crseFR.incrementFine(thisViscFlux[dir],
                               fluxMult, dit(), velComps, 
                               velComps, dir, Side::Lo);
          
          crseFR.incrementFine(thisViscFlux[dir],
                               fluxMult, dit(), velComps, 
                               velComps, dir, Side::Hi);
        }
      }
    
    } // end loop over grids for flux register updates

    // final thing to do -- subtract off dt*gradPi term
    for (dit.reset(); dit.ok(); ++dit) {
      gradPi[dit()] *= m_dt;
      a_uStar[dit()] += gradPi[dit()];
    }
  
    
    // clean up temp storage
    if (crseDataPtr != NULL) {
      delete crseDataPtr;
      crseDataPtr = NULL;
    }
    
    if (crseSolverDataPtr != NULL) {
      delete crseSolverDataPtr;
      crseSolverDataPtr = NULL;
    }

  } // end if viscous

}
