// Example for Trapping Floating Point exceptions

// Not working for some reason?
// Starting with glibc 2.2, the following C99-style (but glibc specific) code is preferred. 
//#define _GNU_SOURCE 1
//
//extern "C" {
//#include <fenv.h>
//}
//static void __attribute__ ((constructor)) trapfpe(void)
//{
//  /* Enable some exceptions.  At startup all exceptions are masked. */
//  feenableexcept(FE_INVALID|FE_DIVBYZERO|FE_OVERFLOW);
//}

// This one works...
// Previous versions of glibc require the following code:
extern "C" {
#include <fpu_control.h>
}
/* IM: Invalid operation mask
 * DM: Denormalized operand mask 
 * ZM: Zero-divide mask
 * OM: Overflow mask 
 * UM: Underflow mask 
 * PM: Precision (inexact result) mask */
static void __attribute__ ((constructor)) trapfpe(void)
{
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_UM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM);
  //fpu_control_t cw = _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_UM);
  fpu_control_t cw = _FPU_DEFAULT;
  _FPU_SETCW(cw);
   /* On x86, this expands to: */
   /* unsigned int cw = 0x037f & ~(0x01 | 0x04 | 0x08); */
   /* __asm__ ("fldcw %0" : : "m" (*&cw));              */ 
}


#include <cstdio>
#include <cfloat>


void func(void) {

  printf(" entering func\n");
  double a = 1.0e-300;

  printf(" %le\n", a/1.0);
  printf(" %le\n", a/2.0);

  for(int i=1; i<1000; i+=50) {
    printf("%d %le\n", i, a);
    a /= (double)i;
  }
  
  printf("%le\n", a);

  for(int i=0; i<2; i++) {
    printf("%d %le\n", i, a);
    a /= (double)i;
  }
  printf(" leaving func\n");
}

int main() {

  printf("DBL_MIN = %le\n", DBL_MIN);
  printf("DBL_MAX = %le\n", DBL_MAX);

  //func();

  trapfpe();
  
  func();

}
