/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "PatchAdvection.H"
#include "PatchAdvectionF_F.H"

#include "OldLoHiCenter.H"
#include "LoHiSide.H"

#include "Timer.H"
extern Timer Everything;
Timer TF_NORMALPREDF  ("NORMALPREDF",       Everything, DTIMEFORT, true);
Timer TF_RIEMANNF2    ("RIEMANNF2",         Everything, DTIMEFORT, true);
Timer TF_UPDATECONSADF("UPDATECONSADF",     Everything, DTIMEFORT, true);

PatchAdvection::PatchAdvection():OldPatchGodunov()
{
}

// Factory method - this object is its own factory.  It returns a pointer
// to new OldPatchGodunov object with its initial and boundary condtions, slope
// parameters, and artificial viscosity information defined.
OldPatchGodunov* PatchAdvection::new_patchGodunov() const
{
  // Make the new object
  PatchAdvection* newPtr = new PatchAdvection();
  
  // Pass the current initial and boundary condition (IBC) object to this new
  // patch integrator so it can create its own IBC object and define it as it
  // needs to (i.e. new domain and grid spacing if necessary)
  newPtr->setPhysIBC(getPhysIBC());
  
  // Set the slope and artificial viscosity parameters
  newPtr->setSlopeParameters(m_useFourthOrderSlopes,
                             m_useFlattening,
                             m_limitSlopes);
  newPtr->setArtificialViscosity(m_useArtificialViscosity,
                                 m_artificialViscosity);
  
  // copy state info
  newPtr->m_numCons = m_numCons;
  newPtr->m_numPrim = m_numPrim;
  newPtr->m_numFlux = m_numFlux;
  newPtr->m_numSlope = m_numSlope;

  OldPatchGodunov* retval = 
    static_cast<OldPatchGodunov*>(newPtr);

  // Return the new object
  return retval;
}

// void PatchAdvection::computeFluxes(FArrayBox&       a_U,
//                                    FArrayBox        a_F[CH_SPACEDIM],
//                                    const FArrayBox& a_S,
//                                    const Real&      a_dt,
//                                    const Box&       a_box)
// {
//   assert(isDefined());
//   assert(a_box == m_currentBox);

//   // Get the number of various variables
//   int numFlux = numFluxes();
//   int numPrim = numPrimitives();

//   // The current box
//   Box curBox = m_currentBox;

//   // Boxes for face centered state - used for the riemann() and
//   // artificialViscosity() calls
//   Box faceBox[SpaceDim];

//   // Boxes for face centered fluxes
//   Box fluxBox[SpaceDim];

//   // Boxes for cell centered state - used for the updatePrim() calls
//   Box ccBox[SpaceDim];

//   for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
//   {
//     // This box is face centered in direction "dir1", is one bigger than the
//     // input box in all directions except "dir1" and stays one cell away from
//     // the domain boundary in "dir1"
//     faceBox[dir1] = curBox;
//     faceBox[dir1].grow(1);
//     faceBox[dir1] &= m_domain;
//     faceBox[dir1].grow(dir1,-1);
//     faceBox[dir1].surroundingNodes(dir1);

//     // This box is face centered in direction "dir1", is one bigger than the
//     // input box in all directions except "dir1"
//     fluxBox[dir1] = curBox;
//     fluxBox[dir1].grow(1);
//     fluxBox[dir1].grow(dir1,-1);
//     fluxBox[dir1] &= m_domain;
//     fluxBox[dir1].surroundingNodes(dir1);

//     assert (a_F[dir1].box().contains(fluxBox[dir1]));


//     // This is the cell centered analog of "fluxBox"
//     ccBox[dir1] = curBox;
//     ccBox[dir1].grow(1);
//     ccBox[dir1].grow(dir1,-1);
//     ccBox[dir1] &= m_domain;
//   }

//   // Primitive variables
//   Box WBox = a_U.box();
//   WBox &= m_domain;

//   FArrayBox W(WBox,numPrim);

//   // Calculate the primitive variables from the conserved variables
//   consToPrim(W,a_U,WBox);

//   // Define the box where slopes will be needed (one larger than the final
//   // update box)
//   Box slopeBox = curBox;
//   slopeBox.grow(1);
//   slopeBox &= m_domain;

//   // Compute flattening once for all slopes if needed
//   FArrayBox flattening;
//   if (useFlattening())
//   {
//     flattening.define(slopeBox,1);
//     computeFlattening(flattening,W,slopeBox);
//   }

//   // Intermediate, extrapolated primitive variables
//   FArrayBox WMinus[SpaceDim];
//   FArrayBox WPlus [SpaceDim];

//   // Initial fluxes
//   FArrayBox F1[SpaceDim];

//   // Compute initial fluxes
//   for (int dir1 = 0; dir1 < SpaceDim; dir1++)
//   {
//     // Size the intermediate, extrapolated primitive variables
//     WMinus[dir1].resize(slopeBox,numPrim);
//     WPlus [dir1].resize(slopeBox,numPrim);

//     // Compute slopes
//     int numSlope = numSlopes();
//     FArrayBox dW(slopeBox,numSlope);

//     slope(dW,W,flattening,dir1,slopeBox);

//     // Compute predictor step to obtain extrapolated primitive variables
//     normalPred(WMinus[dir1],WPlus[dir1],W,dW,a_dt/m_dx,dir1,slopeBox);

//     // If the source term is valid add it to the primitive quantities
//     if (!a_S.box().isEmpty())
//     {
//       incrementWithSource(WMinus[dir1],a_S,0.5*a_dt,slopeBox);
//       incrementWithSource(WPlus [dir1],a_S,0.5*a_dt,slopeBox);
//     }

//     // Solve the Riemann problem and get face-centered values
//     F1[dir1].resize(fluxBox[dir1],numFlux);
//     riemann(F1[dir1],WPlus[dir1],WMinus[dir1],dir1,faceBox[dir1]);

//     // Use the user supplied PhysBC object to obtain boundary values
//     m_bc->fluxBC(F1[dir1],W,WMinus[dir1],dir1,Side::Lo,m_currentTime);
//     m_bc->fluxBC(F1[dir1],W,WPlus [dir1],dir1,Side::Hi,m_currentTime);
//   }

// #if (CH_SPACEDIM == 3)
//   // In 3D, compute some additional intermediate fluxes
//   //
//   // NOTE:  The diagonal entries of this array of fluxes are not
//   // used and will not be defined.
//   FArrayBox F2[SpaceDim][SpaceDim];
  
//   // Compute the intermediate, corrected fluxes in each direction
//   for (int dir1 = 0; dir1 < SpaceDim; dir1++)
//   {
//     // Correct fluxes using fluxes from a different direction
//     for (int dir2 = 0; dir2 < SpaceDim; dir2++)
//     {
//       // A different direction has been found
//       if (dir2 != dir1)
//       {
//         // Temporary primitive variables
//         FArrayBox WMinusTemp(WMinus[dir1].box(),numPrim);
//         FArrayBox WPlusTemp (WPlus [dir1].box(),numPrim);

//         // Copy data for in place modification
//         WMinusTemp.copy(WMinus[dir1]);
//         WPlusTemp .copy(WPlus [dir1]);

//         // Update the current, extrapolated primitive variable using a flux
//         // in a different direction
//         updatePrim(WMinusTemp,WPlusTemp,F1[dir2],(1.0/3.0)*a_dt/m_dx,dir2,ccBox[dir2]);

//         // Solve the Riemann problem and get fluxes
//         F2[dir1][dir2].resize(fluxBox[dir1],numFlux);
//         riemann(F2[dir1][dir2],WPlusTemp,WMinusTemp,dir1,faceBox[dir1]);

//         // Use the user supplied PhysBC object to obtain boundary values
//         m_bc->fluxBC(F2[dir1][dir2],W,WMinusTemp,dir1,Side::Lo,m_currentTime);
//         m_bc->fluxBC(F2[dir1][dir2],W,WPlusTemp ,dir1,Side::Hi,m_currentTime);
//       }
//     }
//   }
// #endif

//   // faceBox and fluxBox are now a bit smaller for the final corrections
//   for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
//     {
//       faceBox[dir1] = curBox;
//       faceBox[dir1].grow(dir1,1);
//       faceBox[dir1] &= m_domain;
//       faceBox[dir1].grow(dir1,-1);
//       faceBox[dir1].surroundingNodes(dir1);
      
//       fluxBox[dir1] = curBox;
//       fluxBox[dir1].surroundingNodes(dir1);
//     }

//   // Do the final corrections to the fluxes
//   for (int dir1 = 0; dir1 < SpaceDim; dir1++)
//   {
//     // Correct the flux using fluxes in the remaining direction(s)
//     for (int dir2 = 0; dir2 < SpaceDim; dir2++)
//     {
//       // A different direction has been found
//       if (dir2 != dir1)
//       {
// #if (CH_SPACEDIM == 2)
//         // In 2D, the current primitive state is updated by a flux in
//         // the other direction
//         updatePrim(WMinus[dir1],WPlus[dir1],F1[dir2],(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
// #elif (CH_SPACEDIM == 3)
//         // In 3D, find a direction different from the two above
//         int dir3 = 3 - dir1 - dir2;

//         // Update the conservative state using both corrected fluxes in
//         // the other two directions
//         updatePrim(WMinus[dir1],WPlus[dir1],F2[dir2][dir3],(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
// #else
//         // Only 2D and 3D should be possible
//         MayDay::Error("LevelGodunov::step - CH_SPACEDIM not 2 or 3!");
// #endif
//       }
//     }

//     // Solve the Riemann problem and get face-centered values; these will
//     // be returned
//     FArrayBox& F = a_F[dir1];
//     // do setval here to prevent uninitialized memory read on some machines
//     F.setVal(666.666);
//     riemann(F,WPlus[dir1],WMinus[dir1],dir1,faceBox[dir1]);

//     // Use the user supplied PhysBC object to obtain boundary values
//     m_bc->fluxBC(F,W,WMinus[dir1],dir1,Side::Lo,m_currentTime);
//     m_bc->fluxBC(F,W,WPlus [dir1],dir1,Side::Hi,m_currentTime);

//     // Apply artificial viscosity if reqested
//     if (useArtificialViscosity())
//     {
//       // Compute the divergence of the velocity for artificial viscosity
//       FArrayBox divVelDir;

//       divVelDir.define(fluxBox[dir1],1);
//       divVelDir.setVal(0.0);

//       divVel(divVelDir,W,dir1,slopeBox);

//       // Apply artificial viscosity to the fluxes
//       artificialViscosity(F,a_U,divVelDir,dir1,faceBox[dir1]);

//       // Use the user supplied PhysBC object to apply artificial viscosity
//       // to the boundary fluxes
//       m_bc->artViscBC(F,a_U,divVelDir,dir1,m_currentTime);
//     }
//   }
// }


// Compute the maximum wave speed
Real PatchAdvection::getMaxWaveSpeed(const FArrayBox& a_U,
                                     const Box&       a_box)
{
  assert(isDefined());

  // note that for this advection problem, the max wavespeed
  // is based entirely on the cell-centered velocity
  Real maxSpeed, speed = 0.0;
  for (int dir=0; dir<SpaceDim; dir++)
    {
      FArrayBox& vel = *m_cellVelPtr;
      speed = vel.norm(0,dir,1);
      if (maxSpeed < speed) maxSpeed = speed;
    }
  
  return maxSpeed;
}

// Number of conserved variables
int PatchAdvection::numConserved()
{
  return m_numCons;
}

// Names of the conserved variables
Vector<string> PatchAdvection::stateNames()
{
  // just hack around this for now
  Vector<string> retval(m_numPrim, "scalar");

  return retval; 
}

// Number of flux variables
int PatchAdvection::numFluxes()
{
  // In some computations there may be more fluxes than conserved variables
  return m_numFlux;
}

// Number of primitive variables
int PatchAdvection::numPrimitives()
{
  // This doesn't equal the number of conserved variables because
  // auxiliary/redundant variable may be computed and stored
  return m_numPrim;
}

//  Number of primitive variables for which slopes are computed
int PatchAdvection::numSlopes()
{
  // This may be less than the number of primitive variables for the
  // reason given in numPrimitives() comments
  return m_numSlope;
}

// Compute the primitive variables from the conserved variables
void PatchAdvection::consToPrim(FArrayBox&       a_W,
                                const FArrayBox& a_U,
                                const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.box().contains(a_box));
  assert(a_W.box().contains(a_box));
  
  // since the primitive and conserved variables are the same in 
  // this set of equations, even I can do this one...
  a_W.copy(a_U, a_box);
  
  
}

// Compute the conserved variables from the primitive variables
void PatchAdvection::primToCons(FArrayBox&       a_U,
                                const FArrayBox& a_W,
                                const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.box().contains(a_box));
  assert(a_W.box().contains(a_box));

  // since primitive variables are conserved in this problem,
  // this is easy
  a_U.copy(a_W, a_box);


}

// Extrapolate the primitive variables to the cell faces
void PatchAdvection::normalPred(FArrayBox&       a_WMinus,
                                FArrayBox&       a_WPlus,
                                const FArrayBox& a_W,
                                const FArrayBox& a_dW,
                                const Real&      a_scale,
                                const int&       a_dir,
                                const Box&       a_box)
{
  assert(isDefined());

  assert (m_cellVelPtr != NULL);
  FArrayBox& cellVel = *m_cellVelPtr;
  assert(cellVel.box().contains(a_box));

  TF_NORMALPREDF.start();
  FORT_NORMALPREDF(CHF_FRA(a_WMinus),
                   CHF_FRA(a_WPlus),
                   CHF_CONST_FRA(a_W),
                   CHF_CONST_FRA(a_dW),
                   CHF_CONST_FRA(cellVel),
                   CHF_CONST_REAL(a_scale),
                   CHF_CONST_INT(a_dir),
                   CHF_BOX(a_box));
  TF_NORMALPREDF.stop();
}

// increment with source term
void PatchAdvection::incrementWithSource(FArrayBox& a_W,
                                         const FArrayBox& a_S,
                                         const Real& a_scale,
                                         const Box& a_box)
{
  assert (isDefined());

  assert (a_W.box().contains(a_box));
  assert (a_S.box().contains(a_box));
  // may eventually want to have version which only acts on a subset
  // of components, but for now, this is required (can always set non-active
  // components to 0):
  assert (a_S.nComp() == a_W.nComp());

  // now we just add in the source term, right? (do i need to do 
  // cons->prim, increment w/ source, prim->cons? -- in this case, since
  // cons == prim, it's not important, but as an example, it would be
  // useful to actually have it correct in the code.
  // 

  // for the moment, skip the cons->prim->cons conversion
  // would be nice if FArrayBox had a scaleMult function which
  // worked on a sub-box (scale-mult functions only work on
  // entire domain) -- this is a problem because of the const-ness of 
  // a_S 
  // this is downright silly
  FArrayBox tempSrc(a_S.box(), a_S.nComp());
  tempSrc.copy(a_S);
  tempSrc *= a_scale;
  int startComp = 0;
  int numComp = a_S.nComp();
  a_W.plus(tempSrc, a_box, startComp, startComp, numComp);

  
}


// Compute a Riemann problem and generate fluxes at the faces
void PatchAdvection::riemann(FArrayBox&       a_F,
                             const FArrayBox& a_WLeft,
                             const FArrayBox& a_WRight,
                             const int&       a_dir,
                             const Box&       a_box)
{
  assert(isDefined());
  
  assert(a_F.box().contains(a_box));
  
  assert(m_advVelPtr != NULL);
  FluxBox& advVel = *m_advVelPtr;
  //assert(advVel.box().contains(a_box));


  // Get the numbers of relevant variables
  int numPrim = numPrimitives();
  int numFlux = numFluxes();

  assert(a_F.nComp() == numFlux);
  assert(a_WLeft.nComp() == numPrim);
  assert(a_WRight.nComp() == numPrim);

  // Cast away "const" inputs so their boxes can be shifted left or right
  // 1/2 cell and then back again (no net change is made!)
  FArrayBox& shiftWLeft  = (FArrayBox&)a_WLeft;
  FArrayBox& shiftWRight = (FArrayBox&)a_WRight;

  // Solution to the Riemann problem
  FArrayBox Wgdnv(a_box,numPrim);

  // Shift the left and right primitive variable boxes 1/2 cell so they are
  // face centered
  shiftWLeft .shiftHalf(a_dir, 1);
  shiftWRight.shiftHalf(a_dir,-1);
  
  assert(shiftWLeft.box().contains(a_box));
  assert(shiftWRight.box().contains(a_box));

  FArrayBox& advVelDir = advVel[a_dir];
  assert(advVelDir.box().contains(a_box));
  
  // Riemann solver computes Wgdnv all edges that are not on the physical
  // boundary.
  TF_RIEMANNF2.start();
  FORT_RIEMANNF(CHF_FRA(Wgdnv),
                CHF_CONST_FRA(shiftWLeft),
                CHF_CONST_FRA(shiftWRight),
                CHF_CONST_FRA1(advVelDir,0),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
  TF_RIEMANNF2.stop();
  // Shift the left and right primitive variable boxes back to their original
  // position
  shiftWLeft .shiftHalf(a_dir,-1);
  shiftWRight.shiftHalf(a_dir, 1);
  
  a_F.copy(Wgdnv);
}

// Update the primitive variables using face-centered variables
void PatchAdvection::updatePrim(FArrayBox&       a_WMinus,
                                FArrayBox&       a_WPlus,
                                const FArrayBox& a_F,
                                const Real&      a_scale,
                                const int&       a_dir,
                                const Box&       a_box)
{
  assert(isDefined());
  assert(a_WPlus.box().contains(a_box));
  assert(a_WMinus.box().contains(a_box));

  // Number of conserved variables
  int numCons = numConserved();

  // Temporary storage for converting primitive variables to conserved
  // variables and back again (after update with the fluxes)
  FArrayBox U(a_box,numCons);

  // for W+-, convert primitive quantities to conservative quantities,
  // update with the fluxes, and convert back
  primToCons(U,a_WMinus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WMinus,U,a_box);

  primToCons(U,a_WPlus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WPlus,U,a_box);
}

// Update the conserved variable using fluxes and a scaling factor
void PatchAdvection::updateCons(FArrayBox&       a_U,
                                const FArrayBox& a_F,
                                const Real&      a_scale,
                                const int&       a_dir,
                                const Box&       a_box)
{
  assert(isDefined());
  
  assert (cellVelPtr() != NULL);

  const FArrayBox& cellVel = *cellVelPtr();
  assert (cellVel.box().contains(a_box));

  TF_UPDATECONSADF.start();
  FORT_UPDATECONSADF(CHF_FRA(a_U),
                     CHF_CONST_FRA(a_F),
                     CHF_CONST_FRA(cellVel),
                     CHF_CONST_REAL(a_scale),
                     CHF_CONST_INT(a_dir),
                     CHF_BOX(a_box));
  TF_UPDATECONSADF.stop();
}

// Update the conserved variable using fluxes and a scaling factor
void PatchAdvection::finalUpdate(FArrayBox&       a_U,
                                 const FArrayBox& a_F,
                                 const Real&      a_scale,
                                 const int&       a_dir,
                                 const Box&       a_box)
{
  // in this case, we just call updateCons..
  updateCons(a_U, a_F, a_scale, a_dir, a_box);
}



// Interval within the primitive variables corresponding to the velocities
// since this doesn't apply to this set of equations, return a bogus interval
Interval PatchAdvection::velocityInterval()
{
  Interval retval(-1,-1);

  return retval;
}

// Component index within the primitive variables of the pressure
// since this doesn't apply to this set of equations, return a bogus value
int PatchAdvection::pressureIndex()
{
  return -1;
}

// Component index within the primitive variables of the bulk modulus
// since this doesn't apply to this set of equations, return a bogus value
int PatchAdvection::bulkModulusIndex()
{
  return -1;
}

// set advection velocity pointer
void 
PatchAdvection::setAdvVelPtr(FluxBox* a_advVelPtr)
{
  m_advVelPtr = a_advVelPtr;
}


// access advection velocity
const FluxBox* 
PatchAdvection::advectionVelPtr() const
{
  return m_advVelPtr;
}

// set cell-centered velocity pointer
void 
PatchAdvection::setCellVelPtr(FArrayBox* a_cellVelPtr)
{
  m_cellVelPtr = a_cellVelPtr;
}


// access advection velocity
const FArrayBox* 
PatchAdvection::cellVelPtr() const
{
  return m_cellVelPtr;
}


// set number of variables
void
PatchAdvection::setNumVar(int a_numVar)
{
  // note that numCons == numPrim == numFlux == numVar
  // in this case.
  m_numCons = a_numVar;
  m_numPrim = a_numVar;
  m_numCons = a_numVar;
  m_numFlux = a_numVar;
  m_numSlope = a_numVar;
}
  
