/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// PoissonOp by DTGraves, Weds, July 21, 1999
// HyprePoissonOp by petermc, 20 Feb 2003

#ifndef HYPREPOISSONOP_H
#define HYPREPOISSONOP_H

#include "PoissonOp.H"

///  version of poisson operator that uses hypre for bottom solver
/** this class contains stuff to solve  div (grad phi) = rhs
 */
class HyprePoissonOp: public PoissonOp
{
public:

  ///
  /**
     virtual constructor workaround
  */
  virtual LevelOp* new_levelop() const;

  ///
  HyprePoissonOp();

  ///
  ~HyprePoissonOp();

  ///
  void clearMemory();

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& Ba,
              const DisjointBoxLayout* base_ba,
              Real  DxLevel, 
              int refratio,
              const Box& domf,
              bool a_homogeneousOnly=false,
              int ncomp = 1);

  ///
  /**
     Full define function.  makes all coarse-fine
     information and sets internal variables 
  */
  void define(
              const DisjointBoxLayout& Ba,
              const DisjointBoxLayout* base_ba,
              Real  DxLevel, 
              int refratio,
              const ProblemDomain& domf,
              bool a_homogeneousOnly=false,
              int ncomp = 1);

  ///
  /**
     Full define function.  define
     from finer levelop.
  */
  void define(
              const LevelOp* opfine,
              int refratio);


private:
  //internally useful functions

  void setHypreValues();

};

#endif
