/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// from NoOpSmoother by DFMartin, Sun, May 5, 2002
// petermc, 10 Jan 2003

#ifndef _HypreSmoother_H_
#define _HypreSmoother_H_

#include <iostream>
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"
#include "BaseBottomSmoother.H"
#include "FluxBox.H"
#include "RealVect.H"

/// Implements hypre solver as a bottom smoother for LevelOp-derived classes
/** The HypreSmoother class implements the hypre solver for 
    LevelOp-derived classes
*/ 
class HypreSmoother : public BaseBottomSmoother
{

  struct HYPRE_impl;

public:

  /// constructor, destructor


  /// default constructor
  HypreSmoother();

  ///  destructor
  virtual ~HypreSmoother();

  ///
  /** this gets around the "no virtual constructor" rule
   */
  virtual BaseBottomSmoother* new_bottomSmoother() const;


  ///
  void clearMemory();

  ///
  void clear_hypre();

  ///
  /** set up solver
   */
  void setup_solver(const DisjointBoxLayout& a_grids, 
                    const DisjointBoxLayout* a_baseGrids, 
                    const ProblemDomain& a_domain, 
                    const RealVect& a_dxLevel,
                    int a_nRefCrse,
                    int a_solver_flag);

  ///
  /** set up matrix in hypre
   */
  void setup_matrix();

  /// solve at this level
  /** solve
  */
  virtual void doBottomSmooth(LevelData<FArrayBox>& a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              LevelOp* a_levelop_ptr);

protected:

  // problem domain at this level
  ProblemDomain m_domain;

  //
  HYPRE_impl* m_impl;

  // flag:  0 for SMG, 1 for PFMG
  int m_solver_flag;

  //
  bool m_verbose;

  // have you filled in the matrix?
  // bool m_initializedSolver;

  // do you have m_acoefs and m_bcoefs?
  bool m_initializedHypre;

  LevelData<FArrayBox>* m_acoefs;

  // LevelData<FArrayBox>* m_bcoefs[CH_SPACEDIM];
  LevelData<FluxBox>* m_bcoefs;

  Real m_alpha;
  
  Real m_beta;
 
  Real m_relTolerance;

  int m_numPre;

  int m_numPost;

  // grids at this level
  DisjointBoxLayout  m_grids;

  //
  RealVect m_dxLevel;

  int m_nRefCrse;

  int m_maxIter;

private:

   ///
  /** Transpose a_fab to a_array.
   */
  void transpose(Real* a_array, const FArrayBox& a_fab) const;
};

#endif
